/*****************************************************************************
*
*	Title:		%M%
*
*	Synopsis:	Display routines.
*
******************************************************************************
*
*	Filename:		%P%
*
*	Author:			Danial Lawrence
*
*	Creation Date:		29/04/91 09:13
*
*	Modification date:	%G% : %U%	<990704.2238>
*
*	Current rev:		%I%
*
*	Special Comments:
*
*	Contents Description:
*
*
* The functions in this file handle redisplay. There are two halves, the
* ones that update the virtual display screen, and the ones that make the
* physical display screen the same as the virtual display screen. These
* functions use hints that are left in the windows by the commands.
*
******************************************************************************
* 
* (C)opyright 1987 by Daniel M. Lawrence
* MicroEMACS 3.8 can be copied and distributed freely for any
* non-commercial purposes. MicroEMACS 3.8 can only be incorporated
* into commercial software with the permission of the current author.
* 
* Modifications to the original file by Jasspa. 
* 
* Copyright (C) 1988 - 1999, JASSPA 
* The MicroEmacs Jasspa distribution can be copied and distributed freely for
* any non-commercial purposes. The MicroEmacs Jasspa Distribution can only be
* incorporated into commercial software with the expressed permission of
* JASSPA.
* 
****************************************************************************/

/*---	Include defintions */

#define	__DISPLAYC				/* Name file */

/*---	Include files */

#include "emain.h"
#include "eskeys.h"		/* External Defintions */

#ifdef _STDARG
#include <stdarg.h>		/* Variable Arguments */
#endif

#if (defined _UNIX) || (defined _DOS) || (defined _WIN32)
#include <sys/types.h>
#include <time.h>
#endif

#ifdef _DOS
#include <pc.h>
#endif

/*---	Local macro definitions */

/*---	External references */

/*---	Local type definitions */

/*
 * Set the virtual cursor to the specified row and column on the virtual
 * screen. There is no checking for nonsense values; this might be a good
 * idea during the early stages.
 */
/*---	Local variable definitions */

VVIDEO        vvideo;                   /* Head of the virtual video list */
LINE          *mline;                   /* message line. */
unsigned short TTsrow = 0;              /* Terminal starting row */

/* Poke screen flags
 * Poke screen maintains the rectangular extents of the region of the 
 * screen that has been damaged. These screen areas are fixed on the next
 * invocation to update(). Note that we only initialise the flag, the other
 * fields should be automatically initialised since we start in a garbaged
 * state.
 * 
 * Note that all poke extents are 'inclusive' values.
 */
static void pokeUpdate (void);          /* The poke screen update function */
static char poke_flag = 0;              /* Boolean TRUE/FALSE flag. TRUE
                                         * when a poke operation has been 
                                         * performed. */
static int poke_xmin;                   /* Minimum column extent of poke   */
static int poke_xmax;                   /* Maximum column extent of poke   */
static int poke_ymin;                   /* Minimum row extent of poke      */
static int poke_ymax;                   /* Maximum row extent of poke      */ 

/*
 * Initialize the data structures used by the display code. The edge vectors
 * used to access the screens are set up. The operating system's terminal I/O
 * channel is set up. All the other things get initialized at compile time.
 * The original window has "WFCHG" set, so that it will get completely
 * redrawn on the first call to "update".
 */
static unsigned char defScheme [2*HCTNUM] =
{
    TTdefFCol, TTdefBCol,           /* Foreground */
    TTdefFCol, TTdefBCol,           /* Current foreground */
    TTdefBCol, TTdefFCol,           /* Region foreground */
    TTdefBCol, TTdefFCol,           /* Current region foreground */
    TTdefBCol, TTdefFCol,           /* Foreground */
    TTdefBCol, TTdefFCol,           /* Current foreground */
    TTdefFCol, TTdefBCol,           /* Region foreground */
    TTdefFCol, TTdefBCol,           /* Current region foreground */
};

void
vtinit(void)
{
    FRAMELINE *flp;                     /* Frame store line pointer */
    int ii, jj ;                        /* Local loop counters */
    unsigned short fsCol, *col ;        /* Frame store colour and pointer to */ 

    if (TTstart() == FALSE)             /* Started ?? */
        meExit (1);                     /* No quit */

    if(((colSchmTbl = (unsigned char *) malloc(2*HCTNUM)) == NULL) ||
#ifdef _TCAPFONT
       ((fntSchmTbl = (unsigned char *) malloc(2*HCTNUM)) == NULL) ||
#endif
       ((vvideo.video = (VIDEO *) malloc((TTmrow)*sizeof(VIDEO))) == NULL) ||
       ((hilCol = (HILCOLOR *) malloc (hilColS*sizeof(HILCOLOR))) == NULL) ||
       ((disLineBuff = (unsigned char *) malloc((disLineSize+32)*sizeof(unsigned char))) == NULL) ||
       ((mline   = (LINE *)  malloc(sizeof(LINE)+TTmcol)) == NULL) ||
       ((mlStore = (char *)  malloc(TTmcol+1)) == NULL))
        meExit(1);
    
    memcpy(colSchmTbl,defScheme,2*HCTNUM);
#ifdef _TCAPFONT
    memset(fntSchmTbl,0,HCTNUM);
    memset(fntSchmTbl+HCTNUM,8,HCTNUM);
#endif
    /* Initialise the virtual video structure. Allocate and reset to 
     * zero. */
    memset(vvideo.video,0,TTmrow*sizeof(VIDEO)) ;
    memset(mline,0,sizeof(LINE)) ;
    vvideo.video [TTnrow].flag = VFMESSL ;
    vvideo.video [TTnrow].line = mline ;
    /* Reset the video list block */
    vvideo.next = NULL;                 /* No next block */
    vvideo.window = NULL;               /* No windows attached */
    
    /* Set the fore and back colours */
    TTaddColor(TTdefFCol,255,255,255) ;
    TTaddColor(TTdefBCol,0,0,0) ;
    
    /* Set up the size of the line */
    mline->l_size = TTncol;             /* Set the length of the line */
    mlStore [0] = '\0';                 /* Reset the length of the store */
    
    /* Frame Store storage
     * The frame store hold's 'n' lines of video information.
     * The frameStore is an array of points to FRAMELINE structures
     * which hold the physical line contents.
     * 
     * Allocate lines in the frame store to hold the video information.
     */
    if ((frameStore = (FRAMELINE *) meMalloc (sizeof (FRAMELINE) * TTmrow)) == NULL)
        meExit (1);
    fsCol = FScolourPair(TTdefFCol,TTdefBCol) ;
    for (flp = frameStore, ii = 0; ii < TTmrow; ii++, flp++)
    {
        if (((flp->text = meMalloc(sizeof(unsigned char) * TTmcol)) == NULL) ||
            ((flp->colour = meMalloc(sizeof(unsigned short) * TTmcol)) == NULL))
            meExit (1);
        /* Fill with data */
        memset (flp->text, ' ', sizeof (char) * TTmcol);
        for(col=flp->colour, jj=0 ; jj<TTmcol ; jj++)
            *col++ = fsCol ;
    }
    
    /* Some windowing systems require a 2 stage start. Kick off the second
     * stage of the start up process after the display memory has been 
     * initialised */
    TTstartStage2();
}

/*
 * Apply selection hilighting to the window
 */
void
shilightWindow(WINDOW *wp)
{
    long lineNo;           /* Toprow line number */
    
    /* Check to see if any selection hilighting is to be removed. */
    if(((selhilight.flags & SELHIL_ACTIVE) == 0) || 
       (wp->w_bufp != selhilight.bp))
        goto remove_hilight;
    
    if((wp->w_flag & (WFSELHIL|WFREDRAW)) == 0)
        return ;
    
    /* Hilight selection has been modified. Work out the start and
     * end points
     */
    if (selhilight.flags & SELHIL_CHANGED)
    {
        selhilight.flags &= ~(SELHIL_CHANGED|SELHIL_SAME);
        
        /* dot ahead of mark line */
        if (selhilight.dlineno > selhilight.mlineno)
        {
            selhilight.sline = selhilight.mlineno;
            selhilight.eline = selhilight.dlineno;
            selhilight.soff = selhilight.mlineoff;
            selhilight.eoff = selhilight.dlineoff;
        }
        /* dot behind mark line */
        else if (selhilight.dlineno < selhilight.mlineno)
        {
            selhilight.sline = selhilight.dlineno;
            selhilight.eline = selhilight.mlineno;
            selhilight.soff = selhilight.dlineoff;
            selhilight.eoff = selhilight.mlineoff;
        }
        /* dot and mark on same line */        
        else
        {
            selhilight.sline = selhilight.mlineno;
            selhilight.eline = selhilight.mlineno;
            /* dot ahead of mark */
            if (selhilight.mlineoff > selhilight.dlineoff)
            {
                selhilight.soff = selhilight.dlineoff;
                selhilight.eoff = selhilight.mlineoff;
            }
            /* dot behind mark line */
            else if (selhilight.mlineoff < selhilight.dlineoff)
            {
                selhilight.soff = selhilight.mlineoff;
                selhilight.eoff = selhilight.dlineoff;
            }
            /* dot and mark on same offset */        
            else
                selhilight.flags |= SELHIL_SAME;
        }
    }        
        
    lineNo = wp->topLineNo ;
    if ((selhilight.flags & SELHIL_SAME) ||
        (lineNo > selhilight.eline) ||
        (lineNo+wp->numTxtRows-1 < selhilight.sline))
    {
remove_hilight:
        if(wp->w_flag & WFSELDRAWN)
        {
            VIDEO          *vptr;             /* Pointer to the video block */
            unsigned short  sline, eline ;    /* physical screen line to update */
        
            /* Determine the video line position and determine the video block that
             * is being used. */
            vptr = wp->w_vvideo->video;       /* Video block */
            sline = wp->firstRow;             /* Start line */
            eline = sline + wp->numTxtRows ;  /* End line */

            while (sline < eline)
            {
                short vflag;
                
                if ((vflag = vptr[sline].flag) & VFSHMSK)
                    vptr[sline].flag = (vflag & ~VFSHMSK)|VFCHNGD;
                sline++;
            }
            wp->w_flag = (wp->w_flag & ~WFSELDRAWN) | WFMAIN ;
        }
    }
    else
    {
        VIDEO          *vptr;             /* Pointer to the video block */
        unsigned short  sline, eline ;    /* physical screen line to update */
        
        /* Determine the video line position and determine the video block that
         * is being used. */
        vptr = wp->w_vvideo->video;       /* Video block */
        sline = wp->firstRow;             /* Start line */
        eline = sline + wp->numTxtRows ;  /* End line */
        
        /* Mark all of the video lines with the update. */
        while (sline < eline)
        {
            short vflag;
            
            vflag = vptr[sline].flag;
            
            if ((lineNo < selhilight.sline) || (lineNo > selhilight.eline))
            {
                if (vflag & VFSHMSK)
                    vptr[sline].flag = (vflag & ~VFSHMSK)|VFCHNGD;
            }
            else
            {
                if (lineNo == selhilight.sline)
                {
                    if (lineNo == selhilight.eline)
                        vptr[sline].flag = (vflag & ~VFSHMSK)|(VFSHBEG|VFSHEND|VFCHNGD);
                    else
                        vptr[sline].flag = (vflag & ~VFSHMSK)|(VFSHBEG|VFCHNGD);
                }
                else if (lineNo == selhilight.eline)
                    vptr[sline].flag = (vflag & ~VFSHMSK)|(VFSHEND|VFCHNGD);
                else if ((vflag & VFSHALL) == 0)
                    vptr[sline].flag = (vflag & ~VFSHMSK)|(VFSHALL|VFCHNGD);
            }
            /* on to the next one */
            sline++;                        /* Next video line */
            lineNo++;                       /* Next line number */
        }
        wp->w_flag |= (WFSELDRAWN|WFMAIN) ;
    }
}

#if 0
/*
 * menuline()
 * Redisplay the menu line.
 * Boolean flag; TRUE = modified; FALSE = refresh.
 */

void
menuline (int flag)
{
    if (TTsrow > 0)
    {
        /* Mark as changed */
        vvideo.video [0].flag = VFMENUL|VFCHNGD;
        vvideo.video [0].endp = TTncol;
        
        /* Change the line if requested */
        if (flag != FALSE)
            vvideo.video [0].line = menuLine;
    }
}
    
/*  updgar:  if the screen is garbage, clear the physical screen and
             the virtual screen and force a full update		*/

static void
updgar(void)
{
    addModeToWindows(WFUPGAR) ;
    
    menuline(FALSE);                    /* Mark menu line for update */
    
    /* Reset the poke flags */
    poke_xmin = TTncol;
    poke_xmax = 0;
    poke_ymin = TTnrow;
    poke_ymax = 0;
    poke_flag = 0;
}

#endif

int
showCursor(int f, int n)
{
    int ii ;
    
    ii = cursorState ;
    if(f)
        cursorState += n ;
    else
        cursorState = 0 ;
    if((cursorState >= 0) && (ii < 0))
    {
        if(cursorBlink)
            /* kick off the timer */
            TThandleBlink(1) ;
        else
            TTshowCur() ;
    }
    else if((cursorState < 0) && (ii >= 0))
        TThideCur() ;
    return TRUE ;
}

int
showRegion(int f, int n)
{
    int absn ;
    
    absn = (n < 0) ? -n:n ;
    
    if((absn == 1) || (absn == 2))
    {
        if(!(selhilight.flags & SELHIL_FIXED))
            return mlwrite(MWABORT|MWCLEXEC,"[No current region]") ;
        if(absn == 2)
        {
            if(selhilight.bp != curbp)
                return mlwrite(MWABORT|MWCLEXEC,"[Current region not in this buffer]") ;
            curwp->w_flag |= WFMOVEL ;
            if(n == 2)
            {
                n = gotoLine(TRUE,selhilight.dlineno+1) ;
                f = selhilight.dlineoff ;
            }
            else
            {
                n = gotoLine(TRUE,selhilight.mlineno+1) ;
                f = selhilight.mlineoff ;
            }
            if((n == TRUE) && (f <= llength(curwp->w_dotp)))
                curwp->w_doto = (unsigned short) f ;
            return n ;
        }
    }
    switch(n)
    {
    case -3:
        selhilight.bp = curbp;                  /* Select the current buffer */
        selhilight.flags = SELHIL_ACTIVE|SELHIL_CHANGED ;
        selhilight.mlineoff = curwp->w_doto;   /* Current mark offset */
        selhilight.mlineno = curwp->line_no;    /* Current mark line number */
        selhilight.dlineoff = curwp->w_doto;   /* Current mark offset */
        selhilight.dlineno = curwp->line_no;    /* Current mark line number */
        break ;
    case -1:
        selhilight.flags &= ~SELHIL_ACTIVE ;
        break ;
    case 0:        /* Show status of hilighting */
        n = 0 ;
        if(selhilight.flags & SELHIL_ACTIVE)
            n |= 1 ;
        if(selhilight.flags & SELHIL_FIXED)
            n |= 2 ;
        if(selhilight.bp == curbp)
        {
            n |= 4 ;
            /* check for d >= dot > m */
            if(((selhilight.dlineno < curwp->line_no) ||
                ((selhilight.dlineno == curwp->line_no) && (selhilight.dlineoff <= curwp->w_doto))) &&
               ((selhilight.mlineno > curwp->line_no) ||
                ((selhilight.mlineno == curwp->line_no) && (selhilight.mlineoff >  curwp->w_doto))))
                n |= 8 ;
            /* check for m >= dot > d */
            else if(((selhilight.mlineno < curwp->line_no) ||
                     ((selhilight.mlineno == curwp->line_no) && (selhilight.mlineoff <= curwp->w_doto))) &&
                    ((selhilight.dlineno > curwp->line_no) ||
                     ((selhilight.dlineno == curwp->line_no) && (selhilight.dlineoff >  curwp->w_doto))))
                n |= 8 ;
        }
        sprintf(resultStr,"%d",n) ;
        return TRUE ;
                  
    case  1:
        selhilight.flags |= SELHIL_ACTIVE;
        if(f || (selhilight.uFlags & SELHILU_KEEP))
            selhilight.flags |= SELHIL_KEEP ;
        break ;
    case 3:
        if(((selhilight.flags & (SELHIL_FIXED|SELHIL_ACTIVE)) == 0) ||
           (selhilight.bp != curbp))
        {
            selhilight.bp = curbp ;
            selhilight.mlineoff = curwp->w_doto;   /* Current mark offset */
            selhilight.mlineno = curwp->line_no;    /* Current mark line number */
        }
        selhilight.flags |= SELHIL_FIXED|SELHIL_CHANGED|SELHIL_ACTIVE ;
        if(selhilight.uFlags & SELHILU_KEEP)
            selhilight.flags |= SELHIL_KEEP ;
        selhilight.dlineoff = curwp->w_doto;   /* Current mark offset */
        selhilight.dlineno = curwp->line_no;    /* Current mark line number */
        break ;
    default:
        return ABORT ;
    }
    addModeToBufferWindows(selhilight.bp, WFSELHIL);
    return TRUE ;
}

void
windCurLineOffsetEval(WINDOW *wp)
{
    if((wp->curLineOff->l_fp == wp->w_dotp) &&
       !(wp->w_dotp->l_flag & LNCHNG))
        return ;
    if(wp->curLineOff->l_size < wp->w_dotp->l_size)
    {
        meFree(wp->curLineOff) ;
        wp->curLineOff = lalloc(wp->w_dotp->l_size) ;
    }
    /* store dotp as the last line done */
    wp->curLineOff->l_fp = wp->w_dotp ;
    if((wp->w_bufp->hiLight != 0) &&
       (hilights[wp->w_bufp->hiLight]->type & HFRPLCDIFF))
        hilightCurLineOffsetEval(wp) ;
    else
    {
        register unsigned char   cc, *ss, *off ;
        register int pos, ii ;
    
        ss = wp->w_dotp->l_text ;
        off = wp->curLineOff->l_text ;
        pos = 0 ;
#ifndef NDEBUG
        if(wp->w_dotp->l_text[wp->w_dotp->l_used] != '\0')
        {
            _meAssert(__FILE__,__LINE__) ;
            wp->w_dotp->l_text[wp->w_dotp->l_used] = '\0' ;
        }
#endif
        while((cc=*ss++) != 0)
        {
            if(isDisplayable(cc))
                ii = 1 ;
            else if(cc == TAB)
                ii = get_tab_pos(pos) + 1;
            else if (cc < 0x20)
                ii = 2 ;
            else
                ii = 4 ;
            *off++ = (unsigned char) ii ;
            pos += ii ;
        }
        *off = 0 ;
    }
}

static void
updCursor(register WINDOW *wp)
{
    register unsigned int ii, jj ;
    int leftMargin;                     /* Base left margin position (scrolled) */
    
    if(wp->w_flag & (WFREDRAW|WFRESIZE))
        /* reset the curLineOff to force a recalc when needed */
        wp->curLineOff->l_fp = NULL ;
    
    /* set ii to cursors screen col & jj to the require screen scroll */
    if(wp->w_doto == 0)
    {
        /* simple express case when offset is zero */
        ii = 0 ;
        jj = 0 ;
        /* must reset the curLineOff if its changed cos the flag will be lost
         * for next time! Nasty bug!! Must check the line is still dotp if we
         * are to keep it cos it could have been freed.
         */
        if((wp->curLineOff->l_fp != wp->w_dotp) || (wp->w_dotp->l_flag & LNCHNG))
            wp->curLineOff->l_fp = NULL ;
    }
    else
    {
        register unsigned char *off ;
        
        windCurLineOffsetEval(wp) ;
        jj = 0 ;
        ii = wp->w_doto ;
        off = wp->curLineOff->l_text ;
        while(ii--)
            jj += *off++ ;
        
        /* jj - is the character offset of the cursor on the screen */
        if((scrollFlag & 0x0f) == 3)
            leftMargin = wp->w_sscroll ;
        else
            leftMargin = wp->w_scscroll;

        if(jj <= leftMargin)
        {
            /* Cursor behind the current scroll position */
            ii = wp->numTxtCols - (((leftMargin - jj) % wp->w_scrsiz) + wp->w_margin);
            if (ii > jj)                /* Make sure we have enough info to show */
                ii = jj;                /* No - shoe the start of the line */
            jj -= ii;                   /* Correct the screen offset position */
        }
        else if (jj >= leftMargin + (wp->numTxtCols-1))
        {
            /* Cursor ahead of the scroll position and off screen */
            /* Move onto the next screen */
            ii = ((jj - leftMargin - wp->numTxtCols) % wp->w_scrsiz) + wp->w_margin;
            jj -= ii;                   /* Screen offset */
        }
        else
        {
            /* Cursor ahead of scroll position and on screen */
            ii = jj - leftMargin ;
            jj = leftMargin ;
        }
    }
    
    wp->w_sccol = ii;               /* Set the cursor column */
    mwRow = wp->firstRow + (wp->line_no - wp->topLineNo) ;
    mwCol = wp->firstCol + ii ;
    if(wp->w_scscroll != jj)        /* Screen scroll correct ?? */
    {
        wp->w_scscroll = jj;           /* Scrolled line offset */
        wp->w_flag |= WFDOT ;
    }
    switch(scrollFlag & 0x0f)
    {
    case 0:
        if(wp->w_sscroll)
        {
            /* Reset scroll column */
            wp->w_sscroll = 0 ;      /* Set scroll column to base position */
            wp->w_flag |= WFREDRAW; /* Force a screen update */
        }
        break ;
    case 1:
        if((wp->w_sscroll != jj) && wp->w_sscroll)
        {
            /* Reset scroll column */
            wp->w_sscroll = 0 ;      /* Set scroll column to base position */
            wp->w_flag |= WFREDRAW; /* Force a screen update */
        }
        break ;
    case 2:
        if(wp->w_sscroll != jj)
        {
            /* Reset scroll column */
            wp->w_sscroll = jj ;      /* Set scroll column to base position */
            wp->w_flag |= WFREDRAW; /* Force a screen update */
        }
        break ;
        /* case 3 leaves the scroll alone */
    }
}

#define DEBUGGING 0
#if DEBUGGING
static char drawno = 'A' ;
#endif

/*
 * renderLine
 * This function renders a non-hilighted text line. 
 */
int
renderLine (unsigned char *s1, int len, int wid)
{
    register unsigned char  cc;
    register unsigned char  *s2 ;
            
    s2 = disLineBuff + wid;
    while(len-- > 0)
    {
        /* the largest character size is a tab which is user definable */
        if (wid >= disLineSize)
        {
            disLineSize += 512 ;
            disLineBuff = realloc(disLineBuff,disLineSize+32) ;
            s2 = disLineBuff + wid ;
        }
        cc = *s1++ ;
        if(isDisplayable(cc))
        {
            wid++ ;
            *s2++ = cc ;
        }
        else if(cc == TAB)
        {
            int ii=get_tab_pos(wid) + 1;
            
            wid += ii ;
            for( ; ii>0 ; ii--)
                *s2++ = ' ' ;
        }
        else if(cc < 0x20)
        {
            wid += 2 ;
            *s2++ = '^' ;
            *s2++ = cc ^ 0x40 ;
        }
        else
        {
            /* Its a nasty character */
            wid += 4 ;
            *s2++ = '\\' ;
            *s2++ = '0' ;
            *s2++ = hexdigits[cc/0x10] ;
            *s2++ = hexdigits[cc%0x10] ;
        }
    }
    return wid;
}

/* row of screen to update, virtual screen image */
static int
updateline(register int row, register VIDEO *vp1, WINDOW *window)
{
    register unsigned char *s1;         /* Text line pointer */
    register unsigned short flag ;      /* Video line flag */
    HILCOLOR *colp;                     /* Color change list */
    unsigned short noColChng ;          /* Number of colour changes */
    unsigned short colpair;             /* Frame store - Colour pair */
    unsigned short  scol;               /* Lines starting column */
    unsigned short  ncol;               /* Lines number of columns */
    unsigned short *fscp;               /* Frame store - colour pointer */
    unsigned char  *fstp;               /* Frame store - text pointer */
                                           
    /* Get column size/offset */
    s1 = vp1->line->l_text;
    if(window != NULL)
    {
        scol = window->firstCol ;
        ncol = window->numTxtCols ;
    }
    else
    {
        scol = 0 ;
        ncol = TTncol ;
    }
    
    if((flag = vp1->flag) & VFMAINL)
    {
        unsigned short color;
        if(vp1->line->l_flag & LNSMASK)
        {
            color = window->w_bufp->lScheme[vp1->line->l_flag & LNSMASK] ;
            /* We have to assume this line is an exception and the hilno & bracket
             * for the next line should be what this line would have been */
            if(window->w_bufp->hiLight && 
               ((vp1[1].hilno != vp1[0].hilno) || (vp1[1].bracket != vp1[0].bracket)))
            {
                vp1[1].flag |= VFCHNGD ;
                vp1[1].hilno = vp1[0].hilno ;
                vp1[1].bracket = vp1[0].bracket ;
            }
            goto hideLineJump ;
        }
#if HILIGHT
        else if(window->w_bufp->hiLight)
        {
            unsigned char tempIsWordMask;
            
            /* The highlighting relies on the correct setting of the
             * isWordMask to find word boundaries. Context save the existing
             * word mask and restore from the buffer. Do the hilighting
             * and then restore the previous setting. */
            tempIsWordMask = isWordMask;  
            isWordMask = window->w_bufp->isWordMask;
            noColChng = hilightLine(vp1) ;
            isWordMask = tempIsWordMask;
           
            /* Export the information from the higlighting request */
            colp = hilCol ;
        }
#endif
        else
        {
            unsigned short lineLen;
            
            color = window->w_bufp->b_scheme;
hideLineJump:
            colp = hilCol ;
            lineLen = vp1->line->l_used;
            
            if (flag & VFCURRL)
                color += HCTCURFORE;
            
            /* Determine if there is any selection hilighting on the line. */
            if (flag & VFSHMSK)
            {
                /* Check for whole line hilighted */
                if (flag & VFSHALL)
                {
                    if(lineLen > 0)
                        colp[0].ccol = renderLine(s1,lineLen,0) ;
                    else
                        colp[0].ccol = 0 ;
                    colp[0].icol = color + HCTSELFORE;
                    noColChng = 1 ;                        
                }
                else
                {
                    register unsigned short wid, len ;
                    
                    /* Region hilight commences in this line */
                    if((flag & VFSHBEG) && (selhilight.soff > 0))
                    {
                        len = selhilight.soff ;
                        wid = renderLine (s1, len, 0) ;
                        colp[0].icol = color ;
                        colp[0].ccol = wid ;
                        noColChng = 1 ;
                    }
                    else
                    {
                        wid = 0 ;
                        len = 0 ;
                        noColChng = 0 ;
                    }
                    
                    /* Region hilight terminates in this line */
                    if (flag & VFSHEND)
                    {
                        if (selhilight.eoff > len)
                        {
                            /* Set up the colour change */
                            wid = renderLine(s1+len,selhilight.eoff-len,wid) ;
                            colp[noColChng].icol = color + HCTSELFORE;
                            colp[noColChng++].ccol = wid ;
                            len = selhilight.eoff ;
                        }
                        colp[noColChng].icol = color ;
                    }
                    else
                        /* Set up the colour change */
                        colp[noColChng].icol = color + HCTSELFORE;
                    
                    /* Render the rest of the line in the standard colour */
                    if (lineLen > len)
                        wid = renderLine(s1+len, lineLen-len,wid) ;
                    colp[noColChng].ccol = wid ;
                    noColChng += 1 ;
                }
            }
            else
            {
                /* Render the rest of the line in the standard colour */
                if (lineLen > 0)
                    colp[0].ccol = renderLine(s1,lineLen,0) ;
                else
                    colp[0].ccol = 0 ;
                colp[0].icol = color ;
                noColChng = 1 ;
            }
        }

        s1 = disLineBuff ;
        {
            register unsigned short scroll ;

            if (flag & VFCURRL)
                scroll = window->w_scscroll ;   /* Current line scroll */
            else
                scroll = window->w_sscroll ;    /* Hard window scroll */
            if(scroll != 0)
            {
                register unsigned short ii ;

                /* Line is scrolled. The effect we want is is any text at all
                 * is on the line we place a dollar at the start of the line
                 * in the last highlighting colour. If the line is empty then
                 * we do not want the dollar.
                 *
                 * Only process the line if it is not empty, this ensures that
                 * we do not insert a dollar where not required. */
                if (colp->ccol > 0)
                {
                    s1 += scroll ;
                    while((colp->ccol < scroll))
                    {
                        if(noColChng == 1)
                        {
                            colp->ccol = scroll ;
                            break ;
                        }
                        colp++ ;
                        noColChng-- ;
                    }

                    for(ii=0 ; ii<noColChng ; ii++)
                        colp[ii].ccol -= scroll ;

                    /* Always insert the dollar ($) - make space for it */
                    if(colp[noColChng-1].ccol <= 1)
                        colp[noColChng-1].ccol = 1 ;
                    *s1 = '$' ;
                }
            }
        }
        if(colp[noColChng-1].ccol >= ncol)
        {
            s1[ncol-1] = '$' ;
            while((noColChng > 1) && (colp[noColChng-2].ccol >= ncol))
                noColChng-- ;
            colp[noColChng-1].ccol = ncol ;
        }
        else
        {
            /* An extra space is added to the last column so that
             * the cursor will be the right colour */
            s1[colp[noColChng-1].ccol] = ' ' ;
            colp[noColChng-1].ccol++ ;
        }
    }
    else
    {
        colp=hilCol ;
        colp->ccol = vp1->line->l_used ;
        noColChng = 1 ;
        if(flag & VFMODEL)
            colp->icol = mdLnScheme + ((flag & VFCURRL) ? HCTCURFORE : HCTFORE);
        else if (flag & VFMENUL)        /* Menu line */
            colp->icol = osdScheme;
        else
            colp->icol = mlScheme;
    }
    
    /* Get the frame store colour and text pointers */
    fscp = &frameStore[row].colour[scol] ;
    fstp = &frameStore[row].text[scol] ;
    
    /************************************************************************
     * 
     * X-WINDOWS
     * 
     ***********************************************************************/
#ifdef _XTERM
    if(!(meSystemCfg & meSYSTEM_CONSOLE))
    {
        register unsigned char bcol, fcol ;
        register int ii, len, col, cno ;
        XGCValues    xgcv ;

        row = (fdepth*row)+ascent ;
        col = cno = 0 ;                 /* Virtual column start */
        
        if (meSystemCfg & meSYSTEM_FONTFIX)
        {
            unsigned char cc, *sfstp=fstp;
            int spFlag, ccol ;
            do {
                ii = colp->icol;
                fcol = colSchmTbl[ii];
                xgcv.foreground = colTable[fcol] ;
                bcol = colSchmTbl[ii^1] ;
                xgcv.background = colTable[bcol] ;
                XChangeGC(xdisplay,xgc,(GCForeground|GCBackground),&xgcv) ;
                ii = colp->ccol ;
                ii -= col;
                len = ii ;
                colp++ ;
                /* Maintain the frame store and copy the string into 
                 * the frame store with the colour information */
                colpair = FScolourPair (fcol, bcol);
                ccol = col ;
                spFlag = 0 ;
                while (--len >= 0)
                {
                    *fscp++ = colpair;
                    if(((cc=s1[col++]) & 0xe0) == 0)
                    {
                        cc = ' ' ;
                        spFlag++ ;
                    }
                    *fstp++ = cc ;
                }
                XDrawImageString(xdisplay,xwindow,xgc, 
                                 (scol+ccol)*fwidth,row,
                                 (char *) sfstp+ccol, ii);
                while(--spFlag >= 0)
                {
                    while (((cc=s1[ccol]) & 0xe0) != 0)
                        ccol++ ;
                    sfstp[ccol] = cc ;
                    XTERMSpecialChar((scol+ccol)*fwidth,row-ascent,cc) ;
                    ccol++ ;
                }
            } while(++cno < noColChng) ;
        }
        else
        {
            do {
                ii = colp->icol;
                fcol = colSchmTbl[ii];
                xgcv.foreground = colTable[fcol] ;
                bcol = colSchmTbl[ii^1] ;
                xgcv.background = colTable[bcol] ;
                XChangeGC(xdisplay,xgc,(GCForeground|GCBackground),&xgcv) ;
                ii = colp->ccol ;
                len = ii-col;
                XDrawImageString(xdisplay,xwindow,xgc, 
                                 (scol+col)*fwidth,row,
                                 (char *) s1+col, len);
                colp++ ;
                
                /* Maintain the frame store and copy the string into 
                 * the frame store with the colour information */
                colpair = FScolourPair (fcol, bcol);
                while (--len >= 0)
                {
                    *fscp++ = colpair;
                    *fstp++ = s1[col++];
                }
            } while(++cno < noColChng) ;
        }
        if(vp1->bcol != bcol)
        {
            ii = ncol ;
            vp1->bcol = bcol ;
        }
        else
            ii = vp1->endp ;
        vp1->endp = col ;
        
        if(ii > col)
        {
            char buff[MAXBUF] ;
            ii -= col ;
            cno = ii ;
            s1 = (unsigned char *) buff ;
            do
            {
                *s1++ = ' ' ;           /* End fill with spaces */
                *fstp++ = ' ';          /* Frame store space fill */
                *fscp++ = colpair;      /* Frame store colour fill */
            }
            while(--cno) ;
            XDrawImageString(xdisplay,xwindow,xgc, 
                             (scol+col)*fwidth, row, buff, ii) ;
        }
#if DEBUGGING
        else
            ii = 0 ;
        XDrawImageString(xdisplay,xwindow,xgc, 
                         (scol+col+ii-1)*fwidth, row, &drawno,1) ;
#endif
    }
    else
#endif
    /************************************************************************
     * 
     * TERMCAP
     * 
     ***********************************************************************/
#ifdef _TCAP
    {
        unsigned char fcol, bcol;
        register int ii, col, cno ;

/*        TCAPputc(0) ;*/
        TCAPgoto(row,scol);	/* Go to start of line. */
/*        TCAPputc(0) ;*/

        col = 0 ;
        for(cno=0 ; cno<noColChng ; cno++)
        {
#if COLOR
#ifdef _TCAPFONT
            if(meSystemCfg & meSYSTEM_ANSICOLOR)
            {
#endif /* _TCAPFONT */
                fcol = colSchmTbl[colp->icol];
                bcol = colSchmTbl[colp->icol^1];
                TCAPcolorSetScheme(fcol,bcol) ;
#ifdef _TCAPFONT
            }
            else if(meSystemCfg & meSYSTEM_FONTS)
            {
#endif /* _TCAPFONT */
#endif /* COLOR  */
#ifdef _TCAPFONT
                fcol = fntSchmTbl[colp->icol];
                bcol = TCAPfontSetScheme(fcol) ;
#if COLOR
            }
            colpair = FScolourPair (fcol,bcol);
#endif /* COLOR  */
#endif /* _TCAPFONT */
            /* Output the character in the specified colour.
             * Maintain the frame store */
            for(ii = colp->ccol ; col<ii ; col++)
            {
                *fscp++ = colpair;
                *fstp++ = *s1;
                TCAPputc(*s1++) ;
            }
            colp++;
        }
        if(vp1->bcol != bcol)
        {
            ii = ncol ;
            vp1->bcol = bcol ;
        }
        else
            ii = vp1->endp ;
        vp1->endp = col ;
        /* Output the end of the line - maintain the frame store */
        for( ; col<ii ; col++)
        {
            *fscp++ = colpair;
            *fstp++ = ' ';
            TCAPputc(' ') ;
        }
#if DEBUGGING
        TCAPputc(drawno) ;
#endif
#if COLOR
#ifdef _TCAPFONT
        if(meSystemCfg & meSYSTEM_ANSICOLOR)
#endif
            TCAPcolorReset() ;
#endif
#ifdef _TCAPFONT
        if(vstate)
            TCAPfontReset() ;
#endif
    }

#endif /* _TCAP */
    /************************************************************************
     * 
     * MS-DOS
     * 
     ***********************************************************************/
#ifdef _DOS
    {
        register unsigned char  cc, bcol ;
        register unsigned short ii, len ;
        
        ii = 0 ;
        do {
            register unsigned short fcol ;
            fcol = colp->icol ;
            bcol = colSchmTbl[fcol^1] ;
            fcol = colSchmTbl[fcol];
            cc = TTCmake(colTable[fcol],colTable[bcol]) ;
            colpair = FScolourPair (fcol, bcol);
            
            /* Update the screen.
             * Maintain the frame store and copy the sting into the 
             * frame store with the colour information */
            len = colp->ccol;
            for(; ii < len; ii++)
            {
                *fscp++ = colpair;
                *fstp++ = *s1;
                ScreenPutChar(*s1++, cc, scol++, row);
            }
            colp++;
        } while(--noColChng) ;
        
        if(vp1->bcol != bcol)
        {
            len = ncol ;
            vp1->bcol = bcol ;
        }
        else
            len = vp1->endp ;
        vp1->endp = ii ;
        /* Update the screen for end of the line. Maintain the 
         * frame store copy. */
        for( ; ii<len ; ii++)
        {
            *fscp++ = colpair;
            *fstp++ = ' ';
            ScreenPutChar(' ', cc, scol++, row);
        }
#if DEBUGGING
        ScreenPutChar(drawno,TTCmake(1,7),scol-1,row) ;
#endif
    }
#endif /* _DOS */
    /*************************************************************************
     * 
     * MS-WINDOWS
     * 
     ************************************************************************/
#ifdef _WIN32
#ifdef _WINCON
    if (meSystemCfg & meSYSTEM_CONSOLE)
    {
        register unsigned char  cc, bcol, fcol ;
        register unsigned short ii, ccol ;
        register int ll ;
        
        ccol = 0 ;
        do {
            ll = colp->icol ;
            fcol = colSchmTbl[ll];
            bcol = colSchmTbl[ll^1] ;
            cc = TTCmake(colTable[fcol],colTable[bcol]) ;
            colpair = FScolourPair (fcol, bcol);
            
            /* Update the screen.
             * Maintain the frame store and copy the sting into the 
             * frame store with the colour information */
            ii = colp->ccol;
            ll = ii - ccol ;
            ConsoleDrawString (s1, cc, scol+ccol, row, ll);
            ccol = ii ;
            while(--ll >= 0)
            {
                *fscp++ = colpair;
                *fstp++ = *s1++;
            }
            colp++;
        } while(--noColChng) ;
        
        if(vp1->bcol != bcol)
        {
            ii = ncol ;
            vp1->bcol = bcol ;
        }
        else
            ii = vp1->endp ;
        vp1->endp = ccol ;
        /* Update the screen for end of the line. Maintain the 
         * frame store copy. */
        if(ccol < ii)
        {
            s1 = fstp ;
            ll = ii - ccol ;
            while(--ll >= 0)
            {
                *fscp++ = colpair;
                *fstp++ = ' ';
            }
            ll = ii - ccol ;
            ConsoleDrawString (s1, cc, scol+ccol, row, ll);
#if DEBUGGING
            ccol = ii ;
#endif
        }
#if DEBUGGING
        ConsoleDrawString (&drawno, cc, scol+ccol, row, 1);
#endif
/*        TTflush() ;*/
    }
    else
#endif
    {
        int offset;                     /* Offset into the line */
        int len;                        /* Local line column */
        
        /* Iterate through the colour changes */
        len = 0;
        do {
            /* Set up the colour maps */
            offset = colp->ccol - len;   /* Get width of coloured block */
            colpair = FScolourPair (colSchmTbl[colp->icol],
                                    colSchmTbl[colp->icol^1]);
            colp++;                     /* Next colour pair */
            len += offset;              /* Increase line length */
            while (--offset >= 0)
            {
                *fscp++ = colpair;
                *fstp++ = *s1++;
            }
        } while (--noColChng > 0);
        
        /* Clear to the end of the line */
        if (vp1->bcol != FScolourBack (colpair))
        {
            offset = ncol - len;
            vp1->bcol = FScolourBack (colpair);
        }
        else if (vp1->endp < len)
            offset = 0;
        else
            offset = vp1->endp - len;
        vp1->endp = len;
        len += offset;
        
        /* Fill in the frame store */
        while (--offset >= 0)
        {
            *fscp++ = colpair;
            *fstp++ = ' ';
        }
#if DEBUGGING
        fstp[-1] = drawno ;
#endif
        /* Invalidate the screen */
        if(flag & VFMAINL)
            /* in the main window just add the area */
            TTaddArea(row,scol,1,len) ;
        else
            /* other areas we want to set and apply */
            TTputs(row,scol,len) ;
    }
#endif

    return TRUE ;
}

/*
 * updateWindow() - update all the lines in a window on the virtual screen
 * The screen now consists of TTnrow array of VIDEO's which point to the
 * actual LINE's (dummy ones if its a modeline) the flags indicate if the
 * LINE is mess or mode or main line and if its current
 */
static void
updateWindow(WINDOW *wp)
{
    BUFFER *bp = wp->w_bufp ;
    VIDEO   *vptr;                    /* Pointer to the video block */
    register LINE *lp ;               /* Line to update */
    register int   row, nrows ;       /* physical screen line to update */
    register unsigned char force ;            
                        
    force = (unsigned char) (wp->w_flag & (WFREDRAW|WFRESIZE)) ;
    /* Determine the video line position and determine the video block that
     * is being used. */
    row   = wp->firstRow ;
    vptr  = wp->w_vvideo->video + row ;  /* Video block */
    nrows = wp->numTxtRows ;
    
    /* Search down the lines, updating them */
    {
        int ii ;
        ii = (wp->line_no - wp->topLineNo) ;
        lp = wp->w_dotp ;
        while(--ii >= 0)
            lp = lback(lp) ;
    }
#if HILIGHT
    if(bp->hiLight)
    {
        if(force)
        {
            vptr->hilno   = bp->hiLight ;
            vptr->bracket = NULL ;
            wp->w_flag |= WFLOOKBK ; 
        }
        if((wp->w_flag & WFLOOKBK) && !TTahead())
        {
            if(hilights[bp->hiLight]->ignore)
                hilLookBack(wp) ;
            wp->w_flag &= ~WFLOOKBK ; 
        }
    }
#endif
#ifdef _WIN32
#ifdef _WINCON
    if (!(meSystemCfg & meSYSTEM_CONSOLE))
#endif
        TTinitArea() ;
#endif
    while(--nrows >= 0)
    {
        register unsigned char update=force|(vptr->flag & VFCHNGD) ;
        
        /*---	 and update the virtual line */
        if(lp == wp->w_dotp)
        {
            if(((vptr->flag & VFTPMSK) != (VFMAINL|VFCURRL)) || 
               (wp->w_flag & WFDOT))
                update = 1 ;
            vptr->flag = (vptr->flag & ~VFTPMSK) | VFMAINL | VFCURRL ;
        }
        else
        {
            if((vptr->flag & VFTPMSK) != VFMAINL)
                update = 1 ;
            vptr->flag = (vptr->flag & ~VFTPMSK) | VFMAINL ;
        }
        if(force)
        {
            if(force & WFRESIZE)
                vptr->endp = wp->numTxtCols ;
            vptr->wind = wp ;
            vptr->line = lp ;
            vptr[1].hilno = bp->hiLight ;
            vptr[1].bracket = NULL ;
        }
        else if((lp->l_flag & LNCHNG) ||
                (vptr->line != lp))
        {
            vptr->line = lp ;
            update = 1 ;
        }
        if(update)
            updateline(row,vptr,wp);
        row++ ;
        vptr++ ;
        if(lp == bp->b_linep)
            break ;
        lp=lforw(lp) ;
    }
    
    if(nrows > 0)
    {
        /* if we are at the end */
        /* store the b_linep l_flag and set to 0 as it may have a $line-scheme */
        unsigned char l_flag=lp->l_flag ;
        unsigned char bcol = vptr[-1].bcol ;
        lp->l_flag = 0 ;
        for( ; (--nrows >= 0) ; row++,vptr++)
        {
            vptr->line = lp ;
            if(force || (vptr->endp > 1) || (vptr->flag & (VFCHNGD|VFCURRL)) ||
               (vptr->bcol != bcol))
            {
                vptr->flag = VFMAINL ;
                vptr->wind = wp ;
                if(force & WFRESIZE)
                    vptr->endp = wp->numTxtCols ;
                updateline(row,vptr,wp);
            }
        }
        lp->l_flag = l_flag ;
    }
#ifdef _WIN32
#ifdef _WINCON
    if (!(meSystemCfg & meSYSTEM_CONSOLE))
#endif
        TTapplyArea() ;
#endif
}

unsigned char
assessModeLine(char *ml)
{
    unsigned char flags=WFMODE|WFRESIZE ;
    char cc ;
    
    while((cc = *ml++) != '\0')
    {
        if(cc == '%')
        {
            switch((cc = *ml++))
            {
            case 'l':
                flags |= WFMOVEL ;
                break ;
            case 'c':
                /* The column often changes with the line,
                 * so add that too
                 */
                flags |= WFMOVEC|WFMOVEL ;
                break ;
            }
        }
    }
    return flags ;
}

/*
 * updateModeLine()
 * Redisplay the mode line for the window pointed to by the "wp". This is the
 * only routine that has any idea of how the modeline is formatted. You can
 * change the modeline format by hacking at this routine. Called by "update"
 * any time there is a dirty window.
 */
static void
updateModeLine(WINDOW *wp)
{
    time_t         clock;		    /* Time in machine format. */
    struct tm	  *time_ptr;	            /* Pointer to time frame. */
    register char *ml, *cp, *ss ;
    register char  cc, lchar ;
    register BUFFER *bp ;
    register int   ii ;		            /* loop index */
    register int   lineLen ;                /* Max length of the line */
    
    /* See if there's anything to do first */
    bp = wp->w_bufp;
    if((ml = bp->modeLineStr) == NULL)
    {
        if((wp->w_flag & modeLineFlags) == 0)
            return ;
        ml = modeLineStr ;
    }
    else if((wp->w_flag & bp->modeLineFlags) == 0)
        return ;
                
    lineLen = wp->numTxtCols ;              /* Max length of line. Only need to
                                             * evaluate this many characters */
    wp->model->l_used = lineLen ;

    if (wp == curwp)                        /* mark the current buffer */
        lchar = windowChars [WCMLCWSEP];    /* Typically '=' */
    else
        lchar = windowChars [WCMLIWSEP];    /* Typically '-' */

    clock = time(0);	                    /* Get system time */
    time_ptr = (struct tm *) localtime (&clock);	/* Get time frame */
    cp = (char *) wp->model->l_text ;
    while((lineLen > 0) && ((cc = *ml++) != '\0'))
    {
        if(cc == '%')
        {
            switch((cc = *ml++))
            {
            case 's':
                /* Horizontal split line character. Only displayed if split
                 * is enabled in the scroll bars */
                if (gsbarmode & WMSPLIT)
                {
                    *cp++ = windowChars [WCHSBSPLIT];
                    lineLen-- ;
                }
                break;
            case 'r':
                /* root user - unix only */
#ifdef _UNIX
                /*
                 * The idea of this nastiness is that the mode line is
                 * changed to give some visual indication when running as root.
                 */
                if(meUid == 0)
                    *cp++ = '#' ;
                else
#endif
                    *cp++ = lchar ;
                lineLen-- ;
                break ;

            case 'u':
                /* buffer changed */
                if(meModeTest(bp->b_mode,MDEDIT))        /* "*" if changed. */
                    *cp++ = windowChars [WCMLBCHNG];     /* Typically '*' */
                else if (meModeTest(bp->b_mode,MDVIEW))  /* "%" if view. */
                    *cp++ = windowChars [WCMLBVIEW];     /* Typically '%' */
                else
                    *cp++ = lchar ;
                lineLen--;
                break ;

            case 'e':
                /* add in the mode flags */
                for (ii = 0; ii < MDNUMMODES; ii++)
                    if(meModeTest(modeLineDraw,ii) &&
                       meModeTest(bp->b_mode,ii))
                    {
                        *cp++ = modeCode[ii] ;
                        if (--lineLen == 0)
                            break;
                    }
                break ;

            case 'k':
                if (kbdmode == RECORD)           /* if playing macro */
                {
                    ss = "REC" ;
                    goto model_copys ;
                }
                break ;

            case 'l':
                ss = meItoa(wp->line_no+1);
                goto model_copys ;

            case 'n':
                ss = meItoa(wp->w_bufp->elineno+1);
                goto model_copys ;
                
            case 'c':
                ss = meItoa(wp->w_doto) ;
                goto model_copys ;

            case 'b':
                ss = bp->b_bname ;
                goto model_copys ;

            case 'f':
                if((bp->b_bname[0] == '*') || (bp->b_fname == NULL))
                    break ;
                ss = bp->b_fname ;
                if((ii=strlen(ss)) > lineLen)
                {
                    *cp++ = '$' ;
                    if(--lineLen == 0)
                        break ;
                    ss += ii-lineLen ;
                }
                goto model_copys ;

            case 'D':
                ss = meItoa(time_ptr->tm_mday);
                goto model_copys ;

            case 'M':
                ss = meItoa(time_ptr->tm_mon+1);
                goto model_copys ;

            case 'Y':
                ss = meItoa(time_ptr->tm_year);
                goto model_copys ;

            case 'h':
                ss = meItoa(time_ptr->tm_hour);
                goto model_2n_copy ;

            case 'm':
                ss = meItoa(time_ptr->tm_min);
model_2n_copy:
                if(ss[1] == '\0')
                {
                    *cp++ = '0' ;
                    if (--lineLen == 0)
                        break;
                }
model_copys:
                while((cc = *ss++) != '\0')
                {
                    *cp++ = cc ;
                    if (--lineLen == 0)
                        break;
                }
                break ;
            default:
                *cp++ = cc ;
                lineLen--;
            }
        }
        else
        {
            if(cc == (char) windowChars [WCMLIWSEP])  /* Is it a '-' ?? */
                *cp++ = lchar ;
            else
                *cp++ = cc ;
            lineLen--;
        }
    }
    while(lineLen-- > 0)              /* Pad to full width. */
        *cp++ = lchar ;
    
    {
        register VIDEO *vptr ;        /* Pointer to the video block */

        /* Determine the video line position and determine the video block that
         * is being used. */
        ii = wp->firstRow + wp->numTxtRows ;
        vptr = wp->w_vvideo->video + ii ;        /* Video block */

        if(wp->w_flag & WFRESIZE)
            vptr->endp = wp->numTxtCols ;
        if(curwp == wp)
            vptr->flag = VFMODEL|VFCURRL ;
        else
            vptr->flag = VFMODEL ;
        vptr->line = wp->model ;
        updateline(ii,vptr,wp) ;
    }
}

/*	reframe: check to see if the cursor is on in the window
        and re-frame it if needed or wanted		*/

static void
reframe(WINDOW *wp)
{
    register long  ii ;

    /* See if the selection hilighting is enabled for the buffer */
    if ((selhilight.flags & SELHIL_ACTIVE) &&
        (selhilight.bp == wp->w_bufp))
        wp->w_flag |= WFSELHIL;
    
#ifdef _IPIPES
    if(meModeTest(wp->w_bufp->b_mode,MDLOCK) &&
       (wp->w_dotp->l_flag & LNMARK))
    {
        meAMARK *ap=wp->w_bufp->b_amark ;
        
        /* Are we at the input line? */
        while((ap != NULL) && (ap->name != 'I'))
            ap = ap->next ;

        if((ap != NULL) &&
           (ap->line == wp->w_dotp) &&
           (ap->offs == wp->w_doto))
        {
            IPIPEBUF *ipipe ;
            
            /* Yes we are at the right place, find the ipipe node */
            ipipe = ipipes ;
            while(ipipe->bp != wp->w_bufp)
                ipipe = ipipe->next ;
            
            if((ii = ipipe->curRow) >= wp->numTxtRows)
                ii = wp->numTxtRows-1 ;
            if((ii = wp->line_no-ii) < 0)
                ii = 0 ;
            if((wp->w_flag & WFFORCE) || (ii != wp->topLineNo))
            {
                wp->topLineNo = ii ;
                /* Force the scroll box to be updated if present. */
                wp->w_flag |= WFSBOX ;
            }
            return ;
        }
    }
#endif
    /* if not a requested reframe, check for a needed one */
    if(!(wp->w_flag & WFFORCE))
    {
        ii = wp->line_no - wp->topLineNo ;
        if((ii >= 0) && (ii < wp->numTxtRows))
            return ;
    }
    /* reaching here, we need a window refresh */
    ii = wp->w_force ;
    
    /* how far back to reframe? */
    if(ii > 0)
    {	/* only one screen worth of lines max */
        if (--ii >= wp->numTxtRows)
            ii = wp->numTxtRows - 1;
    }
    else if(ii < 0)
    {	/* negative update???? */
        ii += wp->numTxtRows;
        if(ii < 0)
            ii = 0;
    }
    else
        ii = wp->numTxtRows / 2;
    
    ii = wp->line_no-ii ;
    if(ii != wp->topLineNo)
    {
        if(ii <= 0)
            wp->topLineNo = 0 ;
        else
            wp->topLineNo = ii ;
        /* Force the scroll box and lookBack to be updated if present. */
        wp->w_flag |= WFSBOX|WFLOOKBK ;
    }
}


/*
 * updateSrollBar
 * Update the scroll bar
 */
static void
updateScrollBar (WINDOW *wp)
{
    int ii;                             /* index into split string */
    int col;                            /* The column */
    int row;                            /* Current row */
    int endrow;                         /* Terminating row */
    unsigned short color;               /* Region colour */
    unsigned char *wbase;               /* Base window character */
    int  len;                           /* Length of bar */
    int  flipBox;                       /* Flip colours for the box */
    
    /* Has this window got a bar present ?? */
    if (!(wp->w_mode & WMVBAR))
        return;                         /* No quit */
    
    /* Sort out the colours we should be using */
    color = sbarScheme + ((curwp == wp) ? HCTCURFORE : HCTFORE);
    
    /* Get the extents of the bar line */
    row = wp->firstRow;                 /* Start of row */
    col = wp->firstCol + wp->numTxtCols;  /* Start column */
    
    if (wp->w_mode & WMVWIDE)
    {
        len = 2;
        wbase = &windowChars[WCVSBSPLIT1];
    }
    else
    {
        len = 1;
        wbase = &windowChars[WCVSBSPLIT];
    }
    
    /* Flipbox is provided where we have not got a good reverse video character
     * for the scroll box (UNIX !!). When WMRVBOX is defined we may use a <SPACE>
     * character for the scroll box and paint in reverse video, this then makes a
     * better scroll box as required. 
     * 
     * Simply get the status and record the offset of the box, we will flip fcol
     * and bcol when we get there. If it is not defined then set to the end of
     * scroll bar + 1. 
     */
    if (wp->w_mode & WMRVBOX)
        flipBox = 1 << (WCVSBBOX-WCVSBSPLIT) ;
    else
        flipBox = 0 ;
    
#ifdef _WIN32
#ifdef _WINCON
    if (!(meSystemCfg & meSYSTEM_CONSOLE))
#endif
        /* Initialize the invalid screen area */
        TTinitArea() ;
#endif /* _WIN32 */
    
    /* Iterate down the length of the split line */
    for (ii = 0; ii <= (WCVSBML-WCVSBSPLIT); ii++, wbase += len)
    {
        color = (color & ~1) ^ (flipBox & 1) ;  /* Select component colour */
        endrow = wp->w_sbpos [ii] ;
        flipBox >>= 1;
        
        /* See if there is anything to do. */
        if ((wp->w_flag & (WFSBSPLIT << ii)) == 0)
        {
            row = endrow ;
            continue ;
        }
        
        /* Perform the update */
#ifdef _WIN32
#ifdef _WINCON
        if (!(meSystemCfg & meSYSTEM_CONSOLE))
#endif
            /* Invalidate the screen */
            TTaddArea (row, col, endrow-row, len);
#endif /* _WIN32 */
        
        for (; row < endrow; row++)
        {
            /* Update the frame store with the character and colour
             * information - this is nearly the same for all platforms
             * If using _TCAPFONT then the colpair is the scheme
             */
            unsigned char  *fstp ;      /* Frame store text pointer */
            unsigned short *fscp ;      /* Frame store colour pointer */
            unsigned short colpair;     /* Colour pair */
            
            fstp = &frameStore [row].text [col];
            fscp = &frameStore [row].colour [col];
            
#ifdef _TCAPFONT
            if(meSystemCfg & meSYSTEM_FONTS)
                colpair = color ;
            else
#endif
                colpair = FScolourPair (colSchmTbl[color], colSchmTbl[color^1]);
                
#ifdef _WIN32
            /***************************************************************
             * 
             * Windows
             * 
             **************************************************************/
            
#ifdef _WINCON
            if (meSystemCfg & meSYSTEM_CONSOLE)
            {
                int cc ;
                cc = TTCmake(colTable[colSchmTbl[color]],colTable[colSchmTbl[color^1]]) ;
                ConsoleDrawString(wbase,cc,col,row,len) ;
/*                TTflush() ;*/
                *fscp = colpair;  /* Assign the colour */
                *fstp = *wbase;   /* Assign the text */
                if(len > 1)
                {
                    fscp[1] = colpair;    /* Assign the colour */
                    fstp[1] = wbase[1];   /* Assign the text */
                }
            }
            else
#endif
            {
                if(len == 1)
                {
                    *fscp = colpair;  /* Assign the colour */
                    *fstp = *wbase;   /* Assign the text */
                }
                else
                {
                    *fscp++ = colpair;  /* Assign the colour */
                    *fscp = colpair;    /* Assign the colour */
                    *fstp++ = *wbase;   /* Assign the text */
                    *fstp = wbase[1];   /* Assign the text */
                }
            }
            
#endif /* _WIN32 */
            
#ifdef _XTERM
            /***************************************************************
             * 
             * X-WINDOWS
             * 
             **************************************************************/
            if(!(meSystemCfg & meSYSTEM_CONSOLE))
            {
                XGCValues xgcv ;
                
                xgcv.foreground = colTable[colSchmTbl[color]] ;
                xgcv.background = colTable[colSchmTbl[color^1]] ;
                XChangeGC(xdisplay,xgc,(GCForeground|GCBackground),&xgcv) ;
                
                if (meSystemCfg & meSYSTEM_FONTFIX)
                {
                    unsigned char cc;
                    *fscp = colpair;  /* Assign the colour */
                    if(!((cc = *wbase) & 0xe0))
                        cc = ' ' ;   /* Assign the text */
                    *fstp = cc ;   /* Assign the text */
                    
                    if(len > 1)
                    {
                        fscp[1] = colpair;  /* Assign the colour */
                        if(!((cc = wbase[1]) & 0xe0))
                            cc = ' ' ;   /* Assign the text */
                        fstp[1] = cc ;   /* Assign the text */
                    }
                    XDrawImageString (xdisplay,xwindow,xgc,(fwidth * col),
                                      (fdepth * row) + ascent,(char *) fstp, len);
                    if(!((cc = *wbase) & 0xe0))
                    {
                        XTERMSpecialChar(fwidth*col,fdepth*row,cc) ;
                        *fstp = cc ;   /* Assign the text */
                    }
                    if((len > 1) && !((cc = *wbase) & 0xe0))
                    {
                        XTERMSpecialChar(fwidth*(col+1),fdepth*row,cc) ;
                        fstp[1] = cc ;   /* Assign the text */
                    }
                }
                else
                {
                    *fscp = colpair;  /* Assign the colour */
                    *fstp = *wbase ;   /* Assign the text */
                    if(len > 1)
                    {
                        fscp[1] = colpair;  /* Assign the colour */
                        fstp[1] = wbase[1] ;
                    }
                    XDrawImageString (xdisplay,xwindow,xgc,(fwidth * col),
                                      (fdepth * row) + ascent,(char *) wbase, len);
                }
            }
            else
#endif
#ifdef _TCAP
            /***************************************************************
             * 
             * TERMCAP
             * 
             **************************************************************/
            {
                unsigned char cc;
                
                TCAPgoto(row, col);   /* Go to correct place. */
#if COLOR
#ifdef _TCAPFONT
                if(meSystemCfg & meSYSTEM_ANSICOLOR)
                {
#endif /* _TCAPFONT */
                    TCAPcolorSetScheme(colSchmTbl[color],colSchmTbl[color^1]) ;
#ifdef _TCAPFONT
                }
                else if(meSystemCfg & meSYSTEM_FONTS)
                {
#endif /* _TCAPFONT */
#endif /* COLOR  */
#ifdef _TCAPFONT
                    TCAPfontSetScheme(fntSchmTbl[color]) ;
#if COLOR
                }
#endif /* COLOR  */
#endif /* _TCAPFONT */
                *fscp = colpair;  /* Assign the colour */
                cc = *wbase;
                *fstp = cc ;   /* Assign the text */
                TCAPputc(cc) ;
                if(len > 1)
                {
                    fscp[1] = colpair;  /* Assign the colour */
                    cc = wbase[1] ;
                    fstp[1] = cc ;   /* Assign the text */
                    TCAPputc(cc) ;
                }
#if COLOR
#ifdef _TCAPFONT
                if(meSystemCfg & meSYSTEM_ANSICOLOR)
#endif
                    TCAPcolorReset() ;
#endif
#ifdef _TCAPFONT
                if (vstate)
                    TCAPfontReset() ;
#endif
            }
#endif /* _TCAP */
#ifdef _DOS
            /***************************************************************
             * 
             * MS-DOS
             * 
             **************************************************************/
            {
                unsigned char ss ;
                int cc ;
                
                cc = TTCmake(colTable[colSchmTbl[color]],colTable[colSchmTbl[color^1]]) ;
                *fscp = colpair;  /* Assign the colour */
                ss = *wbase;
                *fstp = ss ;   /* Assign the text */
                ScreenPutChar(ss,cc,col,row);
                if(len > 1)
                {
                    fscp[1] = colpair;  /* Assign the colour */
                    ss = wbase[1] ;
                    fstp[1] = ss ;   /* Assign the text */
                    ScreenPutChar(ss,cc,col+1,row);
                }
            }
#endif /* _DOS */
        }
    }   
#ifdef _WIN32
#ifdef _WINCON
    if (!(meSystemCfg & meSYSTEM_CONSOLE))
#endif
        /* Apply the invalid area */
        TTapplyArea () ;
#endif /* _WIN32 */
}


/*
 * pokeUpdate()
 * Update all of the areas of the screen that have been modified.
 * This is a simple bounding box test against the extents of the 
 * window. If any part of the poke screen box lies within the window
 * then the lines in the virtual video structure must be marked as 
 * changed in order to refresh them.
 */
static void 
pokeUpdate (void)
{
    WINDOW *wp;                         /* Window pointer */
    VIDEO  *vp1;                        /* Video pointer */
    
    /* Fix up the message line - special case */
    if (poke_ymax == TTnrow)
    {
        vp1 = vvideo.video + TTnrow;
        vp1->flag |= VFCHNGD;
        vp1->endp = TTncol;             /* Force update to EOL        */
    }
    
    /* Fix up the menu line - spacial case */
    if (poke_ymin < TTsrow)
    {
        vvideo.video->flag |= VFCHNGD;
        vvideo.video->endp = TTncol;
    }
    
    /* Fix up the windows */
    for (wp = wheadp; wp != NULL; wp = wp->w_wndp)
    {
        int ii, jj;                      /* Offset/Length counter */
        
        /* Bounding box test agains the window extents */
        if (wp->firstCol > poke_xmax)
            continue;
        if (wp->firstRow > poke_ymax)
            continue;
        if ((wp->firstCol + wp->numCols) <= poke_xmin)
            continue;
        if ((wp->firstRow + wp->numRows) <= poke_ymin)
            continue;
        
        /* Got here - therefore overlaps the window.
         * Determine vertical starting position on the frame. 
         * 
         * ii initially contains the startOffset */
        if (wp->firstRow > poke_ymin)
            ii = wp->firstRow;            /* Start at top of window */
        else
            ii = poke_ymin;             /* Start at top of poke */
        
        /* Get starting point og the Video frame pointer */
        vp1 = (VIDEO *)(wp->w_vvideo->video) + ii;
        
        /* Determine the end position on the frame and work out
         * the number of lines that need to be updated.
         * 
         * ii will contain the (endOffset - startOffset) i.e the 
         * number of video rows to be updated */
        if ((jj=wp->firstRow+wp->numRows-1) <= poke_ymax)
        {
            ii = jj - ii - 1 ;
            wp->w_flag |= WFMODE ;
        }
        else
            ii = poke_ymax - ii;        /* End at poke max */
        
        /* Iterate down the video frame until we have exhausted our
         * row count. Note that we only mark the underlying video structure
         * as changed and NOT the window contents. This ensures that the
         * video is re-drawn without checking the window flags
         */
        do
        {
            vp1->flag |= VFCHNGD;       /* Mark video line as changed */
            vp1->endp = wp->numTxtCols; /* Force update to EOL        */
            vp1++;                      /* Next line                  */
        }
        while (ii-- > 0);               /* Until exhaused line count  */
        
        /* flag the window as needing attension */
        wp->w_flag |= WFMAIN ;
        
        /* Fix up the vertical scroll bar if we have invaded it's space,
         * note that we do not modify the underlying separator bar in the
         * VVIDOE structure since it has not been re-sized. It is only 
         * necessary to mark the scroll bar itself as changed */
        if ((poke_xmax >= (wp->firstCol + wp->numTxtCols)) &&
            (wp->w_mode & WMVBAR))
            wp->w_flag |= WFSBAR;
    }
    
    /* Reset the poke flags for next time. Set to their maximal limits
     * that will force a fail on the next poke screen operation */
    poke_xmin = TTncol;
    poke_xmax = 0;
    poke_ymin = TTnrow;
    poke_ymax = 0;
    poke_flag = 0;
}
        
/*
 * Make sure that the display is right. This is a three part process. First,
 * scan through all of the windows looking for dirty ones. Check the framing,
 * and refresh the screen. Second, make sure that "currow" and "curcol" are
 * correct for the current window. Third, make the virtual and physical
 * screens the same.
 */
/* screenUpdate:	user routine to force a screen update
		always finishes complete update		*/

int
screenUpdate(int f, int n)
{
    register WINDOW *wp;

#if DEBUGGING
    if(drawno++ == 'Z')
        drawno = 'A' ;
#endif
    if(n)
    {
        /* if screen is garbage, re-plot it all */
        if (TTsrow > 0)
        {
            /* Mark menu as changed */
            vvideo.video[0].flag = VFMENUL|VFCHNGD;
            vvideo.video[0].endp = TTncol;
        }
        /* Mark message line as changed */
        vvideo.video[TTnrow].flag |= VFCHNGD ;
        vvideo.video[TTnrow].endp = TTncol;
        /* Reset the poke flags */
        poke_xmin = TTncol;
        poke_xmax = 0;
        poke_ymin = TTnrow;
        poke_ymax = 0;
        poke_flag = 0;
    }
    else
    {
        if(osdDisplayHd != NULL)
            /* else we must store any osd dialogs */
            osdStoreAll() ;

        /* if the screen has been poked then fix it */
        if (poke_flag)                      /* Screen been poked ??           */
            pokeUpdate();                   /* Yes - determine screen changes */
    }
    
    /* Does the title bar need updating? */
#ifdef _WINDOW
    {
        static char *lastBName=NULL ;
        
        if(n || (curbp->b_bname != lastBName))
        {
            lastBName = curbp->b_bname ;
            TTtitleText(curbp->b_bname) ;
        }
    }
#endif
    /* Does the menu line need updating? if so do it! */
    if((TTsrow > 0) && (vvideo.video[0].flag & VFCHNGD))
        osdMainMenuUpdate(n) ;
    
    /* update any windows that need refreshing */
    for (wp = wheadp; wp != NULL; wp = wp->w_wndp)
    {
        if(n)
        {
            wp->l_bufp = wp->w_bufp ;
            wp->w_flag |= WFUPGAR ;
        }
        else if(wp->l_bufp != wp->w_bufp)
        {
            wp->l_bufp = wp->w_bufp ;
            wp->w_flag |= WFMODE|WFREDRAW|WFMOVEL|WFSBOX ;
        }
        /* if top if window is the last line and there's more than
         * one, force refame and draw */
        if((wp->topLineNo == wp->w_bufp->elineno) && wp->topLineNo)
            wp->w_flag |= WFFORCE ;

        /* if the window has changed, service it */
        if(wp->w_flag & (WFMOVEL|WFFORCE))
            reframe(wp) ;	        /* check the framing */

        if(wp->w_flag & (WFREDRAW|WFRESIZE|WFSELHIL|WFSELDRAWN))
            shilightWindow(wp);         /* Update selection hilight */
        
        /* check the horizontal scroll and cursor position */
        if(wp == curwp)
            updCursor(wp) ;
        
        if(wp->w_flag & ~(WFSELDRAWN|WFLOOKBK))
        {
            /* if the window has changed, service it */
            if(wp->w_flag & WFSBOX)
                fixWindowScrollBox(wp);     /* Fix the scroll bars */
            if(wp->w_flag & ~(WFMODE|WFSBAR|WFLOOKBK))
                updateWindow(wp) ;          /* Update the window */
            updateModeLine(wp);  	    /* Update mode line */
            if(wp->w_flag & WFSBAR)
                updateScrollBar(wp);        /* Update scroll bar  */
            wp->w_flag &= WFSELDRAWN ;
        }
        wp->w_force = 0;
    }
    
    /* If forced then sort out the message-line as well */
    if(vvideo.video[TTnrow].flag & VFCHNGD)
    {
        updateline(TTnrow,vvideo.video+TTnrow,NULL) ;
        vvideo.video[TTnrow].flag &= ~VFCHNGD ;
    }
    /* If we are in osd then update the osd menus */
    if(osdDisplayHd != NULL)
        osdRestoreAll(n) ;

    /* update the cursor and flush the buffers */
    resetCursor() ;
    
    TTflush();
    
    /* Now rest all the window line LNCHNG flags */
    for (wp = wheadp; wp != NULL; wp = wp->w_wndp)
    {
        register LINE *flp, *blp ;
        register int   ii, jj ;
        register unsigned char flag ;
        
        ii = (wp->line_no - wp->topLineNo) ;
        jj = (wp->numTxtRows - ii) ;
        blp = flp = wp->w_dotp ;
        while(--ii >= 0)
        {
            blp = lback(blp) ;
            if((flag=blp->l_flag) & LNCHNG)
                blp->l_flag = (flag & ~LNCHNG) ;
        }
        while(--jj >= 0)
        {
            if((flag=flp->l_flag) & LNCHNG)
                flp->l_flag = (flag & ~LNCHNG) ;
            flp = lforw(flp) ;
        }
    }
    return(TRUE);
}

int
update(int flag)    /* force=TRUE update past type ahead? */
{
    unsigned int arg ;
    register int index;
        
    if((alarmState & 0x80) ||
       (!(flag & 0x01) && ((kbdmode == PLAY) || clexec || TTahead())))
        return TRUE ;
    
    if((index = decode_key(ME_SPECIAL|SKEY_redraw,&arg)) >= 0)
        execFuncHidden(ME_SPECIAL|SKEY_redraw,index,0x80000000+sgarbf) ;
    else
        screenUpdate(1,sgarbf) ;
    /* reset garbled status */
    sgarbf = FALSE ;

    return TRUE ;
}

/*
 * pokeScreen.
 * Poke some text on the screen.
 *
 * flags are a bit pattern with the following meanings
 * 
 * 0x01 POKE_NOMARK - Dont remember (or mark) the poke so that at the next
 *                    update this may not be erased (depending on what upates
 *                    are required.
 * 0x02 POKE_NOFLUSH- Don't flush the output terminal, great for splatting a
 *                    lot of pokes up at once (only the last one calls a flush)
 *                    This has limited success on Xterms as they seem to flush
 *                    anyway.
 * 0x04 POKE_COLORS - If not set then fbuf and bbuf point to a single fore and
 *                    back-ground color to be used for the whole poke. If set
 *                    then a fore & back color is given per poke char.
 */
void
pokeScreen(int flags, int row, int col, unsigned char *scheme,
           unsigned char *str)
{
    int            len, schm, off ;
    unsigned char  cc, *ss, fcol, bcol ;
    unsigned short *fscp;               /* Frame store colour pointer */

    /* Normalise to the screen rows */
    if((row < 0) || (row > TTnrow))     /* Off the screen in Y ?? */
        return ;                        /* Yes - quit */
    
    /* Normalise to the screen columns. Chop the string up if necessary
     * and change any bad chars to '.'s
     */
    len = 0 ;
    ss = str ;
    while((cc = *ss) != '\0')
    {
        len++ ;
        if(!isPokable(cc))
            *ss = '.' ;
        ss++ ;
    }
    if(col < 0)
    {
        if((len += col) <= 0)
            return ;
        str -= col ;
        if(flags & POKE_COLORS)
            scheme -= col ;
        col = 0 ;
    }
    else if(col >= TTncol)
        return ;
    if((col+len) >= TTncol)
        len = TTncol - col ;

    /* Sort out the poke flags. Just keep the minimal and maximal extents
     * of the change. We will sort out what needs to be updated on the 
     * next update() call. */
    if((flags & POKE_NOMARK) == 0)
    {
        poke_flag |= 1;
        if (row < poke_ymin)
            poke_ymin = row;
        if (row > poke_ymax)
            poke_ymax = row;
        if (col < poke_xmin)
            poke_xmin = col;
        if ((col+len) > poke_xmax)
            poke_xmax = col+len;
    }
    
    /* Write the text to the frame store. Note that the colour still 
     * needs to be updated. */
    memcpy(&frameStore[row].text[col],str,len) ;    /* Write text in */
    fscp = &frameStore[row].colour[col] ;           /* Get the colour pointer */
    off  = (flags >> 4) & 0x07 ;
    
#ifdef _WIN32
#ifdef _WINCON
    if (!(meSystemCfg & meSYSTEM_CONSOLE))
#endif
        TTputs(row,col,len) ;
#endif
    
    /* Write to the screen */
    if(flags & POKE_COLORS)
    {
        /************************************************************************
         * 
         * X-WINDOWS
         * 
         ***********************************************************************/
#ifdef _XTERM
        if(!(meSystemCfg & meSYSTEM_CONSOLE))
        {
            XGCValues xgcv ;
            
            col *= fwidth ;
            row = (fdepth*row)+ascent ;
            
            if (meSystemCfg & meSYSTEM_FONTFIX)
            {
                unsigned char cc ;
                str-- ;
                while(len--)
                {
                    schm = *scheme++ ;
                    if((schm == 0xff) && ((schm = *scheme++) == 1))
                        schm = 0 ;
                    else
                        schm = TTcolourSchmCheck(schm)*HCTNUM ;
                    schm += off ;
                    fcol = colSchmTbl[schm] ;
                    bcol = colSchmTbl[schm^1] ;
                    /* Update the frame store colour */
                    *fscp++ = FScolourPair(fcol,bcol) ;
                    xgcv.foreground = colTable[fcol] ;
                    xgcv.background = colTable[bcol] ;
                    XChangeGC(xdisplay,xgc,(GCForeground|GCBackground),&xgcv) ;
                    if((cc=*++str) & 0xe0)
                        XDrawImageString(xdisplay,xwindow,xgc,col,row,(char *) str,1);
                    else
                    {
                        static char ss[1]={' '} ;
                        XDrawImageString(xdisplay,xwindow,xgc,col,row,ss,1) ;
                        XTERMSpecialChar(col,row-ascent,cc) ;
                    }
                    col += fwidth ;
                }
            }
            else
            {
                while(len--)
                {
                    schm = *scheme++ ;
                    if((schm == 0xff) && ((schm = *scheme++) == 1))
                        schm = 0 ;
                    else
                        schm = TTcolourSchmCheck(schm)*HCTNUM ;
                    schm += off ;
                    fcol = colSchmTbl[schm] ;
                    bcol = colSchmTbl[schm^1] ;
                    /* Update the frame store colour */
                    *fscp++ = FScolourPair(fcol,bcol) ;
                    xgcv.foreground = colTable[fcol] ;
                    xgcv.background = colTable[bcol] ;
                    XChangeGC(xdisplay,xgc,(GCForeground|GCBackground),&xgcv) ;
                    XDrawImageString(xdisplay,xwindow,xgc,col,row,(char *) str++,1);
                    col += fwidth ;
                }
            }
        }
        else
#endif
            /************************************************************************
             * 
             * TERMCAP
             * 
             ***********************************************************************/
#ifdef _TCAP
        {
/*            TCAPputc(0) ;*/
            TCAPgoto(row, col);	/* Go to correct place. */
/*            TCAPputc(0) ;*/
            while(len--)
            {
                schm = *scheme++ ;
                if((schm == 0xff) && ((schm = *scheme++) == 1))
                    schm = 0 ;
                else
                    schm = TTcolourSchmCheck(schm)*HCTNUM ;
                schm += off ;
#if COLOR
#ifdef _TCAPFONT
                if(meSystemCfg & meSYSTEM_ANSICOLOR)
                {
#endif /* _TCAPFONT */
                    fcol = colSchmTbl[schm] ;
                    bcol = colSchmTbl[schm^1] ;
                    TCAPcolorSetScheme(fcol,bcol) ;
#ifdef _TCAPFONT
                }
                else if(meSystemCfg & meSYSTEM_FONTS)
                {
#endif /* _TCAPFONT */
#endif /* COLOR  */
#ifdef _TCAPFONT
                    fcol = fntSchmTbl[schm];
                    bcol = TCAPfontSetScheme(fcol) ;
#if COLOR
                }
#endif /* COLOR  */
#endif /* _TCAPFONT */
                *fscp++ = FScolourPair(fcol,bcol) ;
                cc = *str++ ;
                TCAPputc(cc) ;
            
                /* Update the frame store colour */
            }
#if COLOR
#ifdef _TCAPFONT
            if(meSystemCfg & meSYSTEM_ANSICOLOR)
#endif
                TCAPcolorReset() ;
#endif
#ifdef _TCAPFONT
            if(vstate)
                TCAPfontReset() ;
#endif
        }
        
#endif /* _TCAP */
        /************************************************************************
         * 
         * MS-DOS
         * 
         ***********************************************************************/
#ifdef _DOS
        {
            while(len--)
            {
                schm = *scheme++ ;
                if((schm == 0xff) && ((schm = *scheme++) == 1))
                    schm = 0 ;
                else
                    schm = TTcolourSchmCheck(schm)*HCTNUM ;
                schm += off ;
                fcol = colSchmTbl[schm] ;
                bcol = colSchmTbl[schm^1] ;
                /* Update the frame store colour */
                *fscp++ = FScolourPair(fcol,bcol) ;
                cc = TTCmake(colTable[fcol],colTable[bcol]) ;
                ScreenPutChar(*str++, cc, col++, row);
            }
        }
#endif /* _DOS */
        /************************************************************************
         * 
         * MS-WINDOWS
         * 
         ***********************************************************************/
#ifdef _WIN32
        {
            while(len--)
            {
                schm = *scheme++ ;
                if((schm == 0xff) && ((schm = *scheme++) == 1))
                    schm = 0 ;
                else
                    schm = TTcolourSchmCheck(schm)*HCTNUM ;
                schm += off ;
                fcol = colSchmTbl[schm] ;
                bcol = colSchmTbl[schm^1] ;
                /* Update the frame store colour */
                *fscp++ = FScolourPair(fcol,bcol) ;
#ifdef _WINCON
                if (meSystemCfg & meSYSTEM_CONSOLE)
                {
                    cc = TTCmake(colTable[fcol],colTable[bcol]) ;
                    ConsoleDrawString(str++, cc, col++, row, 1);
/*                    TTflush() ;*/
                }
#endif
            }
        }
#endif /* _WIN32 */
    }
    else
    {
        unsigned short colpair;             /* Colour pair */
        
        schm = *scheme ;
        schm = TTcolourSchmCheck(schm)*HCTNUM ;
        schm += off ;
        
        /************************************************************************
         * 
         * X-WINDOWS
         * 
         ***********************************************************************/
#ifdef _XTERM
        if(!(meSystemCfg & meSYSTEM_CONSOLE))
        {
            XGCValues xgcv ;
            
            fcol = colSchmTbl[schm] ;
            bcol = colSchmTbl[schm^1] ;
            colpair = FScolourPair (fcol, bcol);
            xgcv.foreground = colTable[fcol] ;
            xgcv.background = colTable[bcol] ;
            XChangeGC(xdisplay,xgc,(GCForeground|GCBackground),&xgcv) ;
            if (meSystemCfg & meSYSTEM_FONTFIX)
            {
                unsigned char cc, *sfstp, *fstp ;
                int ii, spFlag=0 ;
                
                ii = len ;
                sfstp = fstp = &(frameStore[row].text[col]) ;
                row = (fdepth*row)+ascent ;
                while (--len >= 0)
                {
                    *fscp++ = colpair;
                    if(((cc=*fstp++) & 0xe0) == 0)
                    {
                        spFlag++ ;
                        fstp[-1] = ' ' ;
                    }
                }
                XDrawImageString(xdisplay,xwindow,xgc, 
                                 col*fwidth,row,
                                 (char *) sfstp, ii);
                ii = 0 ;
                while(--spFlag >= 0)
                {
                    while (((cc=str[ii]) & 0xe0) != 0)
                        ii++ ;
                    sfstp[ii] = cc ;
                    XTERMSpecialChar((col+ii)*fwidth,row-ascent,cc) ;
                    ii++ ;
                }
            }
            else
            {
                XDrawImageString(xdisplay,xwindow,xgc,(fwidth*col),(fdepth*row)+ascent,
                                 (char *) str, len);
                /* Update the frame store colour */
                while (--len >= 0)
                    *fscp++ = colpair;
            }
        }
        else
#endif
            /************************************************************************
             * 
             * TERMCAP
             * 
             ***********************************************************************/
#ifdef _TCAP
        {
/*            TCAPputc(0) ;*/
            TCAPgoto(row, col);	/* Go to correct place. */
/*            TCAPputc(0) ;*/
#if COLOR
#ifdef _TCAPFONT
            if(meSystemCfg & meSYSTEM_ANSICOLOR)
            {
#endif /* _TCAPFONT */
                fcol = colSchmTbl[schm] ;
                bcol = colSchmTbl[schm^1] ;
                TCAPcolorSetScheme(fcol,bcol) ;
#ifdef _TCAPFONT
            }
            else if(meSystemCfg & meSYSTEM_FONTS)
            {
#endif /* _TCAPFONT */
#endif /* COLOR  */
#ifdef _TCAPFONT
                fcol = fntSchmTbl[schm];
                bcol = TCAPfontSetScheme(fcol) ;
#if COLOR
            }
#endif /* COLOR  */
#endif /* _TCAPFONT */
            colpair = FScolourPair (fcol, bcol);
            
            while(len--)
            {
                cc = *str++ ;
                TCAPputc(cc) ;
                /* Update the frame store colour */
                *fscp++ = colpair;
            }
#if COLOR
#ifdef _TCAPFONT
            if(meSystemCfg & meSYSTEM_ANSICOLOR)
#endif
                TCAPcolorReset() ;
#endif
#ifdef _TCAPFONT
            if(vstate)
                TCAPfontReset() ;
#endif
            /* restore cursor position */
            if((cursorState >= 0) && blinkState)
                TTshowCur() ;
            else
                TThideCur() ;
        }
        
#endif /* _TCAP */
        /************************************************************************
         * 
         * MS-DOS
         * 
         ***********************************************************************/
#ifdef _DOS
        {
            fcol = colSchmTbl[schm] ;
            bcol = colSchmTbl[schm^1] ;
            colpair = FScolourPair (fcol, bcol);
            cc = TTCmake(colTable[fcol],colTable[bcol]) ;
            while(len--)
            {
                ScreenPutChar(*str++, cc, col++, row);
                /* Update the frame store colour */
                *fscp++ = colpair;
            }
        }
#endif /* _DOS */
        /************************************************************************
         * 
         * MS-WINDOWS
         * 
         ***********************************************************************/
#ifdef _WIN32
        {
            fcol = colSchmTbl[schm] ;
            bcol = colSchmTbl[schm^1] ;
            colpair = FScolourPair (fcol, bcol);
#ifdef _WINCON
            if (meSystemCfg & meSYSTEM_CONSOLE) 
            {
                cc = TTCmake(colTable[fcol],colTable[bcol]) ;
                ConsoleDrawString (str, cc, col, row, len);
/*                TTflush() ;*/
            }
#endif
            /* Update the frame store colours */
            while (--len >= 0)
                *fscp++ = colpair;
        }
#endif /* _WIN32 */
    }
    if((flags & POKE_NOFLUSH) == 0)
        TTflush() ;                         /* Force update of screen */
}

/*
 * screenPoke.
 * The macro interface to the pokeScreen function. This accepts a 
 * numeric argument which is a bit map of the form passed into pokeScreen.
 * As the default argument is 1 this will set just the POKE_NOMARK flag
 * which means that the poke will be visible after an ecs x invocation of
 * screen-poke.
 */
int
screenPoke(int f, int n)
{
    int row, col ;
    unsigned char fbuf[MAXBUF] ;
    unsigned char sbuf[MAXBUF] ;

    if((mlreply("row",0,0,(char *) fbuf,MAXBUF) != TRUE) ||
       ((row = meAtoi((char *) fbuf)),(mlreply("col",0,0,(char *) fbuf,MAXBUF) != TRUE)) ||
       ((col = meAtoi((char *) fbuf)),(mlreply("scheme",0,0,(char *) fbuf,MAXBUF) != TRUE)) ||
       (mlreply("string",0,0,(char *) sbuf,MAXBUF) != TRUE))
        return FALSE ;
    if((n & POKE_COLORS) == 0)
        fbuf[0] = (unsigned char) meAtoi((char *) fbuf) ;
    pokeScreen(n,row,col,fbuf,sbuf) ;
    return TRUE ;
}

/*
 * Write out a long integer, in the specified radix. Update the physical cursor
 * position.
 *
 * As usual, if the last argument is not NULL, write characters to it and
 * return the number of characters written.
 */
static	int
mlputi(long i, int r, char *buf)
{
    char	   temp_buf [22];		/* Temporary buffer */
    register char *temp = temp_buf;	/* Temporaty buffer pointer */
    register int   count=0, neg=0 ;

    *temp++ = '\0' ;
    if (i < 0)
    {	/* Number is negetive */
        i = -i;
        neg = 1 ;
    }
    do
    {
        count++;
        *temp++ = hexdigits [i%r];
    } while ((i = i/r) > 0) ;

    if(neg)
    {
        count++ ;
        *temp++ = '-';
    }

    /*---	Write the information out */

    while ((*buf++ = *--temp) != 0)
        ;
    return(count);
}

/*
 * Write a message into the message line. Keep track of the physical cursor
 * position. A small class of printf like format items is handled. Assumes the
 * stack grows down; this assumption is made by the "++" in the argument scan
 * loop. Set the "message line" flag TRUE.
 */
#ifdef _STDARG
int
mlwrite(int flags, char *fmt, ...)
#else
int
mlwrite(int flags, char *fmt, int arg)
#endif
{
    register char	c;
#ifdef _STDARG
    va_list	 ap;
#else
    register char *ap;
#endif
    register VIDEO *vp1;
    register int offset ;		/* offset into ???		*/
    char	 mlw[MAXBUF];		/* what we want to be there	*/
    char	*mlwant;		/* what we want to be there	*/
    char        *s1 ;

    if((alarmState & 0x80) ||
       (clexec && (flags & MWCLEXEC)))
        goto mlwrite_exit ;

    if(mlStatus & MLSTATUS_KEEP)
    {
        char *_ss, *_dd ;
        _ss = (char *) mline->l_text ;
        _dd = mlStore ;
        while((*_dd++ = *_ss++))
            ;
        mlStoreCol = mlCol ;
        mlStatus = (mlStatus & ~MLSTATUS_KEEP) | MLSTATUS_RESTORE ;
    }
    else
        mlStatus |= MLSTATUS_CLEAR ;
        
    if(flags & MWSPEC)
    {
        mlwant = fmt ;
        offset = strlen(mlwant) ;
    }
    else
    {
#ifdef _STDARG
        va_start(ap,fmt);
#else
        ap = (char *) &arg;
#endif
        mlwant = mlw ;
        while ((c = *fmt++) != '\0')
        {
            if(c != '%')
                *mlwant++ = c;
            else
            {
                c = *fmt++;
                switch (c)
                {
                case 'c':
#ifdef _STDARG
                    *mlwant++ = va_arg(ap, int) ;
#else
                    *mlwant++ = *(char *)ap ;
                    ap += sizeof(int);
#endif
                    break;

                case 'd':
#ifdef _STDARG
                    mlwant += mlputi((long)(va_arg(ap, int)),10,mlwant);
#else
                    mlwant += mlputi((long)(*(int *)ap), 10,mlwant);
                    ap += sizeof(int);
#endif
                    break;

                case 'o':
#ifdef _STDARG
                    mlwant += mlputi((long)(va_arg(ap, int)),8,mlwant);
#else
                    mlwant += mlputi((long)(*(int *)ap),8,mlwant);
                    ap += sizeof(int);
#endif
                    break;

                case 'x':
#ifdef _STDARG
                    mlwant += mlputi((long)(va_arg(ap, int)),16,mlwant);
#else
                    mlwant += mlputi((long)(*(int *)ap),16,mlwant);
                    ap += sizeof(int);
#endif
                    break;

                case 'D':
#ifdef _STDARG
                    mlwant += mlputi(va_arg(ap, long), 10,mlwant) ;
#else
                    mlwant += mlputi(*(long *)ap,10,mlwant) ;
                    ap += sizeof(long);
#endif
                    break;

                case 's':
#ifdef _STDARG
                    s1 = va_arg(ap, char *) ;
#else
                    s1 = *(char **)ap ;
                    ap += sizeof(char *);
#endif
                    /* catch a string which is too long */
                    offset = strlen(s1) ;
                    if(offset > TTncol)
                    {
                        mlwant = mlw ;
                        s1 += offset - TTncol ;
                        offset = TTncol ;
                    }
                    strcpy(mlwant,s1) ;
                    mlwant += offset ;
                    break;

                default:
                    *mlwant++ = c;
                    break;
                } /* switch */
            } /* if */
        } /* while */
#ifdef _STDARG
        va_end (ap);
#endif
        *mlwant= '\0' ;
        offset = mlwant - mlw ;
        mlwant = mlw ;
    }
    vp1 = vvideo.video + TTnrow ;
    /* JDG End */
    /* SWP - changed the mlwrite when the string is to long for the ml line
     * so that it always displays the last part. This is for two reasons:
     * 1) If this is an MWSPEC then it must not touch the string, so can't
     *    null terminate.
     * 2) The latter part is usually the more important, i.e. where the 
     *    prompt is
     */
    if(offset > TTncol)
    {
        mlwant += offset-TTncol ;
        offset = TTncol ;
        if(!(flags & MWUSEMLCOL))
            mlCol = TTncol-1 ;
    }
    else if(!(flags & MWUSEMLCOL))
        mlCol = offset ;
    s1 = (char *) mline->l_text ;
    while((*s1++ = *mlwant++))
        ;
    mline->l_used=offset ;
    /* JDG 02/03/97 - This line was removed, however when it is then the 
     * bottom line is not reset properly.
     * SWP - I've commented it out again - this only alleviates another bug
     * which should be/should have been fixed.
     */
    /* vp1->endp = TTncol ;*/
    
    updateline(TTnrow,vp1,NULL);
    mlResetCursor() ;
    TTflush();
    
    if(flags & MWABORT)
        TTbell() ;
    if(flags & MWPAUSE)
    {
        /* Change the value of mlStatus to MLSTATUS_KEEP cos we want to keep
         * the string for the length of the sleep
         */
        unsigned char oldMlStatus = mlStatus ;
        mlStatus = MLSTATUS_KEEP ;
        TTsleep(2000,1)  ;
        mlStatus = oldMlStatus ;
    }
    else if((flags & MWWAIT) || (clexec && (flags & MWCLWAIT)))
    {
        /* Change the value of mlStatus to MLSTATUS_KEEP cos we want to keep
         * the string till we get a key
         */
        unsigned char scheme=(globScheme/HCTNUM), oldMlStatus = mlStatus ;
        pokeScreen(POKE_NOMARK+0x10,TTnrow,TTncol-9,&scheme,
                   (unsigned char *) "<ANY KEY>") ;
        vp1->endp = TTncol ;
        mlStatus = MLSTATUS_KEEP ;
        TTgetc() ;
        mlStatus = oldMlStatus ;
        mlerase(0) ;
    }
    if(!(flags & MWCURSOR))
        resetCursor();
    TTflush();

mlwrite_exit:
    if(!(flags & MWABORT))
        return TRUE ;
    
    if(macbug < -2)
        macbug = 1 ;
    return ABORT ;
}


/* 
 * addColor
 * Add a new colour palett entry, or modify existing colour
 * palette entry.
 *
 * add-colour <index> <red> <green> <blue>
 *
 * Modify the colour  palette  values. The default is black (0) and white (1).
 * Any number of colours may be specified. Colors are specified as rgb values
 * in the range 0-255 where 0 is the black level; 255 is bright.
 *
 * The  <index>  need not  necesserily  commence  from zero and may exceed 256
 * entries. The support  offered by the system is dependent  upon the graphics
 * sub-system.
 */
int
addColor(int f, int n)
{
    unsigned char index, r, g, b ;
    char buff[MAXBUF] ;
    
    if((mlreply("Index",0,0,buff,MAXBUF) == ABORT) ||
       ((index = (unsigned char) meAtoi(buff)),
        (mlreply("Red",0,0,buff,MAXBUF) == ABORT)) ||
       ((r = (unsigned char) meAtoi(buff)),
        (mlreply("Green",0,0,buff,MAXBUF) == ABORT)) ||
       ((g = (unsigned char) meAtoi(buff)),
        (mlreply("Blue",0,0,buff,MAXBUF) == ABORT)))
        return FALSE ;
    b = (unsigned char) meAtoi(buff) ;
    
    n = TTaddColor(index,r,g,b) ;
    if(index == colSchmTbl[globScheme^1])
        TTsetBgcol() ;
    return n ;
}


/* 
 * addColorScheme
 * Get the hilight colour definition.
 */
/* Brought out as add-font also uses the same prompts */
char *schmPrompts[HCTNUM] = 
{
    "Fore",
    "Back",
    "Cur fore",
    "Cur back",
    "Sel fore",
    "Sel back",
    "Cur Sel fore",
    "Cur Sel back",
} ;
int
addColorScheme(int f, int n)
{
    char           buf[MAXBUF] ;
    unsigned char *hcolors ;
    unsigned char *dcolors ;
    int index;
    int ii;
#ifdef _TCAPFONT
    int jj;
#endif
    int col;
    
    /* Get the hilight colour index */
    if ((mlreply("Color schem index",0,0,buf,MAXBUF) != TRUE) ||
        ((index = meAtoi(buf)) < 0))
        return mlwrite(MWABORT|MWPAUSE,"Invalid scheme index number");
    
    /* See if we are replicating a color and validate the existing index */
    if ((f) && ((n > colSchmTblSize) || (n < 0)))
        return mlwrite (MWABORT|MWPAUSE, "Cannot duplicate undefined scheme %d.", n);
    else
        dcolors = colSchmTbl + (n*HCTNUM);
    
    /* Allocate a new table if required */
    if (index >= colSchmTblSize)
    {
        /* Allocate a new table */
        if(((colSchmTbl = meRealloc(colSchmTbl,(index+1) * HCTNUM)) == NULL)
#ifdef _TCAPFONT
           || 
           ((fntSchmTbl = meRealloc(fntSchmTbl,(index+1) * HCTNUM)) == NULL)
#endif
           )
            return mlwrite (MWABORT|MWPAUSE,"Cannot allocate scheme table");
        for(ii=colSchmTblSize ; ii<=index ; ii++)
            memcpy(colSchmTbl+(ii*HCTNUM),defScheme,HCTNUM) ;
#ifdef _TCAPFONT
        memset(fntSchmTbl+(colSchmTblSize*HCTNUM),0,(index-colSchmTblSize+1)*HCTNUM) ;
#endif
        colSchmTblSize = index+1 ;
    }
    hcolors = colSchmTbl+(index*HCTNUM) ;
#ifdef _TCAPFONT
    for (jj = 0; jj < 2 ; jj++,
         (hcolors=fntSchmTbl+(index*HCTNUM)),
         (dcolors=fntSchmTbl+(n*HCTNUM)))
    {
#endif
        /* Get the separate colour indicies from the user. */
        for (ii = 0; ii < HCTNUM; ii++)
        {
            /* Copy entry on duplicate */
            if (f)
                hcolors[ii] = dcolors[ii];
            else
            {       
                /* Prompt for entries */
                if ((mlreply(schmPrompts[ii],0,0,buf,MAXBUF) != TRUE))
                    break;
                if (buf[0] == '*')
                {
                    if(ii < 2)
                        break;
                    col = hcolors[ii - 2];
                }
                else if ((col = meAtoi(buf)) < 0)
                    break;
                else
                {
#ifdef _XTERM
                    if (!(meSystemCfg & meSYSTEM_CONSOLE) && (col >= noColors))
                        break;
#endif
#ifdef _DOS
                    if (col >= noColors)
                        break;
#endif
#ifdef _WIN32
                    if (col > 255)
                        break;
#endif
                }
                hcolors[ii] = col;
            }
        }
        /* Return status to caller */
        if(ii < HCTNUM)
        {
            for(; ii<HCTNUM ; ii++)
                hcolors[ii] = 0 ;
#ifdef _TCAPFONT
            if(jj == 1)
                break ;
#endif
            return mlwrite(MWABORT|MWPAUSE, "Invalid scheme entry");
        }
#ifdef _TCAPFONT
    }
#endif
    if(index == globScheme)
        TTsetBgcol() ;
    return TRUE;
}


/************************** VIRTUAL VIDEO INTERFACES **************************
 *
 * vvideoDetach
 * Detach the window from the virtual video block. If the virtual video 
 * block has no more windows attached then destruct the block itself.
 */

void
vvideoDetach (WINDOW *wp)
{
    VVIDEO *vvptr;
    WINDOW *twp;
    
    vvptr = wp->w_vvideo;               /* Get windows virtual video block */
    meAssert (vvptr != NULL);
    
    if ((twp = vvptr->window) == wp)    /* First window in the chain ?? */
    {
        vvptr->window = wp->w_lvideo;   /* Detach from the head */
        wp->w_vvideo = NULL;
    }
    else                                /* Not first in chain - chase block */
    {
        while (twp != NULL)
        {
            if (twp->w_lvideo == wp)
            {
                twp->w_lvideo = wp->w_lvideo;
                wp->w_vvideo = NULL;
                break;
            }
            else
                twp = twp->w_lvideo;
        }
    }
    
    meAssert (twp != NULL);
    
    /* Clean up the Virtual video frame if it is empty. Note
     * that vvptr == vvideo is handled correctly and is not
     * deleted. */
    if (vvptr->window == NULL)
    {
        VVIDEO *vv;
    
        for (vv = &vvideo; vv != NULL; vv = vv->next)
        {
            if (vv->next == vvptr)
            {
                vv->next = vvptr->next;
                meNullFree (vvptr->video);
                meFree (vvptr);
                break;
            }
        }
    }
}
            
/*
 * vvideoAttach
 * Attach a window to the virtual video block. If no virtual video block
 * is specified then construct a new block and link it into the existing 
 * structure.
 */

int
vvideoAttach (VVIDEO *vvptr, WINDOW *wp)
{
    /* Construct a new virtual video block if required */
    if (vvptr == NULL)
    {
        VIDEO  *vs;                     /* New video structure */
        
        /* Allocate a new video structure */
        if (((vvptr = (VVIDEO *) meMalloc (sizeof (VVIDEO))) == NULL) ||
            ((vs = (VIDEO *) meMalloc(TTmrow*sizeof(VIDEO))) == NULL))
        {
            meNullFree (vvptr);
            return (FALSE);
        }
        
        /* Reset the strucure and chain to the existing vvideo list */
        memset (vs, 0, TTmrow * sizeof(VIDEO));
        memset (vvptr, 0, sizeof (VVIDEO));
        vvptr->video = vs;              /* Attach the video structure */
        vvptr->next = vvideo.next;      /* Chain into vvideo list */
        vvideo.next = vvptr;
        
        /* Reset information in the Window block */
        wp->w_lvideo = NULL;
    }
    
    /* Link the window to the vvideo block */
    wp->w_vvideo = vvptr;
    wp->w_lvideo = vvptr->window;
    vvptr->window = wp;
    return (TRUE);
}
   

/*
 * vvideoEnlarge
 * Enlarge the video frames attached to the virtual video structures.
 * Usually follows an enlargement of the screen. */
int
vvideoEnlarge (int rows)
{
    VVIDEO *vvptr;                      /* Pointer to the video blocks */
    VIDEO  *vs;                         /* Video store line */
    
    for (vvptr = &vvideo; vvptr != NULL; vvptr = vvptr->next)
    {
        /* Try and allocate new video frame */
        if ((vs = (VIDEO *) meMalloc (rows * sizeof (VIDEO))) == NULL)
            return (FALSE);
        
        /* Allocation successful. Reset the contents to zero and swap
           for the existing one */
        memset (vs, 0, rows * sizeof (VIDEO));  /* Reset to zero */
        meFree (vvptr->video);                  /* Dispose of old one */
        vvptr->video = vs;                      /* Swap in new one */
    }
    
    return (TRUE);
}



