/*
-- ----------------------------------------------------------------------------
--
--  			Copyright 1994 Jasspa
--
--  System       : MicroEmacs Jasspa Distribution
--  Module       : Termcap
--  Object Name  : $RCSfile$
--  Revision     : $Revision$
--  Date         : $Date$
--  Author       : $Author$
--
--  Description	
--       Termcap definitions
--
--  Notes
--
--  History
--	
-- ----------------------------------------------------------------------------
--
-- Copyright (C) 1994 - 1999, JASSPA 
--
-- The MicroEmacs Jasspa distribution can be copied and distributed freely for
-- any non-commercial purposes. The MicroEmacs Jasspa Distribution can only be
-- incorporated into commercial software with the expressed permission of
-- JASSPA.
--  
-- ----------------------------------------------------------------------------
*/

/* terminal io include file
 *
 * every system must provide the following variables / constants
 *
 *    TTmrow    - maximum number of rows
 *    TTnrow    - current number of rows
 *    TTmcol    - maximum number of columns
 *    TTncol    - current number of columns
 *    TTmargin  - minimum margin for extended lines
 *    TTscrsiz  - size of scroll region
 *    TTpause   - # times thru update to pause
 *
 * every system must provide the following functions
 *
 *    TTstart   - Creates the terminal (called once at start)
 *    TTend     - Destroyes the terminal (called once at end)
 *    TTopen    - opens the terminal called after any shell call.
 *    TTclose   - closes the terminal called before any shell call.
 *    TTwaitForChar - gets a character
 *    TTflush   - flushes the output terminal
 *    TTmove    - most the cursor to a given position
 *    TThideCur - hide the cursor
 *    TTshowCur - show the cursor
 *    TTNbell   - Do a noisy beep
 *
 * The following must be provided if TYPEAH=TRUE
 *
 *    TTahead   - any characters waiting ?
 *
 * The following must be provided if _TCAPFONT is true
 *
 *    addFont   - adds a user font to the list 
 *    TCAPfontReset  - resets the state
 *    TCAPfontSetScheme - Sets the fonts for the given colour
 *
 * For colour there must be the following functions
 *
 *    TTCreset  - resets the current fore & background colour to default
 *    TCAPcolorSetForeg   - sets the current foreground colour
 *    TCAPcolorSetBackg   - sets the current background colour
 *
 * following must be provided if COLOR=TRUE & CSWITCH=FALSE
 *
 *    TTwrite  - writes a string given the length and the colour 
 *
 * the following are the prototypes for them
 * 
 * int  TTopen(void) ;
 * int  TTclose(void) ;
 * void TTwaitForChar(void) ;
 * void TTflush(void) ;
 * void TTmove(int, int) ;
 * void TTclear(void) ;
 * void TTshowCur(void) ;
 * void TTNbell(void) ;
 * int  TTahead(void) ;
 *
 * int  addFont(int f, int n) ;
 * void TCAPfontReset(void) ;
 * void TCAPfontSetScheme(unsigned short col) ;
 * 
 * TTbreakTest(n) - Test for a break from the keyboard. The numeric argument
 * 'n' is set to 1 if the break test may include 'display update' messages,
 * or 0 if it is only allowed to collect user input. Note that not all platforms
 * support this argument.
 */
#ifndef __TERMIOH
#define __TERMIOH

/***********************************************************************
 * 
 * Common variables, constants and defines
 * 
 **********************************************************************/
/* Default fore and back ground colours, plus valid colour testers */
extern unsigned char  noColors ;
#define TTdefFCol   0
#define TTdefBCol   1
#define TTdefSchm   0
#define TTcolourForeCheck(x) (((x)>=noColors)       ? TTdefFCol : (x))
#define TTcolourBackCheck(x) (((x)>=noColors)       ? TTdefBCol : (x))
#define TTcolourSchmCheck(x) (((x)>=colSchmTblSize) ? TTdefSchm:(x))

/* Standard specisl keys */
#define	BELL	0x07			/* a bell character		*/
#define	TAB	0x09			/* a tab character		*/

extern void setAlarm APRAM((long absTime, long offTime)) ;

/* Standard key input definitions, found in termio.c */
#define KEYBUFSIZ 128                   /* Length of the keyboard buffer */
typedef struct meTRANSKEY {
    int count ;
    unsigned short key ;
    unsigned short map ;
    struct meTRANSKEY *child ;
} meTRANSKEY ;

extern meTRANSKEY TTtransKey ;
extern unsigned short TTkeyBuf[KEYBUFSIZ] ;
extern unsigned char  TTnextKeyIdx, TTlastKeyIdx, TTnoKeys ;
extern unsigned char  TTbreakFlag, TTbreakCnt ;
extern unsigned char  TTallKeys;        /* Report all keys */
extern unsigned short TTgetc(void) ;
extern int TTcurr;                      /* Windows current row. */
extern int TTcurc;                      /* Windows current column */
extern int TTfocus;                     /* Is the window the current focus */
extern void TThandleBlink(int initFlag) ;
extern void TTmove(int r, int c) ;
#define TTinflush()   (TTahead(),TTlastKeyIdx=TTnextKeyIdx,TTnoKeys=0)
extern void addKeyToBuffer(unsigned short cc) ;
extern void doIdlePickEvent(void) ;
#define TTSPECCHARS    32               /* The end of the special chars */
extern char ttSpeChars [TTSPECCHARS];

#define TTBREAKCNT 100                  /* The initial ttbreakCnt value */

#ifdef _UNIX

#define	_SINGLE_TIMER 1		/* Only one itimer available  */
#define	TIMER_MIN     10	/* Only one itimer available  */

extern	int   tputs(char *, int, int (*)(int)) ;
extern	char *tgoto(char *, int, int ) ;

extern unsigned short TTmrow, TTnrow, TTsrow, TTmcol, TTncol, TTmargin, TTscrsiz ;

extern char  *CM, *CL ;

/* Following are functions used by termcap & an xterm */
#define TTNbell()      (putchar(BELL),fflush(stdout))
#define TTdieTest()    if(alarmState & 8) meDie()
#define TTbreakTest(x) ((--TTbreakCnt == 0) &&                         \
                       (((alarmState & 8) && meDie()) ||               \
                        (TTbreakCnt=TTBREAKCNT,TTahead(),TTbreakFlag)))
extern void TTwaitForChar(void) ;
extern void TTsleep(int msec, int intable) ;
#if TYPEAH
extern int  TTahead(void) ;
#endif

/* Following are termcap function */
extern int TCAPstart(void) ;
extern int TCAPopen(void) ;
extern int TCAPclose(void) ;
#define TCAPputc(c)   putchar(c)
#define TCAPgoto(r,c) ((void) tputs(tgoto(CM,c,r),1,putchar))
#ifdef _LINUX
/* Under LINUX with TCAP we need a destructive write to maintain
 * the colour of the text under the cursor. Abstract the TTmove()
 * invocation to ensure that colour is maintained. Note that we rely
 * on the maintainance of the frameStore information to give us the 
 * colour and text information to do the destructive write. */
extern  void TCAPmove(int r, int c) ;
#else
#define TCAPmove(r,c)  TCAPgoto(r,c)
#endif
#define TCAPflush()    fflush(stdout)
#define TCAPclear()    tputs(CL,1,putchar)
extern void TCAPhideCur(void) ;
extern void TCAPshowCur(void) ;
extern void TCAPhandleBlink(void) ;

#if COLOR
extern short Dforgc, Dbacgc ;
extern char RCOLSTR[], FCOLSTR[], BCOLSTR[] ;

extern int  TCAPaddColor(unsigned char index, unsigned char r,
                         unsigned char g, unsigned char b) ;
extern void TCAPcolorSetScheme(unsigned char fcol, unsigned char bcol) ;
#define TCAPcolorReset()      (tputs(RCOLSTR,1,putchar))
#define TCAPcolorSetForeg(c)  ((FCOLSTR[3]=((c)&0x07)+48),(tputs(FCOLSTR,1,putchar)))
#define TCAPcolorSetBackg(c)  ((BCOLSTR[3]=((c)&0x07)+48),(tputs(BCOLSTR,1,putchar)))
#endif

#ifdef _TCAPFONT
extern int           vstate ;
extern void          TCAPfontReset(void) ;
extern unsigned char TCAPfontSetScheme(unsigned char fcol) ;
#endif


#ifdef _XTERM
extern Display *xdisplay ;
extern Window   xwindow ;
extern GC       xgc ;
/* Font size information */
extern int      fdepth, fwidth, ascent ;

extern unsigned long *colTable ;

extern int  XTERMstartStage2(void);
extern void XTERMinitMouse(void);
extern void XTERMmove(int r, int c) ;
extern void XTERMclear(void) ;
extern void XTERMhideCur(void) ;
extern void XTERMshowCur(void) ;
extern int  XTERMaddColor(unsigned char index, unsigned char r,
                          unsigned char g, unsigned char b) ;
extern void XTERMSpecialChar(int x, int y, unsigned char cc) ;
extern void XTERMPaint(int srow, int scol, int erow, int ecol) ;

extern void TTsetBgcol(void) ;
extern int  TTstart(void) ;
#define TTstartStage2() ((!(meSystemCfg & meSYSTEM_CONSOLE)) ? XTERMstartStage2():0)

#define TTend()     ((meSystemCfg & meSYSTEM_CONSOLE) ? TCAPclose():0)
#define TTopen()    ((meSystemCfg & meSYSTEM_CONSOLE) ? TCAPopen():0)
#define TTclose()   ((meSystemCfg & meSYSTEM_CONSOLE) ? TCAPclose():0)
#define TTflush()   ((meSystemCfg & meSYSTEM_CONSOLE) ? TCAPflush():XFlush(xdisplay))
#define TThideCur() ((meSystemCfg & meSYSTEM_CONSOLE) ? TCAPhideCur():XTERMhideCur())
#define TTshowCur() ((meSystemCfg & meSYSTEM_CONSOLE) ? TCAPshowCur():XTERMshowCur())
#define TTaddColor(i,r,g,b) ((meSystemCfg & meSYSTEM_CONSOLE) ? TCAPaddColor(i,r,g,b):XTERMaddColor(i,r,g,b))

/* Some extra function, only available to xterms */
extern void TTtitleText (char *str) ;
extern void TTdepth(int y) ;
extern void TTwidth(int x) ;

#ifdef _CLIPBRD
extern void TTgetClipboard(void);
extern void TTsetClipboard(void);
#endif

#else
/* If no xterm then just use the termcap ones */
#define TTstart    TCAPstart
#define TTstartStage2()                 /* No stage 2 startup */
#define TTopen     TCAPopen
#define TTclose    TCAPclose
#define TTend      TCAPclose
#define TTflush    TCAPflush
#define TThideCur  TCAPhideCur
#define TTshowCur  TCAPshowCur
#define TTaddColor TCAPaddColor
#define TTsetBgcol()

#endif /* _XTERM */

extern int meSigLevel ;
#define meSigHold()    ((++meSigLevel == 0) ? sigprocmask(SIG_BLOCK, &meSigHoldMask, NULL):0)
#define meSigRelease() ((--meSigLevel  < 0) ? sigprocmask(SIG_UNBLOCK, &meSigHoldMask, NULL):0)

#ifdef _CLIENTSERVER
extern void TTopenClientServer(void) ;
extern void TTkillClientServer(void) ;
extern int  TTconnectClientServer(void) ;
extern void TTsendClientServer(char *) ;
#endif

#endif /* _UNIX */


#ifdef _WIN32

#define	_MULTI_TIMER 1		     /* Multiple timers can be set at once */
#ifdef _WINCON
/* strange but true, when running as a console app, you can't assign ids
 * so effectively when a timer pops you have to work out which one has
 * if any. Bummer is you can kill them either! Thank you Bill */
#define	_MULTI_NOID_TIMER 1	     /* Multi timer but no ids */
#endif
#define	TIMER_MIN    1		     /* The minimum timer allowed */
/* executable extension list, in reverse order of dos priority - Must have the end NULL
 * included 4dos's btm files for completeness */
#define noExecExtensions 4
extern char *execExtensions[noExecExtensions] ;

extern HWND ttHwnd ;                 /* This is the window handle */
extern RECT ttRect ;                 /* Area of screen to update */
extern int fdepth, fwidth ;
extern int ascent ;

extern unsigned short TTmrow, TTnrow, TTsrow, TTmcol, TTncol, TTmargin, TTscrsiz ;

extern unsigned long *colTable ;

extern int TTstart(void) ;
extern int TTstartStage2(void);
extern int  TTopen(void) ;
#define TTclose()

#ifdef _WINCON
extern BOOL ConsolePaint(void) ;
extern void ConsoleDrawString(char *s, WORD wAttribute, int x, int y, int len) ;
#define TTCmake(f,b) ((f) | ((b) << 4))

extern int TTend(void) ;
#define TTflush()   ((meSystemCfg & meSYSTEM_CONSOLE) ? ConsolePaint () : UpdateWindow (ttHwnd))
#else
#define TTend()
/*
 * TTflush()
 * Because of the delayed nature of the windows display perform the
 * flush by processing any outstanding PAINT messages ONLY. This
 * will update the display leaving the remaining messages on the
 * input queue.
 */
#define TTflush()   UpdateWindow (ttHwnd)
#endif

extern void TTtitleText (char *str) ;

extern void TTwaitForChar(void) ;
extern void TThideCur(void) ;
extern void TTshowCur(void) ;
#define TTNbell()  MessageBeep(0xffffffff)
extern void TTdepth(int y) ;
extern void TTwidth(int x) ;

#ifdef _CLIPBRD
extern void TTgetClipboard(void);
extern void TTsetClipboard(void);
#endif

#ifdef _CLIENTSERVER
extern int  ttServerToRead;
extern void TTopenClientServer(void) ;
extern void TTkillClientServer(void) ;
extern int  TTcheckClientServer(void) ;
extern int  TTconnectClientServer(void) ;
extern void TTsendClientServer(char *) ;
#endif


extern void TTsleep(int msec, int intable) ;
#if TYPEAH
extern int TTahead(void) ;
#endif
extern int TTaheadFlush(void);
#define TTdieTest()
#define TTbreakTest(x) ((TTbreakCnt-- ==  0) ? (TTbreakCnt=TTBREAKCNT,((x==0)?TTahead():TTaheadFlush()),TTbreakFlag):(TTbreakFlag))

extern int TTaddColor (unsigned char index, unsigned char r, 
                        unsigned char g, unsigned char b);
extern void TTsetBgcol(void);
extern void TTcolour (int fg, int bg);

extern void TTputs(int row, int col, int len) ;

/* init the rect to invalid big areas, they are LONGS so this is okay */
#define TTinitArea()        (ttRect.left=0xfffff,ttRect.right=-1,ttRect.top=0xfffff,ttRect.bottom=-1)

#define TTsetArea(r,c,h,w)                                                   \
(ttRect.left=(c),ttRect.right=(c)+(w),ttRect.top=(r),ttRect.bottom=(r)+(h))

#define TTaddArea(r,c,h,w)                                                   \
do {                                                                         \
    if(ttRect.left > (c))                                                    \
        ttRect.left = (c) ;                                                  \
    if(ttRect.right < (c)+(w))                                               \
        ttRect.right = (c)+(w) ;                                             \
    if(ttRect.top > (r))                                                     \
        ttRect.top = (r) ;                                                   \
    if(ttRect.bottom < (r)+(h))                                              \
        ttRect.bottom = (r)+(h) ;                                            \
} while(0)

extern void TTapplyArea (void) ;
extern int  WinMouseMode (int buttonMask, int highlight, int cursorShape);
extern int  WinLaunchProgram (char *cmd, int flags, char *inFile, char *outFile, int *sysRet);

#endif /* _WIN32 */

#ifdef _DOS

#define	_CONST_TIMER 1		     /* Constantly checked timer */
#define	TIMER_MIN    0		     /* The minimum timer allowed */
/* executable extension list, in reverse order of dos priority - Must have the end NULL
 * included 4dos's btm files for completeness */
#define noExecExtensions 4
extern char *execExtensions[noExecExtensions] ;

extern unsigned short TTmrow, TTmcol, TTnrow, TTsrow, TTncol, TTmargin, TTscrsiz ;

extern unsigned char *colTable ;

extern int  TTstart(void) ;
#define TTstartStage2()                 /* No stage 2 startup */
#define TTend()   TTclose()
extern int  TTopen(void) ;
extern int  TTclose(void) ;

extern void TTwaitForChar(void) ;
#define TTflush()
extern void DOShideCur(void) ;
extern void DOSshowCur(void) ;
extern void TThideCur(void) ;
extern void TTshowCur(void) ;
extern int  bdos(int func, unsigned dx, unsigned al);
#define TTNbell()   bdos(6, BELL, 0);

extern void TTsleep(int msec, int intable) ;
#if TYPEAH
extern int TTahead(void) ;
#endif
#define TTdieTest()
#define TTbreakTest(x) ((TTbreakCnt-- ==  0) && (TTbreakCnt=TTBREAKCNT,TTahead(),TTbreakFlag))

extern unsigned char Cattr ;

extern int TTaddColor (unsigned char index, unsigned char r, 
                        unsigned char g, unsigned char b);
#define TTCmake(f,b)	((f) | ((b) << 4))
#define TTsetBgcol()

#endif /* _DOS */

#if MOUSE
void TTinitMouse(void) ;
#define meCURSOR_DEFAULT   0
#define meCURSOR_ARROW     1
#define meCURSOR_IBEAM     2
#define meCURSOR_CROSSHAIR 3
#define meCURSOR_GRAB      4
#define meCURSOR_WAIT      5
#define meCURSOR_STOP      6
#define meCURSOR_COUNT     7
#endif

/* timer definitions */

/*
 * Number of timers
 */
#define AUTOS_TIMER_ID       0          /* Auto save timer handle         */
#define SLEEP_TIMER_ID       1          /* Sleep timer                    */
#define CALLB_TIMER_ID       2          /* Macro callback timer           */
#define IDLE_TIMER_ID        3          /* Idle timer handle              */
#define CURSOR_TIMER_ID      4          /* Cursor blink timer handle      */
#define MOUSE_TIMER_ID       5          /* Mouse timer handle             */
#define NUM_TIMERS           6          /* Number of timers               */

#define TIMER_SET      0x01             /* Timer is active                */
#define TIMER_EXPIRED  0x02             /* Timer has expired              */

/*
 * Timer block structure.
 */
typedef struct TIMERBLOCK
{
    struct TIMERBLOCK *next;            /* Next block to be scheduled */
    unsigned long      abstime;         /* Absolute time */
    unsigned char      id;              /* Identity of timer */
} TIMERBLOCK;

/*
 * isTimerActive
 * Returns a boolean state (0==FALSE) depending on if the timer
 * is outstanding or not.
 */
#define isTimerActive(id)  (meTimerState[id] & (TIMER_SET|TIMER_EXPIRED))
#define isTimerSet(id)     (meTimerState[id] & TIMER_SET)
#define isTimerExpired(id) (meTimerState[id] & TIMER_EXPIRED)

#define timerClearExpired(id) (meTimerState[(id)] = meTimerState[(id)] & ~TIMER_EXPIRED)

extern unsigned char meTimerState[] ;/* State of the timers, set or expired */
extern long          meTimerTime[] ; /* Absolute time of the timers */
extern TIMERBLOCK   *timers ;        /* Head of timer list             */

/*
 * timerSet
 * Start a timer, reset or continue the timer. In all eventualities
 * the timer is re-started and will run.
 *
 * id     - handle of the timer.
 * tim    - absolute time of alarm time.
 *          if (tim == 0) Absolute time is unknown but value
 *              is needed in meTimerTime.
 *          if (tim < 0)  Absolute time is unknown and value
                is not needed.
 * offset - offset from the current time to alarm.
 */
extern void timerSet (int id, long tim, long offset) ;
extern int  _timerKill (int id) ;
#define timerKill(id) (isTimerSet(id)?_timerKill(id):timerClearExpired(id))

#ifdef _MULTI_NOID_TIMER
extern void timerAlarm(int id) ;
#endif

#if (defined _CONST_TIMER) || (defined _SINGLE_TIMER)
extern void timerCheck(long tim) ;
#endif

/* flag used when idle-drop needs to be called */
#define IDLE_STATE_DROP         0x04  /* Idle timer is running */

#endif /* __TERMIOH */
