/*
 *	SCCS:		%W%		%G%		%U%
 *
 *	Last Modified :	<020699.1622>
 * 
 *****************************************************************************
 * 
 * Modifications to the original file by Jasspa. 
 * 
 * Copyright (C) 1988 - 1999, JASSPA 
 * The MicroEmacs Jasspa distribution can be copied and distributed freely for
 * any non-commercial purposes. The MicroEmacs Jasspa Distribution can only be
 * incorporated into commercial software with the expressed permission of
 * JASSPA.
 * 
 ****************************************************************************/

/*
 * The routines in this file read and write ASCII files from the disk. All of
 * the knowledge about files are here. A better message writing scheme should
 * be used.
 */

#define	__FILEIOC			/* Define program name */

#include "emain.h"

#if (defined _UNIX) || (defined _DOS) || (defined _WIN32)
#ifdef _UNIX
#include <signal.h>
#endif
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#endif

#ifdef _WIN32
#define ffpInvalidVal INVALID_HANDLE_VALUE
#else
#define ffpInvalidVal NULL
#endif

/*
 * char-mask lookup tables
 * 
 * Table 1:
 * 
 * Bits   Function        Use
 * 
 * 0x0f    getType        Used by macro language to determine the arg type
 * 0x10    isDisplayable  Used in display to determine if a character can be directly displayed
 * 0x20    isPokable      Used by screen-poke to determine if a character can be directly poked
 * 0x40    isPrintable    Used by print to determine if a character can be directly printed
 * 0x80    isSpace        Is a white char
 * 
 * 
 * Table 2:
 * 
 * Bits   Function        Use
 * 
 * 0x01    isLower        Is Lower case test
 * 0x02    isUpper        Is Upper case test
 * 0x03    isAlpha        Is Alpha test (Lower | Upper)
 * 0x04    isHexDigit     Is Hex digit char (0-9 | a-f | A-F)
 * 0x07    isAlphaNum     Is alpha numeric char (Upper|Lower|HexDigit)
 * 0x08    isSpllExt      Is Spell Extended character (e.g. '.', '-' etc)
 * 0x0f    isSpllWord     Is Spell word
 * 0xf0    isWord         Is a word letter - selected by $buffer-mask
 * 
 */

unsigned char charMaskTbl1[256] =
{
#if (defined _DOS) || (defined _WIN32)
    0x80, 0x2A, 0x2A, 0x2A, 0x2A, 0x2A, 0x2A, 0x2A, 0x2A, 0xAA, 0xAA, 0xAA, 0xAA, 0xAA, 0x2A, 0x2A, 
    0x2A, 0x2A, 0x2A, 0x2A, 0x2A, 0x2A, 0x2A, 0x2A, 0x2A, 0x2A, 0x2A, 0x2A, 0x2A, 0x2A, 0x2A, 0x2A, 
#else
    0x80, 0x0A, 0x0A, 0x0A, 0x0A, 0x0A, 0x0A, 0x0A, 0x0A, 0x8A, 0x8A, 0x8A, 0x8A, 0x8A, 0x0A, 0x0A, 
    0x0A, 0x0A, 0x0A, 0x0A, 0x0A, 0x0A, 0x0A, 0x0A, 0x0A, 0x0A, 0x0A, 0x0A, 0x0A, 0x0A, 0x0A, 0x0A, 
#endif
    0xFA, 0x76, 0x79, 0x72, 0x74, 0x73, 0x75, 0x7A, 0x7A, 0x7A, 0x77, 0x7A, 0x7A, 0x78, 0x7C, 0x7A, 
    0x78, 0x78, 0x78, 0x78, 0x78, 0x78, 0x78, 0x78, 0x78, 0x78, 0x7B, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 
    0x71, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 
    0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 
    0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 
#if (defined _DOS) || (defined _WIN32)
    0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x3A, 
    0x3A, 0x3A, 0x3A, 0x3A, 0x3A, 0x3A, 0x3A, 0x3A, 0x3A, 0x3A, 0x3A, 0x3A, 0x3A, 0x3A, 0x3A, 0x3A, 
    0x3A, 0x3A, 0x3A, 0x3A, 0x3A, 0x3A, 0x3A, 0x3A, 0x3A, 0x3A, 0x3A, 0x3A, 0x3A, 0x3A, 0x3A, 0x3A, 
#else
    0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x0A, 
    0x0A, 0x0A, 0x0A, 0x0A, 0x0A, 0x0A, 0x0A, 0x0A, 0x0A, 0x0A, 0x0A, 0x0A, 0x0A, 0x0A, 0x0A, 0x0A, 
    0x0A, 0x0A, 0x0A, 0x0A, 0x0A, 0x0A, 0x0A, 0x0A, 0x0A, 0x0A, 0x0A, 0x0A, 0x0A, 0x0A, 0x0A, 0x0A, 
#endif
    0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 
    0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 
    0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 
    0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 
    0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 
    0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A, 0x7A
};

unsigned char charMaskTbl2[256] =
{
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x08, 0x00, 0x00, 0x00, 0x00, 0x00, 0x08, 0x08, 0x00, 
    0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    0x00, 0x06, 0x06, 0x06, 0x06, 0x06, 0x06, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 
    0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x00, 0x00, 0x00, 0x00, 0x10, 
    0x00, 0x05, 0x05, 0x05, 0x05, 0x05, 0x05, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 
    0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
};

unsigned char charCaseTbl[256] = {
    0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08, 0x09, 0x0A, 0x0B, 0x0C, 0x0D, 0x0E, 0x0F,
    0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17, 0x18, 0x19, 0x1A, 0x1B, 0x1C, 0x1D, 0x1E, 0x1F,
    0x20, 0x21, 0x22, 0x23, 0x24, 0x25, 0x26, 0x27, 0x28, 0x29, 0x2A, 0x2B, 0x2C, 0x2D, 0x2E, 0x2F,
    0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37, 0x38, 0x39, 0x3A, 0x3B, 0x3C, 0x3D, 0x3E, 0x3F,
    0x40, 'a',  'b',  'c',  'd',  'e',  'f',  'g',  'h',  'i',  'j',  'k',  'l',  'm',  'n',  'o',  
    'p',  'q',  'r',  's',  't',  'u',  'v',  'w',  'x',  'y',  'z',  0x5B, 0x5C, 0x5D, 0x5E, 0x5F,
    0x60, 'A',  'B',  'C',  'D',  'E',  'F',  'G',  'H',  'I',  'J',  'K',  'L',  'M',  'N',  'O',  
    'P',  'Q',  'R',  'S',  'T',  'U',  'V',  'W',  'X',  'Y',  'Z',  0x7B, 0x7C, 0x7D, 0x7E, 0x7F,
    0x80, 0x81, 0x82, 0x83, 0x84, 0x85, 0x86, 0x87, 0x88, 0x89, 0x8A, 0x8B, 0x8C, 0x8D, 0x8E, 0x8F,
    0x90, 0x91, 0x92, 0x93, 0x94, 0x95, 0x96, 0x97, 0x98, 0x99, 0x9A, 0x9B, 0x9C, 0x9D, 0x9E, 0x9F,
    0xA0, 0xA1, 0xA2, 0xA3, 0xA4, 0xA5, 0xA6, 0xA7, 0xA8, 0xA9, 0xAA, 0xAB, 0xAC, 0xAD, 0xAE, 0xAF,
    0xB0, 0xB1, 0xB2, 0xB3, 0xB4, 0xB5, 0xB6, 0xB7, 0xB8, 0xB9, 0xBA, 0xBB, 0xBC, 0xBD, 0xBE, 0xBF,
    0xC0, 0xC1, 0xC2, 0xC3, 0xC4, 0xC5, 0xC6, 0xC7, 0xC8, 0xC9, 0xCA, 0xCB, 0xCC, 0xCD, 0xCE, 0xCF,
    0xD0, 0xD1, 0xD2, 0xD3, 0xD4, 0xD5, 0xD6, 0xD7, 0xD8, 0xD9, 0xDA, 0xDB, 0xDC, 0xDD, 0xDE, 0xDF,
    0xE0, 0xE1, 0xE2, 0xE3, 0xE4, 0xE5, 0xE6, 0xE7, 0xE8, 0xE9, 0xEA, 0xEB, 0xEC, 0xED, 0xEE, 0xEF,
    0xF0, 0xF1, 0xF2, 0xF3, 0xF4, 0xF5, 0xF6, 0xF7, 0xF8, 0xF9, 0xFA, 0xFB, 0xFC, 0xFD, 0xFE, 0xFF
};   
unsigned char charUserLatinTbl[256] = {
    0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08, 0x09, 0x0A, 0x0B, 0x0C, 0x0D, 0x0E, 0x0F,
    0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17, 0x18, 0x19, 0x1A, 0x1B, 0x1C, 0x1D, 0x1E, 0x1F,
    0x20, 0x21, 0x22, 0x23, 0x24, 0x25, 0x26, 0x27, 0x28, 0x29, 0x2A, 0x2B, 0x2C, 0x2D, 0x2E, 0x2F,
    0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37, 0x38, 0x39, 0x3A, 0x3B, 0x3C, 0x3D, 0x3E, 0x3F,
    0x40, 0x41, 0x42, 0x43, 0x44, 0x45, 0x46, 0x47, 0x48, 0x49, 0x4A, 0x4B, 0x4C, 0x4D, 0x4E, 0x4F,
    0x50, 0x51, 0x52, 0x53, 0x54, 0x55, 0x56, 0x57, 0x58, 0x59, 0x5A, 0x5B, 0x5C, 0x5D, 0x5E, 0x5F,
    0x60, 0x61, 0x62, 0x63, 0x64, 0x65, 0x66, 0x67, 0x68, 0x69, 0x6A, 0x6B, 0x6C, 0x6D, 0x6E, 0x6F,
    0x70, 0x71, 0x72, 0x73, 0x74, 0x75, 0x76, 0x77, 0x78, 0x79, 0x7A, 0x7B, 0x7C, 0x7D, 0x7E, 0x7F,
    0x80, 0x81, 0x82, 0x83, 0x84, 0x85, 0x86, 0x87, 0x88, 0x89, 0x8A, 0x8B, 0x8C, 0x8D, 0x8E, 0x8F,
    0x90, 0x91, 0x92, 0x93, 0x94, 0x95, 0x96, 0x97, 0x98, 0x99, 0x9A, 0x9B, 0x9C, 0x9D, 0x9E, 0x9F,
    0xA0, 0xA1, 0xA2, 0xA3, 0xA4, 0xA5, 0xA6, 0xA7, 0xA8, 0xA9, 0xAA, 0xAB, 0xAC, 0xAD, 0xAE, 0xAF,
    0xB0, 0xB1, 0xB2, 0xB3, 0xB4, 0xB5, 0xB6, 0xB7, 0xB8, 0xB9, 0xBA, 0xBB, 0xBC, 0xBD, 0xBE, 0xBF,
    0xC0, 0xC1, 0xC2, 0xC3, 0xC4, 0xC5, 0xC6, 0xC7, 0xC8, 0xC9, 0xCA, 0xCB, 0xCC, 0xCD, 0xCE, 0xCF,
    0xD0, 0xD1, 0xD2, 0xD3, 0xD4, 0xD5, 0xD6, 0xD7, 0xD8, 0xD9, 0xDA, 0xDB, 0xDC, 0xDD, 0xDE, 0xDF,
    0xE0, 0xE1, 0xE2, 0xE3, 0xE4, 0xE5, 0xE6, 0xE7, 0xE8, 0xE9, 0xEA, 0xEB, 0xEC, 0xED, 0xEE, 0xEF,
    0xF0, 0xF1, 0xF2, 0xF3, 0xF4, 0xF5, 0xF6, 0xF7, 0xF8, 0xF9, 0xFA, 0xFB, 0xFC, 0xFD, 0xFE, 0xFF
}; 
unsigned char charLatinUserTbl[256] = {
    0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08, 0x09, 0x0A, 0x0B, 0x0C, 0x0D, 0x0E, 0x0F,
    0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17, 0x18, 0x19, 0x1A, 0x1B, 0x1C, 0x1D, 0x1E, 0x1F,
    0x20, 0x21, 0x22, 0x23, 0x24, 0x25, 0x26, 0x27, 0x28, 0x29, 0x2A, 0x2B, 0x2C, 0x2D, 0x2E, 0x2F,
    0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37, 0x38, 0x39, 0x3A, 0x3B, 0x3C, 0x3D, 0x3E, 0x3F,
    0x40, 0x41, 0x42, 0x43, 0x44, 0x45, 0x46, 0x47, 0x48, 0x49, 0x4A, 0x4B, 0x4C, 0x4D, 0x4E, 0x4F,
    0x50, 0x51, 0x52, 0x53, 0x54, 0x55, 0x56, 0x57, 0x58, 0x59, 0x5A, 0x5B, 0x5C, 0x5D, 0x5E, 0x5F,
    0x60, 0x61, 0x62, 0x63, 0x64, 0x65, 0x66, 0x67, 0x68, 0x69, 0x6A, 0x6B, 0x6C, 0x6D, 0x6E, 0x6F,
    0x70, 0x71, 0x72, 0x73, 0x74, 0x75, 0x76, 0x77, 0x78, 0x79, 0x7A, 0x7B, 0x7C, 0x7D, 0x7E, 0x7F,
    0x80, 0x81, 0x82, 0x83, 0x84, 0x85, 0x86, 0x87, 0x88, 0x89, 0x8A, 0x8B, 0x8C, 0x8D, 0x8E, 0x8F,
    0x90, 0x91, 0x92, 0x93, 0x94, 0x95, 0x96, 0x97, 0x98, 0x99, 0x9A, 0x9B, 0x9C, 0x9D, 0x9E, 0x9F,
    0xA0, 0xA1, 0xA2, 0xA3, 0xA4, 0xA5, 0xA6, 0xA7, 0xA8, 0xA9, 0xAA, 0xAB, 0xAC, 0xAD, 0xAE, 0xAF,
    0xB0, 0xB1, 0xB2, 0xB3, 0xB4, 0xB5, 0xB6, 0xB7, 0xB8, 0xB9, 0xBA, 0xBB, 0xBC, 0xBD, 0xBE, 0xBF,
    0xC0, 0xC1, 0xC2, 0xC3, 0xC4, 0xC5, 0xC6, 0xC7, 0xC8, 0xC9, 0xCA, 0xCB, 0xCC, 0xCD, 0xCE, 0xCF,
    0xD0, 0xD1, 0xD2, 0xD3, 0xD4, 0xD5, 0xD6, 0xD7, 0xD8, 0xD9, 0xDA, 0xDB, 0xDC, 0xDD, 0xDE, 0xDF,
    0xE0, 0xE1, 0xE2, 0xE3, 0xE4, 0xE5, 0xE6, 0xE7, 0xE8, 0xE9, 0xEA, 0xEB, 0xEC, 0xED, 0xEE, 0xEF,
    0xF0, 0xF1, 0xF2, 0xF3, 0xF4, 0xF5, 0xF6, 0xF7, 0xF8, 0xF9, 0xFA, 0xFB, 0xFC, 0xFD, 0xFE, 0xFF
}; 
unsigned char charMaskFlags[]="luhs1234dpPwaAMLU" ;
unsigned char isWordMask=CHRMSK_DEFWORDMSK;


static int               ffremain ;
       unsigned char    *ffbuf=NULL ;
static unsigned char    *ffcur ;
static unsigned char     ffbinary=0 ;
static unsigned char     ffcrypt=0 ;
static unsigned char     ffauto=0 ;
static unsigned char     ffautoRet=0 ;
static unsigned char     ffnewFile ;
static int               ffread ;

#define AUTO_CRLF  0x01
#define AUTO_CTRLZ 0x02

#ifdef _URLSUPP
#ifdef _WIN32
/* winsock2.h must be included before */
#include <winsock2.h>
#include <io.h>
typedef void (*meATEXIT)(void) ;
static SOCKET ffsock ;
#define meOpenSocket(a,b,c)  WSASocket(a,b,c,NULL,0,0)
#define meCloseSocket closesocket
#else
#include <ctype.h>
#include <sys/socket.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <sys/un.h>
#include <netinet/in.h>
static int ffsock ;
#define meOpenSocket  socket
#define meCloseSocket close
#endif

static struct sockaddr_in meSockAddr ;
static int ffsize ;
#define URL_FTP  0
#define URL_HTTP 1

#ifndef INADDR_NONE
/* This may not be defined, particularly suns */
#define INADDR_NONE -1
#endif


/* `hostname' can be of any form described in inet_addr(3);
 * `port' can be a service name, a port number, or NULL to have the
 * OS choose a port (this is probably only useful for servers);
 * `proto' can be "udp" or "tcp"
 */
static int
make_inet_addr(char *hostname, char *port, char *proto)
{
    struct hostent *hp;
    struct servent *servent;
    
    meSockAddr.sin_family = AF_INET ;
    
    /* First resolve the hostname, then resolve the port */
    if ((meSockAddr.sin_addr.s_addr = inet_addr(hostname)) != INADDR_NONE)
    {
        /* in_addr.s_addr is already set */
    }
    else if(((hp = gethostbyname(hostname)) != NULL) &&
            (hp->h_length <= sizeof(struct in_addr)) &&
            (hp->h_addrtype == AF_INET))
        memcpy(&(meSockAddr.sin_addr.s_addr),hp->h_addr,sizeof(meSockAddr.sin_addr.s_addr));
    else
        return mlwrite(MWABORT|MWPAUSE,"[Unknown host %s]", hostname) ;
    
    if ((servent = getservbyname(port, proto)) != NULL)
    {
        meSockAddr.sin_port = servent->s_port;
    }
    else if ((meSockAddr.sin_port = htons(atoi(port))) == 0)
        return mlwrite(MWABORT|MWPAUSE,"[Bad port number %s]", port) ;
    
    return TRUE ;
}

void
strBase64Encode3(char *dd, char c1, char c2, char c3)
{
    static char base64Table[]="ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/" ;
    dd[0] = base64Table[(c1 >> 2)] ;
    dd[1] = base64Table[((c1 << 4) & 0x30) | ((c2 >> 4) & 0x0f)] ;
    dd[2] = base64Table[((c2 << 2) & 0x3c) | ((c3 >> 6) & 0x03)] ;
    dd[3] = base64Table[(c3 & 0x3f)] ;
}

char *
strBase64Encode(char *dd, char *ss)
{
    char c1, c2, c3 ;
    
    while((c1=*ss++) != '\0')
    {
        if((c2=*ss++) == '\0')
        {
            c3 = '\0' ;
            strBase64Encode3(dd,c1,c2,c3) ;
            dd += 2 ;
            *dd++ = '=' ;
            *dd++ = '=' ;
            break ;
        }
        else if((c3=*ss++) == '\0')
        {
            strBase64Encode3(dd,c1,c2,c3) ;
            dd += 3 ;
            *dd++ = '=' ;
            break ;
        }
        strBase64Encode3(dd,c1,c2,c3) ;
        dd += 4 ;
    }
    *dd = '\0' ;
    return dd ;
}


int
openURL(char *urlName, char *lgnPswd)
{
    char buff[MAXBUF], *addr, *port, *ss, *dd, cc ;
    int  ii ;
    
#ifdef _WIN32
    static int init=0 ;
    if(!init)
    {
        WSADATA wsaData;
        WORD    wVersionRequested;
        
        wVersionRequested = MAKEWORD (1, 1);
    
        if(WSAStartup(wVersionRequested, &wsaData))
            return mlwrite(MWABORT|MWPAUSE,"[Failed to initialise sockets]") ;
        atexit((meATEXIT) WSACleanup) ;
        init = 1 ;
    }
#endif
    ss = urlName ;
    /* skip the http: or ftp: */
    if(ss[0] == 'f')
        ss += 6 ;
    else
        ss += 7 ;
    dd = buff ;
    addr = dd ;
    port = NULL ;
    while(((cc=*ss++) != '\0') && (cc != DIR_CHAR))
    {
        *dd++ = cc ;
        if(cc == ':')
            port = dd ;
        else if(cc == '@')
        {
            lgnPswd = buff ;
            dd[-1] = '\0' ;
            addr = dd ;
            port = NULL ;
        }
    }
    *dd = '\0' ;
    ss-- ;
    if(port == NULL)
        port = "80" ;
    else
        port[-1] = '\0' ;
    
    mlwrite(MWCURSOR,"[Connecting to %s]",addr);
    
    if((ffsock=meOpenSocket(AF_INET,SOCK_STREAM,0)) < 0)
        return mlwrite(MWABORT|MWPAUSE,"[Failed to open socket]") ;
    
    ffp = ffpInvalidVal ;
    ii = 1 ;
    setsockopt (ffsock,SOL_SOCKET,SO_KEEPALIVE,(char *) &ii,sizeof(int));
    
    if(make_inet_addr(addr,port,"tcp") != TRUE)
    {
        meCloseSocket(ffsock) ;
        return ABORT ;
    }
    if(connect(ffsock,(struct sockaddr *) &meSockAddr,sizeof(struct sockaddr_in)) < 0)
    {
        meCloseSocket(ffsock) ;
        return mlwrite(MWABORT|MWPAUSE,"[Failed to connect to %s:%s]",addr,port) ;
    }
    
    mlwrite(MWCURSOR,"[Connected, fetching %s]", ss);

    
    /* send GET message to http */
/*    sprintf(buff,"GET %s",ss) ;*/
    {
        char tbuff[MAXBUF+200] ;
        /*        sprintf(tbuff,"GET %s HTTP/1.0\r\nConnection: Keep-Alive\r\nUser-Agent: Mozilla\r\nHost: %s\r\nAccept: \*\/\*\r\n\r\n",ss,addr) ;*/
        dd = tbuff ;
        dd += sprintf(dd,"GET %s HTTP/1.0\r\nConnection: Keep-Alive\r\nHost: %s",ss,addr) ;
        if(lgnPswd != NULL)
        {
            /* password supplied, encode */
            strcpy(dd,"\r\nAuthorization: Basic ") ;
            dd += 23 ;
            dd = strBase64Encode(dd,lgnPswd) ;
        }
        strcpy(dd,"\r\n\r\n") ;
        if(send(ffsock,tbuff,strlen(tbuff),0) <= 0)
        {
            meCloseSocket(ffsock) ;
            return FALSE ;
        }
    }
    ffsize = 0x7fffffff ;
    /* must now ditch the header, read up to the first blank line */
    {
        for(ii=0;;)
        {
            if(ffremain == 0)
            {
                ffremain = recv(ffsock,ffbuf,FIOBUFSIZ,0) ;
                if(ffremain <= 0)
                {
                    ffremain = 0 ;
                    break ;
                }
                /* ffbuf[ffremain]='\0' ;*/
                /* printf("Got Header: [%s]\n",ffbuf) ;*/
                ffbuf[ffremain]='\n' ;
                ffcur = ffbuf ;
            }
            ffremain-- ;
            cc = *ffcur++ ;
            if(cc == '\n')
            {
                if(ii == 0)
                {
                    ffread = ffremain ;
                    return TRUE ;
                }
                ii = 0 ;
                /* this assumes that all the Centent-Length is in this buffer
                 * this only fails if the header is longer that 2K - not likely */
                if(!strncmp((char *) ffcur,"Content-Length:",15))
                    ffsize = meAtoi((char *) ffcur+15) ;
                else if(!strncmp((char *) ffcur,"Location:",9))
                {
                    /* The requested file is not here, its at the given location */
                    ss = (char *) ffcur+9 ;
                    while(((cc=*ss) == ' ') || (cc == '\t'))
                        ss++ ;
                    dd = ss ;
                    while(((cc=*dd) != '\r') && (cc != '\n'))
                        dd++ ;
                    if(ss != dd)
                    {
                        *dd = '\0' ;
                        /* printf("Got Location: [%s]\n",ss) ;*/
                        meCloseSocket(ffsock) ;
                        ffremain = 0 ;
                        return openURL(ss,lgnPswd) ;
                    }
                }
            }
            else if(cc != '\r')
                ii++ ;
        }
    }
    meCloseSocket(ffsock) ;
    return mlwrite(MWABORT|MWPAUSE,"[Failed to read header of %s]",ss) ;
}
#endif

/*
 * creates the standard bckup name with the extra
 * character(s) being backl, where a '~' is used
 * for backups and '#' for auto-saves
 * returns 1 if fails, else 0 and name in filename buffer
 */
int
createBackupName(char *filename, char *fn, char backl)
{
    char *s, *t ;
    
    if(fn == NULL)
        return 1 ;
    strcpy(filename,fn) ;
    t = filename + strlen(filename) ;
#ifndef _DOS
    if(meSystemCfg & meSYSTEM_DOSFNAMES)
#endif
    {
        s = t ;
        while(--s != filename)
            if((*s == '.') || (*s == DIR_CHAR))
                break ;
        if(*s != '.')
        {
            *t = '.' ;
            t[1] = backl ;
            t[2] = backl ;
            t[3] = backl ;
            t[4] = '\0' ;
        }
#ifndef _DOS
        /* In the Win32 environment. If the extension is longer
         * than 3 characters then replace the last character with
         * the backup character. Otherwise behave like a standard
         * DOS name.
         */
        else if ((t - s) > 3)
            t[-1] = backl;
#endif    
        else
        {
            t = s ;
            if(t[2] == '\0')
                t[2] = backl ;
            t[3] = backl ;
            t[4] = '\0' ;
        }
        if(!strcmp(filename,fn))
            return 1 ;
    }
#ifndef _DOS
    else
    {
        *t++ = backl ;
        *t   = 0 ;
    }
#endif
    return 0 ;
}


/*
 * Read a line from a file and create a new line to hold it.
 * If mode & MDBINARY then loading it in binary mode an a line
 * is upto 16 bytes long.
 * 
 * Returns ABORT if a serious system problem arose (read or malloc
 * failure).
 * Returns Error is nothing was read cos at the end of the file
 * Else success so returns TRUE.
 */
static int	
ffgetline(LINE **line)
{
    unsigned char  buf[16] ;
    register unsigned char *text ;
    register unsigned char *endp ;
    register int            len=0, newl, ii ;
    register LINE          *lp ;
    
    do {
        if(ffremain <= 0)
        {
#ifdef _URLSUPP
            if(ffp == ffpInvalidVal)
            {
                if((ffremain = ffsize-ffread) > 0)
                {
                    if(ffremain > FIOBUFSIZ)
                        ffremain = FIOBUFSIZ ;
                    ffremain = recv(ffsock,ffbuf,ffremain,0) ;
                }
                if(ffremain <= 0)
                {
                    ffremain = -1 ;
                    break ;
                }
            }
            else
#endif
#ifdef _WIN32
            {
                if(ReadFile(ffp,ffbuf,FIOBUFSIZ,&ffremain,NULL) == 0)
                    return mlwrite(MWABORT,"File read error %d",GetLastError());
                
                if(ffremain <= 0)
                {
                    ffremain = -1 ;
                    break ;
                }
            }
#else
            {
                ffremain = fread(ffbuf,1,FIOBUFSIZ,ffp) ;
                if(ffremain <= 0)
                {
                    if(ferror(ffp))
                        return mlwrite(MWABORT,"File read error %d",errno);
                    ffremain = -1 ;
                    break ;
                }
            }
#endif
#if	CRYPT
            if(ffcrypt)
                meCrypt(ffbuf, ffremain);
#endif
            ffbuf[ffremain]='\n' ;
            ffread += ffremain ;
            ffcur = ffbuf ;
        }
        if(ffbinary)
        {
            if(ffremain < (16-len))
            {
                memcpy(buf+len,ffcur,ffremain) ;
                len += ffremain ;
                ffremain = -1 ;
            }
            else
            {
                ii = 16-len ;
                memcpy(buf+len,ffcur,ii) ;
                ffremain -= ii ;
                ffcur += ii ;
                len = 16 ;
            }
        }
        else
        {
            register unsigned char cc ;
            endp = ffcur ;
            newl = 0 ;
            while(((cc=*ffcur++) != '\n') && (cc != '\r'))
                newl++ ;
            /* If we ended in '\r' check for a '\n' */
            if(cc == '\r')
            {
                if(*ffcur == '\n')
                {
                    ffremain -= newl+2 ;
#ifdef _CTRLM
                    if(ffauto)
                        ffautoRet |= AUTO_CRLF ;
                    ffcur[-1] = '\n' ;
#else
                    if(ffauto)
                    {
                        ffautoRet |= AUTO_CRLF ;
                        ffcur[-1] = '\n' ;
                    }
                    else
                        newl++;
#endif
                    ffcur++ ;
                }
                else
                {
                    ffremain -= newl+1 ;
                    ffcur[-1] = '\n' ;
                }
            }
            else
                ffremain -= newl+1 ;
            
            if(len == 0)
            {
                len = newl ;
                if((lp = (LINE *) meMalloc(sizeof(LINE)+len)) == NULL)
                    return ABORT ;
                lp->l_flag = 0 ;
                text = lp->l_text ;
            }
            else
            {
                ii = len ;
                len += newl ;
                if((lp = (LINE *) meRealloc(lp,sizeof(LINE)+len)) == NULL)
                    return ABORT ;
                text = lp->l_text + ii ;
            }
            while((cc = *endp++) != '\n')
            {
                if(cc == '\0')
                    len-- ;
                else
                    *text++ = cc ;
            }
            *text = '\0' ;
        }
    } while(ffremain < 0) ;
    
    if(ffbinary)
    {
        long cpos ;
        
        if(len == 0)
            return FALSE ;
        if((lp = (LINE *) meMalloc(sizeof(LINE)+78)) == NULL)
            return ABORT ;
        lp->l_flag = 0 ;
        text = lp->l_text ;
        
        cpos = (ffread - ffremain - 1) & ~(0x0fL) ;
        
        for(newl=7 ; newl>=0 ; newl--)
        {
            text[newl] = hexdigits[cpos&0x0f] ;
            cpos >>= 4 ;
        }
        text[8] = ':' ;
        text[9] = ' ' ;
                  
        for(newl=0 ; newl<16 ; newl++)
        {
            if(newl < len)
            {
                unsigned char cc = buf[newl] ;
                text[(newl*3)+10] = hexdigits[cc/0x10] ;
                text[(newl*3)+11] = hexdigits[cc%0x10] ;
                text[(newl*3)+12] = ' ' ;
                if(isDisplayable(cc))
                    text[newl+62] = cc ;
                else
                    text[newl+62] = '.' ;
            }
            else
            {
                text[(newl*3)+10] = ' ' ;
                text[(newl*3)+11] = ' ' ;
                text[(newl*3)+12] = ' ' ;
                text[62+newl] = ' ' ;
            }
        }
        text[58] = ' ' ;
        text[59] = '|' ;
        text[60] = ' ' ;
        text[61] = ' ' ;
        text[78] = '\0' ;
        lp->l_size = 78 ;
        lp->l_used = 78 ;
        lp->l_flag = 0 ;			/* line is not marked.	*/
        *line = lp ;
        return TRUE ;
    }
    if(ffremain < 0)  /* had to break */
    {
        if(len == 0)
            return FALSE ;
        if(ffauto && (len == 1) && (lp->l_text[0] == 26))
        {
            ffautoRet |= AUTO_CTRLZ ;
            meFree(lp) ;
            return FALSE ;
        }
        /* Don't flag as the end, otherwise we won't get a proper linep */
        /* mlwrite(MWCURSOR|MWPAUSE,"Warning - Incomplete last line");*/
            
        lp->l_flag = LNNEOL ;				/* line is not marked.	*/
    }
    lp->l_size = len ;
    lp->l_used = len ;
    *line = lp ;

    return TRUE ;
}

int
ffReadFile(char *fname, int flags, BUFFER *bp, LINE *hlp)
{
    register LINE   *lp0 ;
    LINE            *lp1 ;
    register LINE   *lp2 ;
    register int     ss ;
    register int     nline ;
    
    if(bp != NULL)
    {
        ffbinary = (meModeTest(bp->b_mode,MDBINRY) != 0) ;
        ffcrypt  = (meModeTest(bp->b_mode,MDCRYPT) != 0) ;
        ffauto   = (meModeTest(bp->b_mode,MDAUTO) != 0) ;
    }
    else
    {
        ffbinary = 0 ;
        ffcrypt = 0 ;
        ffauto = 1 ;
    }        
    ffautoRet= 0 ;
    ffremain = 0 ;
    ffread = 0 ;

    if(fname != NULL)
    {
#if URLAWAR
        if(isUrlLink(fname))
        {
#ifdef _URLSUPP
#ifdef _UNIX
            /* must stop any signals or the connection will fail */
            meSigHold() ;
#endif
            if(openURL(fname,NULL) != TRUE)
            {
#ifdef _UNIX
                meSigRelease() ;
#endif
                return ABORT ;
            }
            /* cannot return here, must do ffrsetup */
#else
            return mlwrite(MWABORT|MWPAUSE,"[No url support on this platform]") ;
#endif
        }        
        else
#endif
        {
#ifdef _WIN32
            if((ffp=CreateFile(fname,GENERIC_READ,FILE_SHARE_READ|FILE_SHARE_WRITE,NULL,OPEN_EXISTING,
                               FILE_ATTRIBUTE_NORMAL,NULL)) == INVALID_HANDLE_VALUE)
#else
            if ((ffp=fopen(fname, "rb")) == NULL)
#endif
            {
                if(!(flags & READ_SILENT))
                    /* File not found.      */
                    mlwrite(MWCLEXEC,"[%s: No such file]", fname);
                return ABORT ;
            }
        }
    }
#ifdef _UNIX
#ifdef _URLSUPP
    if(ffp != ffpInvalidVal)
#endif
        meSigHold() ;
#endif

    nline = 0;
    lp2 = hlp ;	        /* line after  insert */
    lp0 = lp2->l_bp ;	/* line before insert */
    while((ss=ffgetline(&lp1)) == TRUE)
    {
        /* re-link new line between lp0 and lp1 */
        lp0->l_fp = lp1;
        lp1->l_bp = lp0;
        lp0 = lp1 ;
        nline++ ;
        if(TTbreakTest(1))
        {
            ss = ctrlg(FALSE,1) ;
            break ;
        }
    }
    /* complete the link */
    lp2->l_bp = lp0;
    lp0->l_fp = lp2;

    /* Windows pipeCommand comes in as a pipe because the source file
     * has already been opened, but from now one treat it as a normal
     * file.
     * The output file has been set up as a temporary file with the
     * 'delete when closed' flag.
     */
#ifdef _URLSUPP
    if(ffp == ffpInvalidVal)
    {
        meCloseSocket(ffsock) ;
    }
    else
#endif
    {
#ifdef _WIN32
        if((ffp != GetStdHandle(STD_INPUT_HANDLE)) &&
           (CloseHandle(ffp) == 0) && !(flags & READ_SILENT))
            mlwrite(MWABORT|MWPAUSE,"Error closing file");
#else
#ifdef _UNIX
        if(fname == NULL)
        {
            if(ffp != stdin)
                pclose(ffp);
        }
        else if((fclose(ffp) != 0) && !(flags & READ_SILENT))
            mlwrite(MWABORT|MWPAUSE,"Error closing file");
#else
        if((ffp != stdin) && (fclose(ffp) != 0) && !(flags & READ_SILENT))
            mlwrite(MWABORT|MWPAUSE,"Error closing file");
#endif
#endif
    }
    /* Ignore any close error */
#ifdef _UNIX
    meSigRelease() ;
#endif
    if(bp != NULL)
    {
        bp->elineno += nline ;
        if(ffauto && nline)
        {
            if(ffautoRet & AUTO_CRLF)
                meModeSet(bp->b_mode,MDCRLF) ;
            else
                meModeClear(bp->b_mode,MDCRLF) ;
            if(ffautoRet & AUTO_CTRLZ)
                meModeSet(bp->b_mode,MDCTRLZ) ;
            else
                meModeClear(bp->b_mode,MDCTRLZ) ;
        }
    }
    return ss ;
}


int
ffputBuf(void)
{
#if	CRYPT
    if(ffcrypt)
        meCrypt(ffbuf,ffremain) ;
#endif
#ifdef _WIN32
    {
        int written ;
        if((WriteFile(ffp,ffbuf,ffremain,&written,NULL) == 0) ||
           (written != ffremain))
            return mlwrite(MWABORT,"Write I/O error");
    }
#else
    if(fwrite(ffbuf,1,ffremain,ffp) != ffremain)
        return mlwrite(MWABORT,"Write I/O error");
#endif
    ffremain = 0 ;
    return 0 ;
}

#define ffputc(c)                                                            \
(((ffbuf[ffremain++]=c),(ffremain == FIOBUFSIZ))? ffputBuf():0)
    
int
ffWriteFileOpen(char *fname, int flags, BUFFER *bp)
{
    if(bp != NULL)
    {
#if CRYPT
        if(resetkey(bp) != TRUE)
            return FALSE ;
#endif
    }
    if(fname == NULL)
    {
#ifdef _WIN32
        ffp = GetStdHandle(STD_OUTPUT_HANDLE) ;
#else
        ffp = stdout ;
#endif
        ffnewFile = 0 ;
    }
    else
    {
        if(!(ffnewFile=meTestExist(fname)) && (flags & WRITE_BACKUP))
        {
            char filename[FILEBUF] ;
            int ii ;
        
#ifdef _WIN32
            /* Win95 first version has nasty bug in MoveFile, doing
             * 
             * MoveFile("a.bbb","a.bbb~")
             * CreateFile("a.bbb",..)
             * 
             * Resulted in a.bbb backing up to a.bbb~ and then for some unknown bloody
             * reason opening a.bbb opens a.bbb~ instead of creating a new file.
             * 
             * So alternatively do:
             *
             * MoveFile("a.bbb","a.bbb~.~_~")
             * MoveFile("a.bbb~.~_~","a.bbb~")
             * CreateFile("a.bbb",..)
             * 
             * Which sadly works (we love you bill!)
             * 
             * This is done for all flavours of windows as these drives can be networked
             */
            extern int platformId;
            char filenameOldB[FILEBUF], *filenameOld ;
            
            if(!(meSystemCfg & meSYSTEM_DOSFNAMES))
            {
                strcpy(filenameOldB,fname) ;
                strcat(filenameOldB,".~a~") ;
                filenameOld = filenameOldB ;
                if(!meTestExist(filenameOld) && meUnlink(filenameOld))
                    mlwrite(MWABORT|MWPAUSE,"[Unable to remove backup file %s]", filenameOld) ;
                else if(meRename(fname,filenameOld))
                    mlwrite(MWABORT|MWPAUSE,"[Unable to backup file to %s]",filenameOld) ;
            }
            else
                filenameOld = fname ;
#else
#define filenameOld fname
#endif
#ifndef _DOS
            if(!(meSystemCfg & meSYSTEM_DOSFNAMES) && (keptVersions > 0))
            {
                int ll ;
                
                strcpy(filename,fname) ;
                ll = strlen(filename) ;
                filename[ll++] = '.' ;
                filename[ll++] = '~' ;
                for(ii=0 ; ii<keptVersions ; ii++)
                {
                    sprintf(filename+ll,"%d~",ii) ;
                        if(meTestExist(filename))
                            break ;
                }
                if(ii == keptVersions)
                {
                    if(meUnlink(filename))
                        mlwrite(MWABORT|MWPAUSE,"[Unable to remove backup file %s]", filename) ;
                    ii-- ;
                }
                if(ii)
                {
                    char filename2[FILEBUF] ;
                    strcpy(filename2,filename) ;
                    while(ii > 0)
                    {
                        sprintf(filename2+ll,"%d~",ii--) ;
                        sprintf(filename+ll,"%d~",ii) ;
                        if(meRename(filename,filename2))
                        {
                            mlwrite(MWABORT|MWPAUSE,"[Unable to backup file to %s (%d - %s)]", 
                                    filename2,errno,sys_errlist[errno]) ;
                            if(meUnlink(filename))
                            {
                                mlwrite(MWABORT|MWPAUSE,"[Unable to remove backup file %s]", filename) ;
                                break ;
                            }
                        }
                    }
                }
            }
            else
#endif
                ii = createBackupName(filename,fname,'~') ;

            if(!ii)
            {
                if(!meTestExist(filename) && meUnlink(filename))
                    mlwrite(MWABORT|MWPAUSE,"[Unable to remove backup file %s]", filename) ;
                else if(meRename(filenameOld,filename))
                    mlwrite(MWABORT|MWPAUSE,"[Unable to backup file to %s (%d - %s)]", 
                            filename,errno,sys_errlist[errno]) ;
                else if(flags & 0xffff)
                    meChmod(filename,(flags & 0xffff)) ;
            }
        }
#ifdef _WIN32
        {
            DWORD create ;
            if(flags & WRITE_OPENTRUNC)
                create = TRUNCATE_EXISTING ;
            else if(flags & WRITE_OPENEND)
                create = OPEN_ALWAYS ;
            else
                create = CREATE_ALWAYS ;
            /* Windows must open the file with the correct permissions to support the
             * compress attribute
             */
            if((ffp=CreateFile(fname,GENERIC_WRITE,FILE_SHARE_READ,NULL,create,
                               ((bp == NULL) ? meUmask:bp->stats.stmode),
                               NULL)) == INVALID_HANDLE_VALUE)
                return mlwrite(MWABORT,"Cannot open file [%s] for writing",fname);
            if(flags & WRITE_OPENEND)
                SetFilePointer(ffp,0,NULL,FILE_END) ;
        }
#else
        {
            char *ss ;
            if(flags & WRITE_OPENEND)
                ss = "ab" ;
            else
                ss = "wb" ;
            if((ffp=fopen(fname,ss)) == NULL)
                return mlwrite(MWABORT,"Cannot open file [%s] for writing",fname);
        }
#endif
        mlwrite(MWCURSOR|MWCLEXEC,"[%sriting %s]",
                (flags & WRITE_AUTOSAVE) ? "Auto-w":"W",fname);	/* tell us were writing */
    }
    ffremain = 0 ;
    if(bp != NULL)
    {
        ffbinary = (meModeTest(bp->b_mode,MDBINRY) != 0) ;
        ffcrypt  = (meModeTest(bp->b_mode,MDCRYPT) != 0) ;
        if(!ffbinary &&
#ifdef _CTRLZ
           (!meModeTest(bp->b_mode,MDAUTO) || meModeTest(bp->b_mode,MDCTRLZ))
#else
           (meModeTest(bp->b_mode,MDAUTO) && meModeTest(bp->b_mode,MDCTRLZ))
#endif
           )
            ffauto = 1 ;
        else
            ffauto = 0 ;
    
        if(!ffbinary &&
#ifdef _CTRLM
           (!meModeTest(bp->b_mode,MDAUTO) || meModeTest(bp->b_mode,MDCRLF))
#else
           (meModeTest(bp->b_mode,MDAUTO) && meModeTest(bp->b_mode,MDCRLF))
#endif
           )
            ffautoRet = 1 ;
        else
            ffautoRet = 0 ;
#if NARROW
        if((bp->narrow != NULL) && !(flags & WRITE_IGNRNRRW))
            unnarrowBuffer(bp) ;
#endif
#if	TIMSTMP
        if(!(flags & WRITE_AUTOSAVE))
            set_timestamp(bp);			/* Perform time stamping */
#endif
    }
    else
    {
        ffbinary = 0 ;
        ffcrypt  = 0 ;
        ffauto   = 0 ;
        ffautoRet= 0 ;
    }
#ifdef _UNIX
    meSigHold() ;
#endif

#if (defined _DOS) || (defined _WIN32)
    /* the directory time stamps on dos are not updated so the best
     * we can do is flag that the dirLists are out of date when WE
     * create a new file. Obviously if something else creates a file
     * then that file many be missed in hte completion lists etc
     */
    if(ffnewFile == 1)
    {
        extern meDIRLIST fileNames ;

        curDirList.timeStamp = 1 ;
        fileNames.timeStamp = 1 ;
    }
#endif
    return TRUE ;
}


int
ffWriteFileWrite(register int len, register unsigned char *buff, int eolFlag)
{
    unsigned char cc, dd ;
    
    if(ffbinary)
    {
        while(--len >= 0)
            if(*buff++ == ':')
                break ;
        for(;;)
        {
            while((--len >= 0) && ((cc = *buff++) == ' '))
                ;
            if((len < 0) || (cc == '|'))
                break ;
            if(!isHexDigit(cc) || (--len < 0) ||
               ((dd = *buff++),!isHexDigit(dd)))
                return mlwrite(MWABORT|MWPAUSE,"Binary format error");
            cc = (hexToNum(cc)<<4) | hexToNum(dd) ;
            if(ffputc(cc))
                return ABORT ;
        }
    }
    else
    {
        while(--len >= 0)
        {
            cc=*buff++ ;
            if(ffputc(cc))
                return ABORT ;
        }
        if(eolFlag)
        {
            if(ffautoRet && ffputc('\r'))
                return ABORT ;
            if(ffputc('\n'))
                return ABORT ;
        }
    }
    return TRUE ;
}

int
ffWriteFileClose(int flags, BUFFER *bp)
{
    int ret ;
    
    /* add a ^Z at the end of the file if needed */
    if((ffauto && ffputc(26)) || ffputBuf())
        ret = ABORT ;
#ifdef _WIN32
    else if((ffp != GetStdHandle(STD_OUTPUT_HANDLE)) && (CloseHandle(ffp) == 0))
        ret = mlwrite(MWABORT|MWPAUSE,"Error closing file");
#else
    else if((ffp != stdout) && (fclose(ffp) != 0))
        ret = mlwrite(MWABORT|MWPAUSE,"Error closing file");
#endif
    else
        ret = TRUE ;
#ifdef _UNIX
    meSigRelease() ;
#endif

#if NARROW
    if((bp != NULL) && (bp->narrow != NULL) && !(flags & WRITE_IGNRNRRW))
        redoNarrowInfo(bp) ;
#endif
    return ret ;
}


int
ffWriteFile(char *fname, int flags, BUFFER *bp)
{
    register int   ss ;
    
    if((ss=ffWriteFileOpen(fname,flags,bp)) == TRUE)
    {
        register LINE *lp ;
        long noLines ;
        
        /* store the number of lines now cos the close may narrow the buffer
         * and the number of lines will be wrong */
        noLines = bp->elineno+1 ;
        lp = lforw(bp->b_linep);            /* First line.          */
        while((lp != bp->b_linep) &&
              ((ss=ffWriteFileWrite(llength(lp),ltext(lp),
                                    !(lp->l_flag & LNNEOL))) == TRUE))
            lp = lforw(lp) ;
        ffWriteFileClose(flags,bp) ;
        if(ss == TRUE)
        {
            if(ffnewFile)
                mlwrite(MWCLEXEC,"[New file %s, Wrote %d lines]",fname,noLines);
            else
                mlwrite(MWCLEXEC,"[Wrote %d lines]",noLines);
            return TRUE ;
        }
    }
    return FALSE ;
}


