/*
-- ----------------------------------------------------------------------------
--
--    			Copyright 1999 Jasspa
--
--  System       : MicroEmacs Jasspa Distribution
--  Module       : Macro Handling
--  Object Name  : $RCSfile$
--  Revision     : $Revision$
--  Created      : 1995
--  Author       : Steve Philips
--  Last Modified: <990702.1147> 
--
--  Description	
--
--  Notes
--
--  History
--	
--
-- ----------------------------------------------------------------------------
--   
-- Copyright (C) 1995 - 1999, JASSPA 
-- 
-- The MicroEmacs Jasspa distribution can be copied and distributed freely for
-- any non-commercial purposes. The MicroEmacs Jasspa Distribution can only be
-- incorporated into commercial software with the expressed permission of
-- JASSPA.
-- 
-- ----------------------------------------------------------------------------
*/
#include "emain.h"
#include "efunc.h"

/*
 * Begin a keyboard macro.
 * Error if not at the top level in keyboard processing. Set up variables and
 * return.
 */
int
startKbdMacro(int f, int n)
{
    if (kbdmode != STOP)
    {
        mlwrite(0,"Macro already active");
        return FALSE ;
    }
    mlwrite(0,"[Start macro]");
    kbdptr = &lkbdptr[0];
    kbdlen = 0 ;
    kbdmode = RECORD;
    addModeToWindows(WFMODE) ;  /* and update ALL mode lines */
    return TRUE ;
}

/*
 * End keyboard macro. Check for the same limit conditions as the above
 * routine. Set up the variables and return to the caller.
 */
int
endKbdMacro(int f, int n)
{
    if (kbdmode == PLAY)
        return TRUE ;
    if (kbdmode == RECORD)
    {
        addModeToWindows(WFMODE) ;  /* and update ALL mode lines */
        mlwrite(0,"[End macro]");
        kbdmode = STOP;

        lkbdlen = kbdlen ;
        return TRUE ;
    }
    return mlwrite(MWABORT,"Macro not active");
}

/*
 * Execute a macro.
 * The command argument is the number of times to loop. Quit as soon as a
 * command gets an error. Return TRUE if all ok, else FALSE.
 */
int
execKbdMacro(int f, int n)
{
    if (kbdmode != STOP)
        return mlwrite(MWABORT,"Macro already active!");
    if (n <= 0)
        return TRUE ;
    kbdptr = lkbdptr ;
    kbdlen = lkbdlen ;
    kbdoff = 0 ;
    kbdrep = n;         /* remember how many times to execute */
    kbdmode = PLAY;     /* start us in play mode */
    return TRUE ;
}

int
stringExec(int f, int n, unsigned char *macro)
{
    unsigned char *okbdptr;
    unsigned char oldcle ;      /* old contents of clexec flag */
    char         *oldestr ;	/* original exec string */
    int  okbdlen, okbdoff, okbdmode, okbdrep ;
    int  oldexec, ii ;
    
    okbdptr = kbdptr ;
    okbdoff = kbdoff ;
    okbdlen = kbdlen ;
    okbdrep = kbdrep ;
    okbdmode = kbdmode ;
    oldcle = clexec;			/* save old clexec flag */
    clexec = FALSE;			/* in cline execution */
    oldestr = execstr;	/* save last ptr to string to execute */
    execstr = NULL ;
    oldexec = execlevel ;
    execlevel = 0;			/* Reset execution level */
    
    kbdptr = macro ;
    kbdoff = 0 ;
    kbdlen = strlen((char *) macro) ;
    kbdrep = n ;
    kbdmode = PLAY;     /* start us in play mode */
    ii = TRUE ;
    while((kbdrep > 1) || (kbdoff < kbdlen))
    {
        if(TTbreakFlag || (kbdmode != PLAY))
        {
            ii = ABORT ;
            break ;
        }
        doOneKey() ;
    }
    kbdptr = okbdptr ;
    kbdoff = okbdoff ;
    kbdlen = okbdlen ;
    kbdrep = okbdrep ;
    kbdmode = okbdmode ;
    execlevel = oldexec ;
    clexec = oldcle;			/* restore clexec flag */
    execstr = oldestr;
    
    return ii ;
}

int
execString(int f, int n)
{
    unsigned char sbuf[MAXBUF] ;
    
    if(mlreply("Enter string", 0, 0, (char *) sbuf, MAXBUF) != TRUE)
        return 0 ;
    return stringExec(f,n,sbuf) ;
}

/*
 * Asks the user if they wish to continue macro execution.
 * does nothing when recording, must STOP the macro execution so 
 * mlyesno goes to the keyboard.
 */
int
macroQuery(int f, int t)
{
    if(kbdmode == PLAY)
    {
        int   rr ;
        
        /* force a screen update */
        update(TRUE);
        kbdmode = STOP ;
        if((rr = mlyesno("Continue macro")) == ABORT)
            return ABORT ;
        if(rr == FALSE)
        {
            kbdrep-- ;
            kbdoff = 0 ;
        }
        if(kbdrep)
            kbdmode = PLAY ;
        else
            return FALSE ;
    }
    else if(kbdmode != RECORD)
        return mlwrite(MWABORT,"Not defining macro") ;
    return TRUE ;
}
    

static meMACRO *
createMacro(char *name)
{
    char     buff[MAXBUF] ;
    register meMACRO *mac ;
    register LINE    *hlp ;
    register short    idx ;
    
    /* If the macro name has not been give then try and get one */
    if((name == NULL) && 
       (mlreply("Enter macro name", MLCOMMAND, 0, buff, MAXBUF) == TRUE) && (buff[0] != '\0'))
        name = buff ;
    
    if((name == NULL) || ((hlp = lalloc(0)) == NULL))
        return NULL ;
    
    /* if it already exists */
    if((idx = decode_fncname(name,1)) >= 0)
    {
        /* if function return error, else clear the buffer */ 
        if(idx < CK_MAX)
        {
            mlwrite(MWABORT|MWPAUSE,"Error! can't re-define a base function") ;
            meFree(hlp) ;
            return NULL ;
        }
        mac = getMacro(idx) ;
        if(!(mac->hlp->l_flag & MACEXEC))
        {
            if(mac->hlp->l_flag & MACFILE)
            {
                if(meNullFree(mac->fname))
                    mac->fname = NULL ;
            }
            freeLineLoop(mac->hlp,1) ;
        }
    }
    else
    {
        if(!(cmdTableSize & 0x1f))
        {
            /* run out of room in hte command table, malloc more */
            extern meCMD *__cmdTable[] ;
            meCMD **nt ;
            if((nt = (meCMD **) meMalloc((cmdTableSize+0x20)*sizeof(meCMD *))) == NULL)
                return NULL ;
            memcpy(nt,cmdTable,cmdTableSize*sizeof(meCMD *)) ;
            if(cmdTable != __cmdTable)
                free(cmdTable) ;
            cmdTable = nt ;
        }
        if(((mac = (meMACRO *) meMalloc(sizeof(meMACRO))) == NULL) ||
           ((mac->name = strdup(name)) == NULL))
        {
            meFree(hlp) ;
            return NULL ;
        }
        mac->varList.head = NULL ;
        mac->varList.count = 0 ;
        mac->id = cmdTableSize ;
        mac->fname = NULL ;
        mac->callback = 0xffffffff ;
        cmdTable[cmdTableSize++] = (meCMD *) mac ;
        mac->hnext = NULL ;
        /* insert macro into the hash table and alphabetic list */
        {
            extern unsigned long cmdHashFunc (register char *cmdName) ;
            unsigned long key ;
            meCMD **cmd ;
            
            key = cmdHashFunc(name) ;
            cmd = &(cmdHash[key]) ;
            while(*cmd != NULL)
                cmd = &((*cmd)->hnext) ;
            *cmd = (meCMD *) mac ;
            cmd = &(cmdHead) ;
            while((*cmd != NULL) && (strcmp((*cmd)->name,name) < 0))
                cmd = &((*cmd)->anext) ;
            mac->anext = *cmd ;
            *cmd = (meCMD *) mac ;
        }
    }
    mac->hlp = hlp ;
    hlp->l_fp = hlp ;
    hlp->l_bp = hlp ;
    return mac ;
}
/* macroDefine:	Set up a macro buffer and flag to store all
		executed command lines there			*/

int
macroDefine(int f, int n)
{
    register meMACRO *mac ;	/* pointer to macro */
    
    if((mac=createMacro(NULL)) == NULL)
        return FALSE ;
    if(n == 0)
        mac->hlp->l_flag |= MACHIDE ;
    
    /* and set the macro store pointers to it */
    mcStore = 1 ;
    lpStore = mac->hlp ;
    return TRUE ;
}


/* macroHelpDefine:
 * Set up a macro help definition
 */
int
macroHelpDefine(int f, int n)
{
    char     name[MAXBUF] ;
    char     sect[20] ;
    
    if(mlreply("Enter name", MLCOMMAND, 0, name+4, MAXBUF-4) != TRUE)
        return FALSE ;
    sect[0] = '\0' ;
    if(mlreply("Enter section", 0, 0, sect, 20) == ABORT)
        return FALSE ;
    /* and set the macro store pointers to it */
    lpStore = lforw(&helpStore) ;
    name[0] = '!' ;
    name[1] = (f) ? '0'+n:' ' ;
    if(sect[0] == '\0')
    {
        name[2] = ' ' ; 
        name[3] = ' ' ; 
    }
    else
    {
        name[2] = sect[0] ; 
        name[3] = (sect[1] == '\0') ? ' ':sect[1] ;
    }
    addLine(lpStore,name) ;
    mcStore = 2 ;
    return TRUE ;
}


/* macroFileDefine:
 * Set up a macro buffer and flag to store all
 * executed command lines there			*/

int
macroFileDefine(int f, int n)
{
    char     fname[MAXBUF] ;
    register meMACRO *mac ;	/* pointer to macro */
    int      ii=0 ;
    
    if(mlreply("Enter file", MLCOMMAND, 0, fname, MAXBUF) != TRUE)
        return FALSE ;
    while((mac=createMacro(NULL)) != NULL)
    {
        if(n == 0)
            mac->hlp->l_flag |= MACHIDE ;
        mac->hlp->l_flag |= MACFILE ;
        mac->fname = strdup(fname) ;
        ii++ ;
    }
    if(ii == 0)
    {
        if((mac=createMacro(fname)) == NULL)
            return FALSE ;
        if(n == 0)
            mac->hlp->l_flag |= MACHIDE ;
        mac->hlp->l_flag |= MACFILE ;
    }
    return TRUE ;
}


int
nameKbdMacro(int f, int n)
{
    char buf[MAXBUF] ;
    int  ss ;
    
    if (kbdmode != STOP)
        return mlwrite(MWABORT,"Macro already active!");
    if(lkbdlen <= 0)
        return mlwrite(MWABORT,"No macro defined!") ;
    if((ss=macroDefine(FALSE, TRUE)) == TRUE)
    {
        strcpy(buf,"execute-string \"") ;
        n = expandexp(lkbdlen,lkbdptr,MAXBUF-2,16,buf,-1,NULL,1) ;
        strcpy(buf+n,"\"") ;
        addLine(lpStore,buf) ;
    }
    mcStore = 0 ;
    lpStore = NULL ;
    return ss ;
}

meMACRO *
userGetMacro(char *buf, int len)
{
    register short   idx ;
    
    if(mlreply("Enter macro name ", MLCOMMAND,2,buf,len) == TRUE)
    {        
        if((idx = decode_fncname(buf,0)) < 0)
            mlwrite(MWABORT,"%s not defined",buf) ;
        else if(idx < CK_MAX)
            mlwrite(MWABORT,"%s is a command",buf) ;
        else
            return getMacro(idx) ;
    }
    return NULL ;
}

int
insMacro(int f, int n)
{
    WINDOW          *wp ;
    register LINE   *lp, *slp ;
    meMACRO         *mac ;
    char             buf[MAXBUF] ;
    long             nol, lineNo, len ;
    int              ii ;
    
    strcpy(buf,"define-macro ") ;
    if((mac=userGetMacro(buf+13, MAXBUF-13)) == NULL)
        return FALSE ;
    
    if((ii=bchange()) != TRUE)               /* Check we can change the buffer */
        return ii ;
    curwp->w_doto = 0 ;
    slp = curwp->w_dotp ;
    lineNo = curwp->line_no ;
    nol = addLine(slp,buf) ;
    len = strlen(buf) + 9 ;
    lp = lforw(mac->hlp);            /* First line.          */
    while (lp != mac->hlp)
    {
        nol += addLine(slp,(char *) lp->l_text) ;
        len += llength(lp) + 1 ;
        lp = lforw(lp);
    }
    nol += addLine(slp,"!emacro") ;
    curbp->elineno += nol ;
    for (wp=wheadp; wp!=NULL; wp=wp->w_wndp)
        if (wp->w_bufp == curbp)
        {
            if(wp->line_no >= lineNo)
                wp->line_no += nol ;
            if(wp->mlineno >= lineNo)
                wp->mlineno += nol ;
            wp->w_flag |= WFMAIN|WFMOVEL ;
        }
#if MEUNDO
    if(meModeTest(curbp->b_mode,MDUNDO))
        meUndoAddInsChars(len) ;
#endif
    return TRUE ;
}
  

