/*****************************************************************************
*
*	Title:		%M%
*
*	Synopsis:	
*
******************************************************************************
*
*	Filename:		%P%
*
*	Author:			Danial Lawrence
*
*	Creation Date:		07/05/91 08:19		<160699.0735>
*
*	Modification date:	%G% : %U%
*
*	Current rev:		%I%
*
*	Special Comments:	
*
*	Contents Description:	
*
*	The routines in this file deal with the region, that magic space
*	between "." and mark. Some functions are commands. Some functions are
*	just for internal use.
*
****************************************************************************** 
* 
* Modifications to the original file by Jasspa. 
* 
* Copyright (C) 1988 - 1999, JASSPA 
* 
* The MicroEmacs Jasspa distribution can be copied and distributed freely for
* any non-commercial purposes. The MicroEmacs Jasspa Distribution can only be
* incorporated into commercial software with the expressed permission of
* JASSPA.
* 
****************************************************************************/

/*---	Include defintions */

#define	__REGIONC			/* Define filename */

/*---	Include files */

#include "emain.h"

/*
 * This routine figures out the
 * bounds of the region in the current window, and
 * fills in the fields of the "REGION" structure pointed
 * to by "rp". Because the dot and mark are usually very
 * close together, we scan outward from dot looking for
 * mark. This should save time. Return a standard code.
 * Callers of this routine should be prepared to get
 * an "ABORT" status; we might make this have the
 * conform thing later.
 */
int
getregion(register REGION *rp)
{
    LINE *lp, *elp;
    long  size;
    
    if (curwp->w_markp == NULL)
        return noMarkSet() ;
    if(curwp->w_dotp == curwp->w_markp)
    {
        rp->linep = curwp->w_dotp;
        if (curwp->w_doto < curwp->w_marko) 
        {
            rp->offset = curwp->w_doto;
            rp->size = (long)(curwp->w_marko-curwp->w_doto);
        }
        else
        {
            rp->offset = curwp->w_marko;
            rp->size = (long)(curwp->w_doto-curwp->w_marko);
        }
        rp->line_no = curwp->line_no;
        return TRUE ;
    }
    if(curwp->line_no < curwp->mlineno)
    {
        elp = curwp->w_markp ;
        lp = curwp->w_dotp ;
        rp->line_no = curwp->line_no ;
        size = curwp->w_doto ;
        rp->offset = size ;
        size = curwp->w_marko + llength(lp) - size -
                  curwp->line_no + curwp->mlineno ;
    }
    else
    {
        elp = curwp->w_dotp ;
        lp = curwp->w_markp ;
        rp->line_no = curwp->mlineno ;
        size = curwp->w_marko ;
        rp->offset = size ;
        size = curwp->w_doto + llength(lp) - size +
                  curwp->line_no - curwp->mlineno ;
    }
    rp->linep = lp ;
    while((lp=lforw(lp)) != elp)
        size += llength(lp) ;
    rp->size = size ;
    return TRUE ;
}

/*
 * Kill the region. Ask "getregion"
 * to figure out the bounds of the region.
 * Move "." to the start, and kill the characters.
 * Bound to "C-W".
 */
int
killRegion(int f, int n)
{
    REGION region;
    
    if(getregion(&region) != TRUE)
        return FALSE ;
    if(bchange() != TRUE)               /* Check we can change the buffer */
        return FALSE ;
    curwp->w_dotp   = region.linep;
    curwp->w_doto   = region.offset;   
    curwp->line_no  = region.line_no;
    
    return ldelete(region.size,3);
}

/*
 * Copy all of the characters in the
 * region to the kill buffer. Don't move dot
 * at all. This is a bit like a kill region followed
 * by a yank. Bound to "M-W".
 */
int	
copyRegion(int f, int n)
{
    LINE          *linep;
    int            left, ii ;
    unsigned char *ss, *dd ;
    REGION         region ;
    
    if((ii=getregion(&region)) != TRUE)
        return ii ;
    if(lastflag != CFKILL)                /* Kill type command.   */
        ksave();
    if((left=region.size) == 0)
    {
        thisflag = CFKILL ;
        return TRUE ;
    }
    
    if((dd = kaddblock(left)) == NULL)
        return FALSE ;
    
    linep = region.linep;                 /* Current line.        */
    if((ii = region.offset) == llength(linep))
        goto add_newline ;                 /* Current offset.      */
    ss = linep->l_text+ii ;
    ii = llength(linep) - ii ;
    goto copy_middle ;
    while(left)
    {
        ss = linep->l_text ;
        ii = llength(linep) ;
copy_middle:
        /* Middle of line.      */
        if(ii > left)
        {
            ii = left ;
            left = 0 ;
        }
        else
            left -= ii ;
        while(ii--)
            *dd++ = *ss++ ;
add_newline:
        if(left != 0)
        {
            *dd++ = meNLCHAR ;
            linep = lforw(linep);
            left-- ;
        }
    }
    thisflag = CFKILL ;
    return TRUE ;
}

/*
 * Lower case region. Zap all of the upper
 * case characters in the region to lower case. Use
 * the region code to set the limits. Scan the buffer,
 * doing the changes. Call "lchange" to ensure that
 * redisplay is done in all buffers. Bound to
 * "C-X C-L".
 */
int
lowerRegion(int f, int n)
{
    LINE *linep ;
    int   loffs ;
    long  lline ;
    register char  c;
    register int   s;
    REGION         region;
    
    if((s=getregion(&region)) != TRUE)
        return (s);
    if((s=bchange()) != TRUE)               /* Check we can change the buffer */
        return s ;
    linep = curwp->w_dotp ;
    loffs = curwp->w_doto ;
    lline = curwp->line_no ;
    curwp->w_dotp = region.linep ;
    curwp->w_doto = region.offset ;
    curwp->line_no = region.line_no ;
    while (region.size--)
    {
        if((c = lgetc(curwp->w_dotp, curwp->w_doto)) == '\0')
        {
            curwp->w_dotp = lforw(curwp->w_dotp);
            curwp->w_doto = 0;
            curwp->line_no++ ;
        }
        else
        {
            if(isUpper(c))
            {
                lchange(WFMAIN);
#if MEUNDO
                meUndoAddRepChar() ;
#endif
                c = toggleCase(c) ;
                lputc(curwp->w_dotp, curwp->w_doto, c);
            }
            (curwp->w_doto)++ ;
        }
    }
    curwp->w_dotp = linep ;
    curwp->w_doto = loffs ;
    curwp->line_no = lline ;
    return TRUE ;
}

/*
 * Upper case region. Zap all of the lower
 * case characters in the region to upper case. Use
 * the region code to set the limits. Scan the buffer,
 * doing the changes. Call "lchange" to ensure that
 * redisplay is done in all buffers. Bound to
 * "C-X C-U".
 */
int
upperRegion(int f, int n)
{
    LINE *linep ;
    int   loffs ;
    long  lline ;
    register char  c;
    register int   s;
    REGION         region;
    
    if((s=getregion(&region)) != TRUE)
        return (s);
    if((s=bchange()) != TRUE)               /* Check we can change the buffer */
        return s ;
    linep = curwp->w_dotp ;
    loffs = curwp->w_doto ;
    lline = curwp->line_no ;
    curwp->w_dotp = region.linep ;
    curwp->w_doto = region.offset ;
    curwp->line_no = region.line_no ;
    while (region.size--)
    {
        if((c = lgetc(curwp->w_dotp, curwp->w_doto)) == '\0')
        {
            curwp->w_dotp = lforw(curwp->w_dotp);
            curwp->w_doto = 0;
            curwp->line_no++ ;
        }
        else
        {
            if(isLower(c))
            {
                lchange(WFMAIN);
#if MEUNDO
                meUndoAddRepChar() ;
#endif
                c = toggleCase(c) ;
                lputc(curwp->w_dotp, curwp->w_doto, c);
            }
            (curwp->w_doto)++ ;
        }
    }
    curwp->w_dotp = linep ;
    curwp->w_doto = loffs ;
    curwp->line_no = lline ;
    return TRUE ;
}

