/*****************************************************************************
*
*	Title:		%M%
*
*	Synopsis:	Spell checking routines
*
******************************************************************************
*
*	Filename:		%P%
*
*	Author:			Steve Phillips
*
*	Creation Date:		21/12/94 		<020699.1636>
*
*	Modification date:	%G% : %U%
*
*	Current rev:		%I%
*
*	Special Comments:	
*
*
*****************************************************************************
* 
* Copyright (C) 1994 - 1999, JASSPA 
* The MicroEmacs Jasspa distribution can be copied and distributed freely for
* any non-commercial purposes. The MicroEmacs Jasspa Distribution can only be
* incorporated into commercial software with the expressed permission of
* JASSPA.
* 
****************************************************************************/

#define	__SPELLC			/* Define filename */

#include <string.h>
#include "emain.h"

#if SPELL

#define DTACTIVE	0x01
#define DTCHNGD		0x02
#define DTCREATE	0x04

#define SP_REG_ROOT  "/history/spell"   /* Root of spell in registry */
#define SP_REGI_SAVE "autosave"         /* Registry item - autosave */

#define TBLINCSIZE 512
#define NOTBLSIZES 11

unsigned long tableSizes[NOTBLSIZES+1] = 
{   211,   419,   811,  1601,  3203, 4801, 6421, 
    8419, 10831, 13001, 16103, 20011
} ;

#define FRSTSPELLRULE 'A'
#define LASTSPELLRULE 'z'
#define NOSPELLRULES  LASTSPELLRULE-FRSTSPELLRULE+1
#define RULE_PREFIX   1
#define RULE_SUFFIX   2
#define RULE_MIXABLE  4

typedef struct meSPELLRULE {
    struct meSPELLRULE *next ;
    unsigned char *ending ;
    unsigned char *remove ;
    unsigned char *append ;
    unsigned char endingLen ;
    unsigned char removeLen ;
    unsigned char appendLen ;
    signed   char changeLen ;
    /* rule is used by find-words for continuation &
     * guessList to eliminate suffixes
     */
    unsigned char rule ;
} meSPELLRULE ;

unsigned char meRuleFlags[NOSPELLRULES]={0} ;
unsigned char meRuleScore[NOSPELLRULES]={0} ;
meSPELLRULE  *meRuleTable[NOSPELLRULES]={0} ;

typedef unsigned char meDICTADDR[3] ;
#define meMAXWORDLEN 127
#define meMAXGUESSES 20
#define meMAXSCORE   60

typedef struct {
    meDICTADDR    next ;
    unsigned char wordLen ;     
    unsigned char flagLen ;     
    unsigned char data[1] ;
} meDICTWORD ;

typedef struct {
    meDICTADDR    next ;
    unsigned char wordLen ;     
    unsigned char flagLen ;     
    unsigned char data[meMAXWORDLEN+meMAXWORDLEN] ;
} meDICTADDWORD ;

#define meDICTWORDSIZE  ((int)((meDICTWORD *)0)->data)

typedef struct meDICTIONARY {
    unsigned char   flags ;
    char           *fname ;
    unsigned long   noWords ;
    unsigned long   tableSize ;
    unsigned long   dSize ;
    unsigned long   dUsed ;
    meDICTADDR     *table ;
    struct meDICTIONARY *next ;
} meDICTIONARY ;

typedef unsigned char meWORDBUF[meMAXWORDLEN] ;
static int           longestPrefixChange=0 ;
static int           longestSuffixRemove=0 ;
static meDICTIONARY *dictHead=NULL ;
static meDICTIONARY *dictIgnr=NULL ;
static meDICTWORD   *lastIgnr=NULL ;
static meDICTWORD   *wordCurr=NULL ;
static int           caseFlags=0 ;
static int           hyphenCheck=1 ;
static int           maxScore=meMAXSCORE ;

/* find-words static variables */
static unsigned char *sfwCurMask=NULL ;
static meDICTIONARY  *sfwCurDict=NULL ;
static int            sfwCurIndx=0 ;
static meDICTWORD    *sfwCurWord=NULL ;
static meSPELLRULE   *sfwPreRule=NULL ;
static meSPELLRULE   *sfwSufRule=NULL ;
static int            sfwFlags=0 ;

#define SPELL_ERROR           0x80

#define SPELL_CASE_FUPPER     0x01
#define SPELL_CASE_FLOWER     0x02
#define SPELL_CASE_CUPPER     0x04
#define SPELL_CASE_CLOWER     0x08

#define toLatinLower(c)       (isUpper(toUserFont(c)) ? toLatinFont(toggleCase(toUserFont(c))):(c))
#define toggleLatinCase(c)    toLatinFont(toggleCase(toUserFont(c)))

#define meEntryGetAddr(ed)       (((ed)[0] << 16) | ((ed)[1] << 8) | (ed)[2])
#define meEntrySetAddr(ed,off)   (((ed)[0] = (unsigned char) ((off)>>16)),((ed)[1] = (unsigned char) ((off)>> 8)),((ed)[2] = (unsigned char) (off)))
#define meWordGetErrorFlag(wd)   (wd->flagLen &  SPELL_ERROR)
#define meWordSetErrorFlag(wd)   (wd->flagLen |= SPELL_ERROR)
#define meWordGetAddr(wd)        meEntryGetAddr((wd)->next)
#define meWordSetAddr(wd,off)    meEntrySetAddr((wd)->next,off)
#define meWordGetWordLen(wd)     (wd->wordLen)
#define meWordSetWordLen(wd,len) (wd->wordLen=len)
#define meWordGetWord(wd)        (wd->data)
#define meWordGetFlagLen(wd)     (wd->flagLen & ~SPELL_ERROR)
#define meWordSetFlagLen(wd,len) (wd->flagLen = meWordGetErrorFlag(wd) | len)
#define meWordGetFlag(wd)        (wd->data+meWordGetWordLen(wd))


unsigned long
hashFunc(unsigned long tableSize, unsigned char *word, int len)
{
    register unsigned long h=0, g ;
    register unsigned char c ;
    
    while(--len >= 0)
    {
        c = *word++ ;
        h = (h << 4) + c ;
        if((g=h & 0xf0000000) != 0)
        {
            h ^= (g >> 24) ;
            h ^= g ;
        }
    }
    return (h % tableSize) ;
}


meDICTWORD *
lookupWord(meDICTIONARY *dict, unsigned char *word, int len)
{
    unsigned long  n, off ;
    meDICTWORD    *ent ;
    int            s, ll ;
    
    n = hashFunc(dict->tableSize,word,len) ;
    off = meEntryGetAddr(dict->table[n]) ;
    
    while(off != 0)
    {
        ent = (meDICTWORD *) (off + (unsigned long) dict->table) ;
        ll = meWordGetWordLen(ent) ;
        if(ll >= len)
        {
            if(ll > len)
                break ;
            s = strncmp((char *) meWordGetWord(ent),(char *)word,ll) ;
            if(s > 0)
                break ;
            if(s == 0)
                return ent ;
        }
        off = meWordGetAddr(ent) ;
    }
    return NULL ;
}

void
dictionaryDeleteWord(meDICTIONARY *dict, unsigned char *word, int len)
{
    meDICTWORD  *ent ;
    if((ent = lookupWord(dict,word,len)) != NULL)
    {
        /* Remove the old entry */
        meWordGetWord(ent)[0] = '\0' ;
        (dict->noWords)-- ;
        dict->flags |= DTCHNGD ;
    }
}

int
dictionaryAddWord(meDICTIONARY *dict, meDICTWORD *wrd)
{
    unsigned long  n, len, off ;
    meDICTWORD    *ent, *lent, *nent ;
    unsigned char  wlen, flen, *word ;
    
    word = meWordGetWord(wrd) ;
    wlen = meWordGetWordLen(wrd) ;
    flen = meWordGetFlagLen(wrd) ;
    if((ent = lookupWord(dict,word,wlen)) != NULL)
    {
        if(flen == 0)
            return TRUE ;
        /* see if the word list already exists in the rule list */
        if(!meWordGetErrorFlag(ent) &&
           !meWordGetErrorFlag(wrd) &&
           ((off=meWordGetFlagLen(ent)) > 0))
        {
            unsigned char *ff, *ef ;
            ef = meWordGetFlag(ent) ;
            ff = meWordGetFlag(wrd) ;
            ff[flen] = '\0' ;
            if(strstr((char *)ef,(char *)ff) != NULL)
                return TRUE ;
            ff[flen++] = '_' ;
            memcpy(ff+flen,ef,off) ;
            flen += off ;
        }
        /* Remove the old entry */
        meWordGetWord(ent)[0] = '\0' ;
        (dict->noWords)-- ;
    }
    dict->flags |= DTCHNGD ;
    len = meDICTWORDSIZE + wlen + flen ;
        
    if((dict->dUsed+len) > dict->dSize)
    {
        dict->dSize = dict->dUsed+len+TBLINCSIZE ;
        if((dict->table = (meDICTADDR *) meRealloc(dict->table,dict->dSize)) == NULL)
            return FALSE ;
    }
    nent = (meDICTWORD *) (dict->dUsed + (unsigned long) dict->table) ;
    nent->wordLen = wrd->wordLen ;
    nent->flagLen = wrd->flagLen ;
    meWordSetFlagLen(nent,flen) ;
    memcpy(meWordGetWord(nent),word,wlen) ;
    memcpy(meWordGetFlag(nent),meWordGetFlag(wrd),flen) ;
    
    lent = NULL ;
    n = hashFunc(dict->tableSize,word,wlen) ;
    off  = meEntryGetAddr(dict->table[n]) ;
    while(off != 0)
    {
        ent = (meDICTWORD *) (off + (unsigned long) dict->table) ;
        if(meWordGetWordLen(ent) >= wlen)
        {
            if((meWordGetWordLen(ent) > wlen) ||
               strncmp((char *) meWordGetWord(ent),(char *) word,wlen) > 0)
                break ;
        }
        off = meEntryGetAddr(ent->next) ;
        lent = ent ;
    }
    if(lent == NULL)
        meEntrySetAddr(dict->table[n],dict->dUsed) ;
    else
        meWordSetAddr(lent,dict->dUsed) ;
    meWordSetAddr(nent,off) ;
    dict->dUsed += len ;
    (dict->noWords)++ ;
    if(dict == dictIgnr)
        lastIgnr = nent ;
    return TRUE ;
}


static void
rehashDictionary(meDICTIONARY *dict)
{
    meDICTWORD     *ent ;
    meDICTADDR     *table, *tbl ;
    unsigned long  tableSize, oldtableSize, ii, 
                   dUsed, dSize, noWords, off ;
    
    oldtableSize = dict->tableSize ;
    noWords = dict->noWords >> 2 ;
    for(ii=0 ; ii<NOTBLSIZES ; ii++)
        if(tableSizes[ii] > noWords)
            break ;
    if((tableSize = tableSizes[ii]) == oldtableSize)
        return ;
    dUsed = sizeof(meDICTADDR)*tableSize ;
    dSize = dict->dUsed+dUsed ;
    
    if((table = (meDICTADDR *) meMalloc(dSize)) == NULL)
        return ;
    memset(table,0,dUsed) ;
    tbl = dict->table ;
    dict->dUsed = dUsed ;
    dict->dSize = dSize ;
    dict->table = table ;
    dict->tableSize = tableSize ;
    dict->noWords = 0 ;
    for(ii=0 ; ii<oldtableSize ; ii++)
    {
        off = meEntryGetAddr(tbl[ii]) ;
        while(off != 0)
        {
            ent = (meDICTWORD *) (off + (unsigned long) tbl) ;
            /* Check this word has not been erased, if not then add */
            if(meWordGetWord(ent)[0] != '\0')
                dictionaryAddWord(dict,ent) ;
            off = meWordGetAddr(ent) ;
        }
    }
    free(tbl) ;
}

            
int
loadDictionary(meDICTIONARY *dict)
{
    FILE           *fp ;
    meDICTADDR     *table, info[2] ;
    unsigned long   dSize, dUsed ;
    
    if((fp = fopen(dict->fname,"rb")) == NULL)
        return mlwrite(MWABORT,"Failed to open dictionary [%s]",dict->fname) ;
    fseek(fp,0,SEEK_END) ;
    dUsed = ftell(fp) ;
    fseek(fp,0,SEEK_SET) ;
    if((fgetc(fp) != 0xED) || (fgetc(fp) != 0xF1))
        return mlwrite(MWABORT,"[%s does not have correct id string]",dict->fname) ;
    fread(info,sizeof(meDICTADDR),2,fp) ;
    dUsed -= ftell(fp) ;
    dSize = dUsed + TBLINCSIZE ;
    
#ifdef BUILD_INSURE_VERSIONS
    table = (meDICTADDR *) meMalloc(10) ;
#endif
    if ((table = (meDICTADDR *) meMalloc(dSize)) == NULL)
        return FALSE ;
    dict->table = table ;
    dict->dUsed = dUsed ;
    dict->dSize = dSize ;
    dict->noWords = meEntryGetAddr(info[0]) ;
    dict->tableSize = meEntryGetAddr(info[1]) ;
    if((dSize = fread(table,1,dUsed,fp)) != dUsed)
    {
        free(table) ;
        return mlwrite(MWABORT,"failed to read dictionary %s",dict->fname) ;
    }
    fclose(fp) ;
    
    dict->flags |= DTACTIVE ;
    rehashDictionary(dict) ;
    
    return TRUE ;
}


int
initDictionaries(void)
{
    meDICTIONARY   *dict, *ldict ;
    
    ldict = NULL ;
    dict = dictHead ;
    if(dict == NULL)
        return mlwrite(MWABORT,"[No dictionaries]") ;
    while(dict != NULL)
    {
        if(!(dict->flags & DTACTIVE) &&
           (loadDictionary(dict) != TRUE))
        {
            if(ldict == NULL)
                dictHead = dict->next ;
            else
                ldict->next = dict->next ;
            meFree(dict->fname) ;
            meFree(dict) ;
            return FALSE ;
        }
        ldict = dict ;
        dict = ldict->next ;
    }
    if(dictIgnr == NULL)
    {
        meDICTADDR    *table ;
        unsigned long  dSize, dUsed ;
        
        dUsed = sizeof(meDICTADDR)*tableSizes[0] ;
        dSize = dUsed + TBLINCSIZE ;
        if(((dictIgnr = (meDICTIONARY *) meMalloc(sizeof(meDICTIONARY))) == NULL) ||
           ((table = (meDICTADDR *) meMalloc(dSize)) == NULL))
            return FALSE ;
        memset(table,0,dUsed) ;
        dictIgnr->dSize = dSize ;
        dictIgnr->dUsed = dUsed ;
        dictIgnr->table = table ;
        dictIgnr->noWords = 0 ;
        dictIgnr->tableSize = tableSizes[0] ;
        dictIgnr->flags = DTACTIVE ;
        dictIgnr->fname = NULL ;
        dictIgnr->next = NULL ;
    }
    return TRUE ;
}

meDICTIONARY *
findDictionary(int flag)
{
    meDICTIONARY *dict ;
    char     fname[FILEBUF], tmp[FILEBUF] ;
    int      found ;
    
    if(mlreply("Dictionary name",MLFILECASE,0,tmp,FILEBUF) != TRUE)
        return FALSE ;
    
    if(!fileLookup(tmp,".edf",meFL_CHECKDOT|meFL_USESRCHPATH,fname))
    {
        strcpy(fname,tmp) ;
        found = 0 ;
    }
    else
        found = 1 ;
    dict = dictHead ;
    while(dict != NULL)
    {
        if(!fnamecmp(dict->fname,fname))
            return dict ;
        dict = dict->next ;
    }
    if(!(flag & 1) ||
       (!found && !(flag & 2)))
    {
        mlwrite(MWABORT,"[Failed to find dictionary %s]",fname) ;
        return NULL ;
    }
    if(((dict = (meDICTIONARY *) meMalloc(sizeof(meDICTIONARY))) == NULL) ||
       ((dict->fname = strdup(fname)) == NULL))
        return NULL ;
    dict->table = NULL ;
    dict->flags = 0 ;
    dict->next = dictHead ;
    dictHead = dict ;
    if(!found || (flag & 4))
    {
        meDICTADDR        *table ;
        unsigned long   tableSize, dSize, dUsed ;
        tableSize = tableSizes[0] ;
        dUsed = sizeof(meDICTADDR)*tableSize ;
        dSize = dUsed + TBLINCSIZE ;
        if((table = (meDICTADDR *) meMalloc(dSize)) == NULL)
            return NULL ;
        memset(table,0,dUsed) ;
        dict->dSize = dSize ;
        dict->dUsed = dUsed ;
        dict->table = table ;
        dict->noWords = 0 ;
        dict->tableSize = tableSize ;
        dict->flags = (DTACTIVE | DTCREATE) ;
    }
    return dict ;
}

void
wordDump(meDICTWORD *ent, unsigned char *buff)
{
    unsigned char  *ss=buff ;
    int len ;
    
    len = meWordGetWordLen(ent) ;
    memcpy(ss,meWordGetWord(ent),len) ;
    ss += len ;
    if((len = meWordGetFlagLen(ent)) > 0)
    {
        if(meWordGetErrorFlag(ent))
            *ss++ = '>' ;
        else
            *ss++ = '/' ;
        memcpy(ss,meWordGetFlag(ent),len) ;
        ss += len ;
    }
    *ss = '\0' ;
}

int
addDict(int f, int n)
{
    meDICTIONARY *dict ;
    
    if     (n == 0) f = 7 ;
    else if(n >  0) f = 3 ;
    else            f = 1 ;
    if((dict = findDictionary(f)) == NULL)
        return FALSE ;
    if(n < 0)
    {
        meDICTADDR    *tbl ;
        unsigned long  tableSize, ii, off ;
        
        if(meModeTest(curbp->b_mode,MDVIEW))
            /* don't allow character insert if in read only */
            return (rdonly()) ;
        if(initDictionaries() != TRUE)
            return ABORT ;
        
        lsinsert(0,"Dictionary: ") ;
        lsinsert(0,dict->fname) ;
        lnewline() ;
        lnewline() ;
        
        tableSize = dict->tableSize ;
        tbl = dict->table ;
        curwp->w_doto = 0 ;
        for(ii=0 ; ii<tableSize ; ii++)
        {
            off = meEntryGetAddr(tbl[ii]) ;
            while(off != 0)
            {
                meDICTWORD    *ent ;
                unsigned char  buff[MAXBUF] ;
                
                ent = (meDICTWORD *) (off + (unsigned long) tbl) ;
                off = meWordGetAddr(ent) ;
                if(meWordGetWord(ent)[0] != '\0')
                {
                    wordDump(ent,buff) ;
                    lsinsert(0,(char *) buff) ;
                    lnewline() ;
                }
            }
        }
    }
    return TRUE ;
}


int
addSpellRule(int f, int n)
{
    meSPELLRULE  *rr ;
    if(n == 0)
    {
        unsigned char *ss=NULL ;
        for(n=0 ; n<NOSPELLRULES ; n++)
        {
            while((rr=meRuleTable[n]) != NULL)
            {
                meRuleTable[n] = rr->next ;
                if(ss != rr->ending)
                {
                    ss = rr->ending ;
                    free(ss) ;
                }
                free(rr->remove) ;
                free(rr->append) ;
                free(rr) ;
            }
            meRuleFlags[n] = 0 ;
        }
        hyphenCheck = 1 ;
        maxScore = meMAXSCORE ;
        longestPrefixChange = 0 ;
        /* reset the find-words */
        if(sfwCurMask != NULL)
        {
            free(sfwCurMask) ;
            sfwCurMask = NULL ;
        }
    }
    else
    {
        meSPELLRULE   *pr ;
        unsigned char *ss, cc, bb ;
        int  rule ;
        char buff[MAXBUF] ;
        
        if((rule = mlCharReply("Rule flag: ",0,NULL,NULL)) < 0)
            return ABORT ;
        if((f == FALSE) && (rule == '-'))
        {
            if((rule = mlyesno("Enable hyphen")) == ABORT)
                return ABORT ;
            hyphenCheck = rule ;
            return TRUE ;
        }
        else if((f == FALSE) && (rule == '#'))
        {
            if(mlreply("Guess score",MLNOSPACE,0,buff,MAXBUF) != TRUE)
                return ABORT ;
            maxScore = meAtoi(buff) ;
            return TRUE ;
        }
        if((rule < FRSTSPELLRULE) || (rule > LASTSPELLRULE) || (rule == '_'))
            return mlwrite(MWABORT,"[Invalid spell rule flag]") ;
            
        if(((rr = meMalloc(sizeof(meSPELLRULE))) == NULL) || 
           (mlreply("Ending",MLNOSPACE,0,buff,MAXBUF) != TRUE) ||
           ((rr->ending = (unsigned char *) strdup(buff)) == NULL) ||
           (mlreply("Remove",MLNOSPACE,0,buff,MAXBUF) != TRUE) ||
           ((rr->remove = (unsigned char *) strdup(buff)) == NULL) ||
           (mlreply("Append",MLNOSPACE,0,buff,MAXBUF) != TRUE) ||
           ((rr->append = (unsigned char *) strdup(buff)) == NULL))
            return ABORT ;
        rule -= FRSTSPELLRULE ;
        rr->removeLen = strlen((char *) rr->remove) ;
        rr->appendLen = strlen((char *) rr->append) ;
        rr->changeLen = rr->appendLen - rr->removeLen ;
        f = 0 ;
        bb = 1 ;
        ss = rr->ending ;
        while((cc=*ss++) != '\0')
        {
            if(cc == '[')
                bb = 0 ;
            else if(cc == ']')
                bb = 1 ;
            if(bb)
                f++ ;
        }
        if(!bb)
            return mlwrite(MWABORT,"[Spell rule - ending has no closing ']']") ;
        rr->endingLen = f ;
        if(n < 0)
        {
            meRuleFlags[rule] = RULE_PREFIX ;
            if(rr->changeLen > longestPrefixChange)
                longestPrefixChange = rr->changeLen ;
            n = -n ;
        }
        else
            meRuleFlags[rule] = RULE_SUFFIX ;
        if(n & 2)
            meRuleFlags[rule] |= RULE_MIXABLE ;
        if((pr = meRuleTable[rule]) == NULL)
        {
            rr->next = meRuleTable[rule] ;
            meRuleTable[rule] = rr ;
        }
        else
        {
            for(;; pr = pr->next)
            {
                if(!strcmp((char *) rr->ending,(char *) pr->ending))
                {
                    rr->ending = pr->ending ;
                    break ;
                }
                if(pr->next == NULL)
                    break ;
            }
            rr->next = pr->next ;
            pr->next = rr ;
        }
    }
    return TRUE ;
}

int
saveDictionary(meDICTIONARY *dict, int n)
{
    REGHANDLE   reg ;
    int         ii ;
    
    if(!(dict->flags & DTCHNGD))
        return TRUE ;
    
    /* Never auto-save created dictionaries */
    if((dict->flags & DTCREATE) ||
       ((n & 0x01) && 
        (((reg=regFind(NULL,SP_REG_ROOT "/" SP_REGI_SAVE))==NULL) || (regGetLong(reg,0) == 0))))
    {
        char prompt[MAXBUF] ;
        int  ret ;
        strcpy(prompt,"Save dictionary ") ;
        strcat(prompt,dict->fname) ;
        if((ret = mlyesno(prompt)) == ABORT)
            return ctrlg(FALSE,1) ;
        if(ret == FALSE)
            return TRUE ;
        if(dict->flags & DTCREATE)
        {
            char fname[FILEBUF], *pp, *ss ;
            /* if the ctreated dictionary does not have a complete path then get one */
            if(strrchr(dict->fname,DIR_CHAR) == NULL)
            {
                ss = dict->fname ;
                if(inputFileName("Save to directory",fname,1) != TRUE)
                    return ABORT ;
                pp = fname + strlen(fname) ;
                if(pp[-1] != DIR_CHAR)
                    *pp++ = DIR_CHAR ;
                strcpy(pp,ss) ;
            }
            else
                strcpy(fname,dict->fname) ;
            
            if(((ii=strlen(fname)) < 4) ||
#ifdef _INSENSE_CASE
               icstrcmp(fname+ii-4,".edf")
#else
               strcmp(fname+ii-4,".edf")
#endif
               )
                strcpy(fname+ii,".edf") ;
            free(dict->fname) ;
            dict->fname = strdup(fname) ;
        }
    }
    if((ii=ffWriteFileOpen(dict->fname,0,NULL)) == TRUE)
    {    
        unsigned char header[8] ;
                  
        header[0] = 0xED ;
        header[1] = 0xF1 ;
        meEntrySetAddr(header+2,dict->noWords) ;
        meEntrySetAddr(header+5,dict->tableSize) ;
        
        if((ii=ffWriteFileWrite(8,header,0)) == TRUE)
           ii = ffWriteFileWrite(dict->dUsed,(unsigned char *) dict->table,0) ;
        ffWriteFileClose(0,NULL) ;
        if(ii == TRUE)
        {
            dict->flags &= ~DTCHNGD ;
            return TRUE ;
        }
    }
    return mlwrite(MWABORT|MWPAUSE,"[Failed to write dictionary %s]",dict->fname) ;
}

int
saveDict(int f, int n)
{
    meDICTIONARY   *dict ;
    
    if(n & 0x02)
    {
        dict = dictHead ;
        while(dict != NULL)
        {
            if(saveDictionary(dict,n) != TRUE)
                return ABORT ;
            dict = dict->next ;
        }
    }
    else
    {
        /* when saving a single disable the prompt */
        if(((dict = findDictionary(0)) == NULL) ||
           (saveDictionary(dict,0) != TRUE))
            return FALSE ;
    }

    return TRUE ;
}

void
freeDictionary(meDICTIONARY *dict)
{
    meNullFree(dict->table) ;
    meNullFree(dict->fname) ;
    free(dict) ;
}

/* deleteDict
 * 
 * Argument is a bitmask where:
 * 
 * 0x01 prompt before losing changes
 * 0x02 delete ignore dictionary
 * 0x04 delete all non-ignore dictionaries
 */
int
deleteDict(int f, int n)
{
    /* reset the find-words */
    if(sfwCurMask != NULL)
    {
        free(sfwCurMask) ;
        sfwCurMask = NULL ;
    }
    if(n & 0x04)
    {
        if(dictIgnr != NULL)
        {
            freeDictionary(dictIgnr) ;
            lastIgnr = NULL ;
            dictIgnr = NULL ;
        }
        if(!(n & 0x02))
            /* just remove the ignore */
            return TRUE ;
    }
    
    if(n & 0x02)
    {
        meDICTIONARY *dd ;
        
        while((dd=dictHead) != NULL)
        {
            if((n & 0x01) && (saveDictionary(dd,0x01) != TRUE))
                return FALSE ;
            dictHead = dd->next ;
            freeDictionary(dd) ;
        }
    }
    else
    {
        meDICTIONARY *dict, *dd ;
    
        if((dict = findDictionary(0)) == NULL)
            return FALSE ;
        
        if((n & 0x01) && (saveDictionary(dict,0x01) != TRUE))
            return FALSE ;
        if(dict == dictHead)
            dictHead = dict->next ;
        else
        {
            dd = dictHead ;
            while(dd->next != dict)
                dd = dd->next ;
            dd->next = dict->next ;
        }
        freeDictionary(dict) ;
    }
    return TRUE ;
}


int
getCurWord(meWORDBUF word)
{
    register unsigned char  *dp, c, alnm, nalnm=1 ;
    register int sz=0, alphaCnt=0 ;
    
    dp = (unsigned char *) word ;
    do
    {
        alnm = nalnm ;
        c = lgetc(curwp->w_dotp,curwp->w_doto) ;
        curwp->w_doto++ ;
        *dp++ = c ;
        if(isAlpha(c))
            alphaCnt = 1 ;
        nalnm = isAlphaNum(c) ;
    } while((++sz < meMAXWORDLEN) && 
            (nalnm || (alnm && isSpllExt(c)))) ;
    
    dp[-1] = 0 ;
    curwp->w_doto-- ;
    if(!alphaCnt)
        return -1 ;
    
    return sz - 1 ;
}

#define wordAddPrefixRule(bwd,rr)         memcpy(bwd,rr->append,rr->appendLen) ;
#define wordRemoveAddPrefixRule(bwd,rr)                                          \
do {                                                                             \
    if(rr->removeLen)                                                            \
        memcpy(bwd,rr->remove, rr->removeLen) ;                                  \
} while(0)

#define wordGetSuffixRuleEnd(wd,wlen,rr)  (wd+wlen-rr->removeLen)
#define wordAddSuffixRule(ewd,rr)         strcpy((char *) ewd,(char *) rr->append)
#define wordRemoveAddSuffixRule(ewd,rr)   strcpy((char *) ewd,(char *) rr->remove)

meDICTWORD *
wordTrySuffixes(unsigned char *word, int wlen, unsigned char prefixRule)
{
    meDICTIONARY  *dict ;
    meSPELLRULE   *rr ;
    meDICTWORD    *wd ;
    unsigned char  flags ;
    int ii ;
    
    flags = (prefixRule == 0) ? RULE_SUFFIX:(RULE_SUFFIX|RULE_MIXABLE) ;
    
    /* Try all the suffix rules to see if we can derive the word from another */
    for(ii=0 ; ii<NOSPELLRULES ; ii++)
    {
        if((meRuleFlags[ii] & flags) == flags)
        {
            rr = meRuleTable[ii] ;
            while(rr != NULL)
            {
                unsigned char cc ;
                int alen, nlen, clen, elen ;
                clen = rr->changeLen ;
                elen = rr->endingLen ;
                if(wlen >= elen+clen)
                {
                    nlen = wlen - clen ;
                    alen = strlen((char *) rr->append) ;
                    if(!strncmp((char *) rr->append,(char *) word+wlen-alen,alen))
                    {
                        cc = word[nlen] ;
                        wordRemoveAddSuffixRule(word+wlen-alen,rr) ;
                        if((elen == 0) ||
                           regexpStrCmp((char *) word+nlen-elen,(char *) rr->ending,1))
                        {
                            dict = dictHead ;
                            while(dict != NULL)
                            {
                                if((wd=lookupWord(dict,word,nlen)) != NULL)
                                {
                                    /* check that the word found allows this rule */
                                    unsigned char *flag, *sflag, ff ;
                                    int len, slen ;
                                    
                                    sflag = flag = meWordGetFlag(wd) ;
                                    slen = len = meWordGetFlagLen(wd) ;
                                    while(--len >= 0)
                                    {
                                        if((ff=*flag++) == '_')
                                        {
                                            sflag = flag ;
                                            slen = len ;
                                        }
                                        else if(ff == ii+FRSTSPELLRULE)
                                        {
                                            if(prefixRule == 0)
                                            {
                                                /* no prefix - found it, return */
                                                word[nlen] = cc ;
                                                return wd ;
                                            }
                                            /* check for the prefix rule as well if given */
                                            while(--slen >= 0)
                                            {
                                                if((ff=*sflag++) == '_')
                                                    break ;
                                                if(ff == prefixRule)
                                                {
                                                    /* found it, return */
                                                    word[nlen] = cc ;
                                                    return wd ;
                                                }
                                            }
                                        }
                                    }
                                }
                                dict = dict->next ;
                            }
                        }
                        word[nlen] = cc ;
                        memcpy(word+wlen-alen,rr->append,alen) ;
                    }
                }
                rr = rr->next ;
            }
        }
    }
    return NULL ;
}

meDICTWORD *
wordTryPrefixes(unsigned char *word, int wlen)
{
    meDICTIONARY *dict ;
    meSPELLRULE  *rr ;
    meDICTWORD   *wd=NULL ;
    int ii ;
    
    /* Now try all the prefix rules to see if we can derive the word */
    for(ii=0 ; ii<NOSPELLRULES ; ii++)
    {
        if(meRuleFlags[ii] & RULE_PREFIX)
        {
            rr = meRuleTable[ii] ;
            while(rr != NULL)
            {
                unsigned char *nw, cc ;
                int nlen, elen ;
                
                if((wlen >= rr->appendLen) &&
                   !strncmp((char *) rr->append,(char *) word,rr->appendLen))
                {
                    nlen = wlen - rr->changeLen ;
                    nw = word + rr->changeLen ;
                    wordRemoveAddPrefixRule(nw,rr) ;
                    if((elen=rr->endingLen) > 0)
                    {
                        cc = nw[elen] ;
                        nw[elen] = '\0' ;
                        if(!regexpStrCmp((char *) nw,(char *) rr->ending,1))
                        {
                            nw[elen] = cc ;
                            wordAddPrefixRule(word,rr) ;
                            rr = rr->next ;
                            continue ;
                        }
                        nw[elen] = cc ;
                    }
                    /* Try the word on its own first */
                    dict = dictHead ;
                    while(dict != NULL)
                    {
                        if((wd=lookupWord(dict,nw,nlen)) != NULL)
                        {
                            /* check that the word found allows this rule */
                            unsigned char *flag ;
                            int len ;
                            
                            flag = meWordGetFlag(wd) ;
                            len = meWordGetFlagLen(wd) ;
                            while(--len >= 0)
                                if(*flag++ == ii+FRSTSPELLRULE)
                                    /* yes, lets get out of here */
                                    return wd ;
                        }
                        dict = dict->next ;
                    }
                    if(meRuleFlags[ii] & RULE_MIXABLE)
                    {
                        /* Now try all the suffixes for this prefix */
                        if((wd = wordTrySuffixes(nw,nlen,(unsigned char) (ii+FRSTSPELLRULE))) != NULL)
                            /* yes, lets get out of here */
                            return wd ;
                    }
                    wordAddPrefixRule(word,rr) ;
                }
                rr = rr->next ;
            }
        }
    }
    return NULL ;
}

/* checkWordSearch
 * 
 * Looks in all dictionaries for the given word, then tries all
 * the rules. Returns:
 *   FALSE  Word not found
 *   TRUE   found and word is good
 *   ABORT  found and word is erroneous
 */
int
checkWordSearch(unsigned char *word, int wlen)
{
    meDICTIONARY *dict ;
    meDICTWORD   *wd ;
    
    if((wd=lookupWord(dictIgnr,word,wlen)) == NULL)
    {
        dict = dictHead ;
        while(dict != NULL)
        {
            if((wd=lookupWord(dict,word,wlen)) != NULL)
                break ;
            dict = dict->next ;
        }
        if(wd == NULL)
        {
            /* Now try all the suffixes on the original word */
            wd = wordTrySuffixes(word,wlen,0) ;
        
            if(wd == NULL)
            {
                /* Now try all the prefixes on the original word */
                wd = wordTryPrefixes(word,wlen) ;
                if(wd == NULL)
                    return FALSE ;
            }
        }
    }
    wordCurr = wd ;
    if(meWordGetErrorFlag(wd))
        return ABORT ;
    return TRUE ;
}

/* checkWord
 * 
 * Checks
 * 1) The word is a word
 * 2) Tries checkWordSearch
 * 3) If upper case then capitalises and tries checkWordSearch
 * 4) If capitalised the makes lower case and tries checkWordSearch
 * 
 * If all the above fail then it fails. Returns:
 *   FALSE  Word is okay
 *   TRUE   found and word is good
 *   ABORT  found and word is erroneous
 */
int
checkWord(unsigned char *word, int wlen)
{
    unsigned char *ss, cc ;
    int ii ;
    
    if(caseFlags == 0)
        /* No letters in the word -> not a word -> word okay */
        return TRUE ;
    
    if((ii=checkWordSearch(word,wlen)) != FALSE)
        return ii ;
    
    /* If all upper then capitalise */
    if(caseFlags == (SPELL_CASE_FUPPER|SPELL_CASE_CUPPER))
    {
        ii = wlen ;
        ss = word+1 ;
        while(--ii > 0)
        {
            cc = *ss ;
            *ss++ = toggleLatinCase(cc) ;
        }
        if(checkWordSearch(word,wlen) == TRUE)
            return TRUE ;
    }
    if(caseFlags & SPELL_CASE_FUPPER)
    {
        cc = *word ;
        *word = toggleLatinCase(cc) ;
        if(checkWordSearch(word,wlen) == TRUE)
            return TRUE ;
        *word = cc ;
    }
    /* We failed to find it, restore the case if we changed it */
    if(caseFlags == (SPELL_CASE_FUPPER|SPELL_CASE_CUPPER))
    {
        ii = wlen ;
        ss = word+1 ;
        while(--ii > 0)
        {
            cc = *ss ;
            *ss++ = toggleLatinCase(cc) ;
        }
    }
    return FALSE ;
}

/* checkCurWord
 * 
 * Check for the given word in all dictionaries, returns:
 *   FALSE  Word not found
 *   TRUE   found and word is good
 *   ABORT  found and word is erroneous
 */
int
checkCurWord(unsigned char *word)
{
    unsigned char *ss ;
    int ii, len, hyphen ;
    
    len = strlen((char *) word) ;
    
    if((ii=checkWord(word,len)) != FALSE)
        return ii ;
    if(isSpllExt(word[len-1]))
    {
        len-- ;
        if(checkWord(word,len) == TRUE)
            return TRUE ;
    }
    if(hyphenCheck)
    {
        /* check for hyphenated words such as wise-cracks */
        for(ss=word,hyphen=0 ; len ; len--,ss++)
        {
            if(*ss == '-')
            {
                if((ss != word) &&
                   (checkWord(word,(unsigned char)(ss-word)) != TRUE))
                    return FALSE ;
                word = ss+1 ;
                hyphen = 1 ;
            }
        }
        if(hyphen && 
           (((len=ss-word)==0) || (checkWord(word,(unsigned char)len) == TRUE)))
        {
            wordCurr = NULL ;
            return TRUE ;
        }
    }
    return FALSE ;
}    

unsigned char *
wordApplyPrefixRule(unsigned char *wd, meSPELLRULE *rr)
{
    int            len ;
    unsigned char *nw, cc ;
    
    if((len=rr->endingLen) != 0)
    {
        cc = wd[len] ;
        wd[len] = '\0' ;
        if(!regexpStrCmp((char *) wd,(char *) rr->ending,1))
        {
            wd[len] = cc ;
            return NULL ;
        }
        wd[len] = cc ;
    }
    nw = wd - rr->changeLen ;
    wordAddPrefixRule(nw,rr) ;
    return nw ;
}

unsigned char *
wordApplySuffixRule(unsigned char *wd, int wlen, meSPELLRULE *rr)
{
    unsigned char *ew, len ;
    
    ew = wd+wlen ;
    if((len=rr->endingLen) != 0)
    {
        if((len > wlen) || !regexpStrCmp((char *) ew-len,(char *) rr->ending,1))
            return NULL ;
    }
    ew -= rr->removeLen ;
    wordAddSuffixRule(ew,rr) ;
    return ew ;
}

int
wordTestSuffixRule(unsigned char *wd, int wlen, meSPELLRULE *rr)
{
    unsigned char len ;
    
    if(((len=rr->endingLen) != 0) &&
       ((len > wlen) || !regexpStrCmp((char *) wd+wlen-len,(char *) rr->ending,1)))
       return 0 ;
    return 1 ;
}


#if 0
#define __LOG_FILE
#endif

#ifdef __LOG_FILE
FILE *fp ;
#endif

/* wordGetGuessScore
 * 
 * Given the user source word (swd) and a dictionary word (dwd) the 
 */
int
wordGetGuessScore(unsigned char *swd, int slen, unsigned char *dwd, int dlen, int testFlag)
{
    int  scr=0 ;
    unsigned char cc, lcc='\0', dd, ldd='\0' ;
    
#ifdef __LOG_FILE
    {
        cc = swd[slen] ;
        swd[slen] = '\0' ;
        dd = dwd[dlen] ;
        dwd[dlen] = '\0' ;
        fprintf(fp,"WordScore [%s] with [%s] ",swd,dwd) ;
        swd[slen] = cc ;
        dwd[dlen] = dd ;
    }
#endif
    for(;;)
    {
        if(slen == 0)
        {
            while(--dlen >= 0)
            {
                dd = *dwd++ ;
                if(!isSpllExt(dd))
                    /* ignore missed ' s etc */
                    scr += 22 ;
            }
            break ;
        }
        if(dlen == 0)
        {
            while(--slen >= 0)
            {
                cc = *swd++ ;
                if(!isSpllExt(cc))
                    /* ignore missed ' s etc */
                    scr += 25 ;
            }
            break ;
        }
        cc = *swd ;
        dd = *dwd ;
        dd = toLatinLower(dd) ;
        if(dd == cc) 
        {
            swd++ ;
            slen-- ;
            lcc = cc ;
            dwd++ ;
            dlen-- ;
            ldd = dd ;
            continue ;
        }
        if((dlen > 1) && (dwd[1] == cc))
        {
            if((slen > 1) && (swd[1] == dd) && 
               ((dlen <= slen) || (dwd[2] != dd)))
            {
                /* transposed chars */
                scr += 23 ;
                swd += 2 ;
                slen -= 2 ;
                lcc = dd ;
            }
            else
            {
                if(ldd == dd)
                    /* missing double letter i.e. leter -> letter */
                    scr += 21 ;
                else if(!isSpllExt(dd))
                    /* missed letter i.e. ltter -> letter, ignore missed ' s etc */
                    scr += 22 ;
                swd++ ;
                lcc = cc ;
                if(testFlag && (slen > 1))
                    /* if this is only a test on part of the word the adjust the
                     * other words length as well to keep the end point aligned
                     * this is because of base word tests. when testing the base
                     * of anomaly with anommoly the comparison is between [anommol]
                     * with [anomaly] the realignment caused by the double m means
                     * that it should be between [anommol] and [anomal]
                     */
                    slen -= 2 ;
                else
                    slen-- ;
            }
            dwd += 2 ;
            dlen -= 2 ;
            ldd = cc ;
        }
        else if((slen > 1) && (swd[1] == dd))
        {
            if(lcc == cc)
                /* extra double letter i.e. vowwel -> vowel */
                scr += 24 ;
            else if(!isSpllExt(cc))
                /* extra letter i.e. lertter -> letter, ignore an extra ' etc. */
                scr += 25 ;
            swd += 2 ;
            slen -= 2 ;
            lcc = dd ;
            dwd++ ;
            ldd = dd ;
            if(testFlag && (dlen > 1))
                /* if this is only a test on part of the word the adjust the
                 * other words length as well to keep the end point aligned
                 * this is because of base word tests. when testing the base
                 * of anomaly with anommoly the comparison is between [anommol]
                 * with [anomaly] the realignment caused by the double m means
                 * that it should be between [anommol] and [anomal]
                 */
                dlen -= 2 ;
            else
                dlen-- ;
        }
        else
        {
            scr += 27 ;
            swd++ ;
            slen-- ;
            lcc = cc ;
            dwd++ ;
            dlen-- ;
            ldd = dd ;
        }
        if(scr >= maxScore)
            break ;
    }
#ifdef __LOG_FILE
    fprintf(fp,"scr = %d, len %d\n",scr,dlen-slen) ;
#endif
    return (scr >= maxScore) ? maxScore:scr ;
}

void
wordAddToGuessList(unsigned char *word, int curScr,
                   meWORDBUF *words, int *scores)
{
    int ii, jj ;
    
    for(ii=0 ; ii<meMAXGUESSES ; ii++)
    {
        if(curScr < scores[ii])
        {
            for(jj=meMAXGUESSES-2 ; jj>=ii ; jj--)
            {
                scores[jj+1] = scores[jj] ;
                strcpy((char *) words[jj+1],(char *) words[jj]) ;
            }
            scores[ii] = curScr ;
            strcpy((char *) words[ii],(char *) word) ;
            break ;
        }
        /* check for duplicate */
        if((curScr == scores[ii]) && !strcmp((char *) words[ii],(char *) word))
            break ;
    }
}

int
wordAddSuffixesToGuessList(unsigned char *sword, int slen,
                           unsigned char *bword, int blen,
                           unsigned char *flags, int noflags,
                           meWORDBUF *words, int *scores, 
                           int bscore, int pflags)
{
    meSPELLRULE   *rr ;
    unsigned char *nwd, *lend=NULL ;
    int ii ;
    
    /* try all the allowed suffixes */
    while(--noflags >= 0)
    {
        ii = flags[noflags] - FRSTSPELLRULE ;
        if(((bscore+meRuleScore[ii]) < scores[meMAXGUESSES-1]) &&
           ((meRuleFlags[ii] & pflags) == pflags))
        {
            rr = meRuleTable[ii] ;
            while(rr != NULL)
            {
                if(bscore+rr->rule < scores[meMAXGUESSES-1])
                {
                    int curScr, nlen ;
                    if(lend != rr->ending)
                    {
                        lend = rr->ending ;
                        if(!wordTestSuffixRule(bword,blen,rr))
                        {
                            do {
                                rr = rr->next ;
                            } while((rr != NULL) && (rr->ending == lend)) ;
                            continue ;
                        }
                    }
                    nwd = wordGetSuffixRuleEnd(bword,blen,rr) ;
                    wordAddSuffixRule(nwd,rr) ;
                    nlen = blen+rr->changeLen ;
                    if((curScr = wordGetGuessScore(sword,slen,bword,nlen,0)) < maxScore)
                        wordAddToGuessList(bword,curScr,words,scores) ;
                    wordRemoveAddSuffixRule(nwd,rr) ;
                    if(TTbreakTest(0))
                        return 1 ;
                }
                rr = rr->next ;
            }
        }
    }
    return 0 ;
}


int
wordScoreSuffixes(unsigned char *word, int olen)
{
    meSPELLRULE *rr ;
    int ii, jj, scr, rscr, bscr=maxScore ;
    
    longestSuffixRemove = 0 ;
    if(isSpllExt(word[olen-1]))
        /* if the last letter is a '.' then ignore it, i.e. "anommolies."
         * should compare suffixes with the "ies", not "es."
         */
        olen-- ;
    for(ii=0 ; ii<NOSPELLRULES ; ii++)
    {
        if(meRuleFlags[ii] & RULE_SUFFIX)
        {
            rscr = maxScore ;
            rr = meRuleTable[ii] ;
            while(rr != NULL)
            {
                jj = rr->appendLen ;
                if((jj < olen) &&
                   ((scr = wordGetGuessScore(word+olen-jj,jj,
                                             rr->append,jj,1)) < maxScore))
                {
                    rr->rule = scr ;
                    if(scr < rscr)
                        rscr = scr ;
                    if(rr->removeLen > longestSuffixRemove)
                        longestSuffixRemove = rr->removeLen ;
                }
                else
                    rr->rule = maxScore ;
                rr = rr->next ;
            }
            meRuleScore[ii] = rscr ;
            if(rscr < bscr)
                bscr = rscr ;
        }
    }
    return bscr ;
}

int
wordGetGuessList(unsigned char *word, int olen, meWORDBUF *words)
{
    meDICTIONARY  *dict ;
    meDICTWORD    *ent ;
    meWORDBUF      wbuff ;
    unsigned char *wb, *ww ;
    unsigned long  ii, off ;
    int            scores[meMAXGUESSES], curScr, bsufScr, noWrds ;
    int            wlen, mlen ;
    
#ifdef __LOG_FILE
    fp = fopen("log","w+") ;
#endif
    for(noWrds=0 ; noWrds<meMAXGUESSES ; noWrds++)
    {
        scores[noWrds] = maxScore ;
        words[noWrds][0] = '\0' ;
    }
    mlen = (olen < 3) ? olen:3 ;
    /* check for two words with a missing ' ' or '-' */
    if(olen > 1)
    {
        unsigned char cc ;
        
        ww = wbuff ;
        for(wlen=1 ; wlen < olen ; wlen++)
        {
            cc = resultStr[wlen] ;
            memcpy(wbuff,resultStr+1,wlen) ;
            wbuff[wlen] = '\0' ;
            if(((wlen == 1) && isSpllExt(cc)) ||
               (checkCurWord(wbuff) == TRUE))
            {
                strcpy((char *) wbuff,resultStr+wlen+1) ;
                if(checkCurWord(wbuff) == TRUE)
                {
                    memcpy(wbuff,resultStr+1,wlen) ;
                    strcpy((char *) wbuff+wlen+1,resultStr+wlen+1) ;
                    wbuff[wlen] = ' ' ;
                    wordAddToGuessList(wbuff,22,words,scores) ;
                    if(hyphenCheck && !isSpllExt(cc))
                    {
                        wbuff[wlen] = '-' ;
                        wordAddToGuessList(wbuff,22,words,scores) ;
                    }
                    if(scores[meMAXGUESSES-1] <= 22)
                        break ;
                }
            }
        }
    }
    wb = wbuff+longestPrefixChange ;
    if(caseFlags & (SPELL_CASE_FUPPER|SPELL_CASE_CUPPER))
    {
        unsigned char cc ;
        ww = word ;
        while((cc = *ww) != '\0')
            *ww++ = toLatinLower(cc) ;
    }
    bsufScr = wordScoreSuffixes(word,olen) ;
    dict = dictHead ;
    while(dict != NULL)
    {
        for(ii=0 ; ii<dict->tableSize ; ii++)
        {
            off = meEntryGetAddr(dict->table[ii]) ;
            while(off != 0)
            {
                ent = (meDICTWORD *) (off + (unsigned long) dict->table) ;
                /* Check this word has not been erased or is an error word,
                 * if not then do guess */
                if(!meWordGetErrorFlag(ent) && ((ww=meWordGetWord(ent))[0] != '\0'))
                {
                    int noflags ;
                    
                    wlen = meWordGetWordLen(ent) ;
                    if((noflags=meWordGetFlagLen(ent)) > 0)
                    {
                        unsigned char *flags ;
                        meSPELLRULE *rr ;
                        int jj, ff, nlen ;
                        
                        memcpy(wb,ww,wlen) ;
                        wb[wlen] = '\0' ;
                        flags = meWordGetFlag(ent) ;
                        if((jj = wlen - longestSuffixRemove) < mlen)
                            jj = (wlen < mlen) ? wlen:mlen ;
                        else if(jj > olen)
                            jj = olen ;
                        curScr = wordGetGuessScore(word,jj,wb,jj,1) ;
                        if(curScr < scores[meMAXGUESSES-1])
                        {
                            /* try all suffixes on the word, wordAddSuffixesToGuessList returns
                             * true if the user aborted */
                            if(((curScr+bsufScr) < scores[meMAXGUESSES-1]) &&
                               wordAddSuffixesToGuessList(word,olen,wb,wlen,flags,noflags,words,
                                                          scores,curScr,RULE_SUFFIX))
                                return -1 ;
                            if((curScr = wordGetGuessScore(word,olen,wb,wlen,0)) < scores[meMAXGUESSES-1])
                                wordAddToGuessList(wb,curScr,words,scores) ;
                            if(TTbreakTest(0))
                                return -1 ;
                        }
                        /* try all the allowed prefixes */
                        ff = noflags ;
                        while(--ff >= 0)
                        {
                            if((jj = flags[ff]) == '_')
                                noflags = ff ;
                            else if(meRuleFlags[(jj-=FRSTSPELLRULE)] & RULE_PREFIX)
                            {
                                rr = meRuleTable[jj] ;
                                while(rr != NULL)
                                {
                                    if((ww = wordApplyPrefixRule(wb,rr)) != NULL)
                                    {
                                        nlen = wlen+rr->changeLen ;
                                        if((jj = nlen - longestSuffixRemove) < mlen)
                                            jj = (nlen < mlen) ? nlen:mlen ;
                                        else if(jj > olen)
                                            jj = olen ;
                                        curScr = wordGetGuessScore(word,jj,ww,jj,1) ;
                                        if(curScr < scores[meMAXGUESSES-1])
                                        {
                                            /* try all suffixes on the word */ 
                                            if(((curScr+bsufScr) <  scores[meMAXGUESSES-1]) && (meRuleFlags[jj] & RULE_MIXABLE))
                                            {
                                                int f1, f2 ;
                                                for(f1=f2=0 ; f2<ff ; )
                                                    if(flags[f2++] == '_')
                                                        f1 = f2 ;
                                                if(wordAddSuffixesToGuessList(word,olen,ww,nlen,flags+f1,noflags-f1,
                                                                              words,scores,curScr,RULE_SUFFIX|RULE_MIXABLE))
                                                    return -1 ;
                                            }
                                            if((curScr = wordGetGuessScore(word,olen,ww,nlen,0)) < scores[meMAXGUESSES-1])
                                                wordAddToGuessList(ww,curScr,words,scores) ;
                                            if(TTbreakTest(0))
                                                return -1 ;
                                        }
                                        wordRemoveAddPrefixRule(wb,rr) ;
                                    }
                                    rr = rr->next ;
                                }
                            }
                        }
                    }
                    else if((curScr = wordGetGuessScore(word,olen,ww,wlen,0)) < scores[meMAXGUESSES-1])
                    {
                        memcpy(wb,ww,wlen) ;
                        wb[wlen] = '\0' ;
                        wordAddToGuessList(wb,curScr,words,scores) ;
                    }
                }
                off = meWordGetAddr(ent) ;
            }
        }
        dict = dict->next ;
    }
    for(noWrds=0 ; (noWrds<meMAXGUESSES) && (scores[noWrds] < maxScore) ; noWrds++)
        ;
    {
        unsigned char cc ;
        cc = word[olen-1] ;
        if(isSpllExt(cc))
        {
            /* if the last letter is a '.' etc then loop through the guesses adding one */
            for(ii=0 ; ii<noWrds ; ii++)
            {
                wlen = strlen((char *) words[ii]) ;
                /* must check that its not there already */
                if(words[ii][wlen-1] != cc)
                {
                    words[ii][wlen]   = cc ;
                    words[ii][wlen+1] = '\0' ;
                }
            }
        }
    }
    return noWrds ;
}


#define SPELLWORD_GET     0x01       /* Get the word from the user */
#define SPELLWORD_GETNEXT 0x02       /* Keep going till we come to a problem */
#define SPELLWORD_ADD     0x04       /* Add the given word */
#define SPELLWORD_IGNORE  0x08       /* Add word to the ignore dictionary */
#define SPELLWORD_ERROR   0x10       /* The given word is an erroneous word */
#define SPELLWORD_GUESS   0x20       /* Return the words guest list */
#define SPELLWORD_DERIV   0x40       /* Return the words derivatives */
#define SPELLWORD_DOUBLE  0x80       /* Check for double word */
#define SPELLWORD_INFO    0x100      /* Get info on word */
#define SPELLWORD_DELETE  0x200      /* Delete the given word */

void
spellWordToLatinFont(unsigned char *dd, unsigned char *ss)
{
    unsigned char cc ;
    cc = *ss++ ;
    if(isUpper(cc))
        caseFlags = SPELL_CASE_FUPPER ;
    else if(isLower(cc))
        caseFlags = SPELL_CASE_FLOWER ;
    else
        caseFlags = 0 ;
    *dd++ = toLatinFont(cc) ;
    while((cc=*ss++) != '\0')
    {
        if(isUpper(cc))
            caseFlags |= SPELL_CASE_CUPPER ;
        else if(isLower(cc))
            caseFlags |= SPELL_CASE_CLOWER ;
        *dd++ = toLatinFont(cc) ;
    }
    *dd = '\0' ;
}


void
spellWordToUserFont(unsigned char *dd, unsigned char *ss)
{
    unsigned char cc ;
    
    cc = *ss++ ;
    cc = toUserFont(cc) ;
    if((caseFlags & SPELL_CASE_FUPPER) || (caseFlags == (SPELL_CASE_FLOWER|SPELL_CASE_CUPPER)))
        cc = toUpper(cc) ;
    *dd++ = cc ;
    while((cc=*ss++) != '\0')
    {
        cc = toUserFont(cc) ;
        if(caseFlags == (SPELL_CASE_FUPPER|SPELL_CASE_CUPPER))
            cc = toUpper(cc) ;
        *dd++ = cc ;
    }
    *dd = '\0' ;
}

int
spellWord(int f, int n)
{
    meWORDBUF word ;
    int       len ;
    
    if(initDictionaries() != TRUE)
        return ABORT ;
    
    if(n & SPELLWORD_GET)
    {
        if(mlreply("Enter word", MLNOSPACE,0,resultStr+1,meMAXWORDLEN) != TRUE)
            return ABORT ;
        spellWordToLatinFont(word,(unsigned char *) resultStr+1) ;
    }
    else if(n & SPELLWORD_GETNEXT)
    {
        unsigned char  cc, chkDbl, curDbl ;
        unsigned short soff, eoff ;
        meWORDBUF lword ;
        
        chkDbl = (n & SPELLWORD_DOUBLE) ;
        curDbl = 0 ;
        while((curwp->w_doto > 0) && isAlphaNum(lgetc(curwp->w_dotp,curwp->w_doto)))
            curwp->w_doto-- ;
        for(;;)
        {
            while(((cc = lgetc(curwp->w_dotp,curwp->w_doto)) != '.') && 
                  !isAlphaNum(cc))
            {
                if(!isSpace(cc))
                    curDbl = 0 ;
                if(WforwChar(curwp, 1) == FALSE)
                {
                    resultStr[0] = 'F' ;
                    resultStr[1] = '\0' ;
                    return TRUE ;
                }
            }
            soff = curwp->w_doto ;
            if(getCurWord((unsigned char *) resultStr+1) < 0)
                continue ;
            eoff = curwp->w_doto ;
            if(curDbl && !icstrcmp((char *) lword,resultStr+1))
            {
                resultStr[0] = 'D' ;
                setShowRegion(curbp,curwp->line_no,soff,curwp->line_no,eoff) ;
                curwp->w_flag |= WFMOVEL|WFSELHIL ;
                return TRUE ;
            }
            spellWordToLatinFont(word,(unsigned char *) resultStr+1) ;
            if(checkCurWord(word) != TRUE)
                break ;
            if(chkDbl)
            {
                /* store the last word for double word check */
                strcpy((char *) lword,resultStr+1) ;
                curDbl = 1 ;
            }
        }
        setShowRegion(curbp,curwp->line_no,soff,curwp->line_no,eoff) ;
        curwp->w_flag |= WFMOVEL|WFSELHIL ;
    }
    else
    {
        unsigned char  cc ;
        unsigned short off, soff, eoff ;
        
        soff = off = curwp->w_doto ;
        for(;;soff--)
        {
            cc = lgetc(curwp->w_dotp,soff) ;
            if((soff==0) || isSpllWord(cc))
                    break ;
        }
        if(!isSpllWord(cc))
            return FALSE ;
        while(soff > 0)	/* Back up */
        {
            --soff ;
            cc = lgetc(curwp->w_dotp,soff) ;
            if(!isSpllWord(cc))
            {
                soff++ ;
                break ;
            }
            if(!isAlphaNum(cc) && 
               ((soff == 0) || !isAlphaNum(lgetc(curwp->w_dotp,soff-1))))
                break ;
        }
        /* if the first character is not alphanumeric or a '.' then move
         * on one, this stops misspellings of 'quoted words'
         */
        if(((cc = lgetc(curwp->w_dotp,soff)) != '.') && !isAlphaNum(cc))
            soff++ ;
        curwp->w_doto = soff ;
        len = getCurWord((unsigned char *) resultStr+1) ;
        eoff = curwp->w_doto ;
        curwp->w_doto = off ;
        setShowRegion(curbp,curwp->line_no,soff,curwp->line_no,eoff) ;
        curwp->w_flag |= WFMOVEL|WFSELHIL ;
        if(len < 0)
        {
            resultStr[0] = 'N' ;
            return TRUE ;
        }
        spellWordToLatinFont(word,(unsigned char *) resultStr+1) ;
    }
    len = strlen((char *) word) ;
    if(n & SPELLWORD_ADD)
    {
        meDICTADDWORD  wdbuff ;
        meDICTIONARY  *dict ;
        meDICTWORD    *wd= (meDICTWORD *) &wdbuff ;
        
        memcpy(meWordGetWord(wd),word,len) ;
        meWordSetWordLen(wd,len) ;
        
        if(mlreply("Enter flags", MLNOSPACE,0,(char *) word,meMAXWORDLEN) != TRUE)
            return ABORT ;
        len = strlen((char *) word) ;
        memcpy(meWordGetFlag(wd),word,len) ;
        /* set flag len directly to clear the ERROR flag */
        wd->flagLen = len ;
        if(n & SPELLWORD_ERROR)
            meWordSetErrorFlag(wd) ;
        if(n & SPELLWORD_IGNORE)
            dict = dictIgnr ;
        else
            dict = dictHead ;
        return dictionaryAddWord(dict,wd) ;
    }
    else if(n & SPELLWORD_DELETE)
    {
        meDICTIONARY  *dict ;
        
        if(n & SPELLWORD_IGNORE)
            dictionaryDeleteWord(dictIgnr,word,len) ;
        else
        {
            dict = dictHead ;
            while(dict != NULL)
            {
                dictionaryDeleteWord(dict,word,len) ;
                dict = dict->next ;
            }
        }
        return TRUE ;
    }
    else if(n & SPELLWORD_DERIV)
    {
        meSPELLRULE  *rr ;
        unsigned char buff[MAXBUF], *wd, *nwd, *swd ;
        
        if(n & SPELLWORD_INFO)
        {
            /* dump all derivatives into the current buffer */
            int ii ;
            
            if(meModeTest(curbp->b_mode,MDVIEW))
                /* don't allow character insert if in read only */
                return (rdonly()) ;
            wd = buff+longestPrefixChange ;
            strcpy((char *) wd,(char *) word) ;
            if(checkCurWord(wd) == TRUE)
            {
                lsinsert(0,(char *) wd) ;
                lnewline() ;
            }
            for(ii=0 ; ii<NOSPELLRULES ; ii++)
            {
                rr = meRuleTable[ii] ;
                f = (meRuleFlags[ii] & RULE_PREFIX) ? 1:0 ;
                while(rr != NULL)
                {
                    if(f)
                        nwd = wordApplyPrefixRule(wd,rr) ;
                    else if((swd = wordApplySuffixRule(wd,len,rr)) != NULL)
                        nwd = wd ;
                    else
                        nwd = NULL ;
                    if(nwd != NULL)
                    {
                        strcpy((char *) word,(char *) nwd) ;
                        if(checkCurWord(word) == TRUE)
                        {
                            lsinsert(0,(char *) nwd) ;
                            lnewline() ;
                            if(meRuleFlags[ii] == (RULE_PREFIX|RULE_MIXABLE))
                            {
                                meSPELLRULE  *sr ;
                                int jj, ll ;
                                ll = len + rr->changeLen ;
                                for(jj=0 ; jj<NOSPELLRULES ; jj++)
                                {
                                    if(meRuleFlags[jj] == (RULE_SUFFIX|RULE_MIXABLE))
                                    {                                    
                                        sr = meRuleTable[jj] ;
                                        while(sr != NULL)
                                        {
                                            if((swd = wordApplySuffixRule(nwd,ll,sr)) != NULL)
                                            {
                                                strcpy((char *) word,(char *) nwd) ;
                                                if(checkCurWord(word) == TRUE)
                                                {
                                                    lsinsert(0,(char *) nwd) ;
                                                    lnewline() ;
                                                }
                                                wordRemoveAddSuffixRule(swd,sr) ;
                                            }
                                            sr = sr->next ;
                                        }
                                    }
                                }
                            }
                        }
                        if(f)
                            wordRemoveAddPrefixRule(wd,rr) ;
                        else
                            wordRemoveAddSuffixRule(swd,rr) ;
                    }
                    rr = rr->next ;
                }
            }
            return TRUE ;
        }
        if(((f = mlCharReply("Rule flag: ",0,NULL,NULL)) < FRSTSPELLRULE) || (f > LASTSPELLRULE))
            return ABORT ;
        f -= FRSTSPELLRULE ;
        wd = buff+longestPrefixChange ;
        strcpy((char *) wd,(char *) word) ;

        rr = meRuleTable[f] ;
        f = (meRuleFlags[f] & RULE_PREFIX) ? 1:0 ;
        while(rr != NULL)
        {
            if(f)
                nwd = wordApplyPrefixRule(wd,rr) ;
            else if((nwd = wordApplySuffixRule(wd,len,rr)) != NULL)
                nwd = wd ;
            if(nwd != NULL)
            {
                spellWordToUserFont((unsigned char *) resultStr,nwd) ;
                return TRUE ;
            }
            rr = rr->next ;
        }
        resultStr[0] = '\0' ;
        return TRUE ;
    }
    else if(n & SPELLWORD_GUESS)
    {
        meWORDBUF words[meMAXGUESSES] ;
        unsigned  char *ss, *dd ;
        int nw, cw, ll, ii ;
        
        if((nw = wordGetGuessList(word,len,words)) < 0)
            return ABORT ;
        dd = (unsigned char *) resultStr ;
        *dd++ = '|' ;
        cw = 0 ;
        ll = MAXBUF-3 ;
        while((cw < nw) && ((ii=strlen((char *) (ss=words[cw++]))) < ll))
        {
            spellWordToUserFont(dd,ss) ;
            dd += ii ;
            *dd++ = '|' ;
            ll -= ii+1 ;
        }
        *dd = '\0' ;
        return TRUE ;
    }
    wordCurr = NULL ;
    if(((f=checkCurWord(word)) == ABORT) && (wordCurr != NULL))
    {
        resultStr[0] = 'A' ;
        f = meWordGetFlagLen(wordCurr) ;
        memcpy(resultStr+1,meWordGetFlag(wordCurr),f) ;
        resultStr[f+1] = '\0' ;
    }
    else if(f == TRUE)
        resultStr[0] = 'O' ;
    else
        resultStr[0] = 'E' ;
    if((n & SPELLWORD_INFO) && (wordCurr != NULL))
        wordDump(wordCurr,(unsigned char *) resultStr+1) ;
    return TRUE ;
}


void
findWordsInit(char *mask)
{
    if(sfwCurMask != NULL)
    {
        free(sfwCurMask) ;
        sfwCurMask = NULL ;
    }
    if(initDictionaries() != TRUE)
        return ;
    
    sfwCurWord = NULL ;
    sfwCurDict = dictHead ;
    sfwPreRule = NULL ;
    sfwSufRule = NULL ;
    if((sfwCurMask = meMalloc(strlen(mask)+1)) != NULL)
    {
        unsigned char *ww, cc ;
        spellWordToLatinFont(sfwCurMask,(unsigned char *) mask) ;
        /* must convert to lower as the regexp works in user font */
        ww = sfwCurMask ;
        while((cc = *ww) != '\0')
            *ww++ = toLatinLower(cc) ;
        sfwFlags = (((cc = *sfwCurMask) == '?') || (cc == '*') || (cc == '[')) ;
    }
}

char *
findWordsNext(void)
{
    unsigned long off ;
    unsigned char *wp, *pwp, *swp, *ww, *flags ;
    int len, flen, preRule, sufRule ;
    
    if(sfwCurMask == NULL)
        return abortm ;
    wp = (unsigned char *) evalResult+longestPrefixChange ;
    if(sfwCurWord != NULL)
    {
        /* Get the word */
        len = meWordGetWordLen(sfwCurWord) ;
        ww = meWordGetWord(sfwCurWord) ;
        for(flen=0 ; flen<len ; flen++)
            wp[flen] = toLatinLower(ww[flen]) ;
        wp[len] = '\0' ;
        /* Get the flags */
        flen = meWordGetFlagLen(sfwCurWord) ;
        flags = meWordGetFlag(sfwCurWord) ;
        if(sfwPreRule != NULL)
        {
            pwp = wordApplyPrefixRule(wp,sfwPreRule) ;
            len += sfwPreRule->changeLen ;
            preRule = sfwPreRule->rule ;  
            if(sfwSufRule != NULL)
            {
                sufRule = sfwSufRule->rule ;  
                goto sfwJumpGotPreGotSuf ;
            }
            goto sfwJumpGotPreNotSuf ;
        }
        if(sfwSufRule != NULL)
        {
            sufRule = sfwSufRule->rule ;  
            goto sfwJumpNotPreGotSuf ;
        }
        goto sfwJumpNotPreNotSuf ;
    }
    while(sfwCurDict != NULL)
    {
        for(sfwCurIndx=0 ; sfwCurIndx<sfwCurDict->tableSize ; sfwCurIndx++)
        {
            off = meEntryGetAddr(sfwCurDict->table[sfwCurIndx]) ;
            while(off != 0)
            {
                sfwCurWord = (meDICTWORD *) (off + (unsigned long) sfwCurDict->table) ;
                if(!meWordGetErrorFlag(sfwCurWord) && 
                   ((ww=meWordGetWord(sfwCurWord))[0] != '\0'))
                {
                    /* Get the word */
                    len = meWordGetWordLen(sfwCurWord) ;
                    for(flen=0 ; flen<len ; flen++)
                        wp[flen] = toLatinLower(ww[flen]) ;
                    wp[len] = '\0' ;
                    if((sfwFlags || (wp[0] == sfwCurMask[0])) &&
                       regexpStrCmp((char *) wp,(char *) sfwCurMask,1))
                    {
                        spellWordToUserFont(wp,wp) ;
                        return (char *) wp ;
                    }
                    /* Get the flags */
sfwJumpNotPreNotSuf:
                    if((flen = meWordGetFlagLen(sfwCurWord)) > 0)
                    {
                        flags = meWordGetFlag(sfwCurWord) ;
                        if(sfwFlags || (wp[0] == sfwCurMask[0]))
                        {
                            /* try all the allowed suffixes */
                            for(sufRule=0 ; sufRule<flen ; sufRule++)
                            {
                                if(meRuleFlags[flags[sufRule]-FRSTSPELLRULE] & RULE_SUFFIX)
                                {
                                    sfwSufRule = meRuleTable[flags[sufRule]-FRSTSPELLRULE] ;
                                    while(sfwSufRule != NULL)
                                    {
                                        if((swp = wordApplySuffixRule(wp,len,sfwSufRule)) != NULL)
                                        {
                                            if(regexpStrCmp((char *) wp,(char *) sfwCurMask,1))
                                            {
                                                sfwSufRule->rule = sufRule ;  
                                                spellWordToUserFont(wp,wp) ;
                                                return (char *) wp ;
                                            }
                                            wordRemoveAddSuffixRule(swp,sfwSufRule) ;
                                        }
sfwJumpNotPreGotSuf:
                                        sfwSufRule = sfwSufRule->next ;
                                    }
                                }
                            }
                        }
                        for(preRule=0 ; preRule<flen ; preRule++)
                        {
                            if(meRuleFlags[flags[preRule]-FRSTSPELLRULE] & RULE_PREFIX)
                            {
                                sfwPreRule = meRuleTable[flags[preRule]-FRSTSPELLRULE] ;
                                while(sfwPreRule != NULL)
                                {
                                    if((sfwFlags || (sfwPreRule->append[0] == sfwCurMask[0])) &&
                                       ((pwp = wordApplyPrefixRule(wp,sfwPreRule)) != NULL))
                                    {
                                        len += sfwPreRule->changeLen ;
                                        if(regexpStrCmp((char *) pwp,(char *) sfwCurMask,1))
                                        {
                                            sfwPreRule->rule = preRule ;  
                                            spellWordToUserFont(pwp,pwp) ;
                                            return (char *) pwp ;
                                        }
sfwJumpGotPreNotSuf:
                                        if(meRuleFlags[flags[preRule]-FRSTSPELLRULE] & RULE_MIXABLE)
                                        {
                                            /* try all the allowed suffixes */
                                            for(sufRule=preRule ; sufRule>=0 ; sufRule--)
                                                if(flags[sufRule] == '_')
                                                    break ;
                                            sufRule++ ;
                                            for( ; sufRule<flen ; sufRule++)
                                            {
                                                if(flags[sufRule] == '_')
                                                    break ;
                                                if(meRuleFlags[flags[sufRule]-FRSTSPELLRULE] == (RULE_SUFFIX|RULE_MIXABLE))
                                                {
                                                    sfwSufRule = meRuleTable[flags[sufRule]-FRSTSPELLRULE] ;
                                                    while(sfwSufRule != NULL)
                                                    {
                                                        if((swp = wordApplySuffixRule(pwp,len,sfwSufRule)) != NULL)
                                                        {
                                                            if(regexpStrCmp((char *) pwp,(char *) sfwCurMask,1))
                                                            {
                                                                sfwPreRule->rule = preRule ;  
                                                                sfwSufRule->rule = sufRule ;  
                                                                spellWordToUserFont(pwp,pwp) ;
                                                                return (char *) pwp ;
                                                            }
                                                            wordRemoveAddSuffixRule(swp,sfwSufRule) ;
                                                        }
sfwJumpGotPreGotSuf:
                                                        sfwSufRule = sfwSufRule->next ;
                                                    }
                                                }
                                            }
                                        }
                                        len -= sfwPreRule->changeLen ;
                                        wordRemoveAddPrefixRule(wp,sfwPreRule) ;
                                    }
                                    sfwPreRule = sfwPreRule->next ;
                                }
                            }
                        }
                    }
                }
                off = meWordGetAddr(sfwCurWord) ;
            }
        }
        sfwCurDict = sfwCurDict->next ;
    }
    sfwCurWord = NULL ;
    return emptym ;
}

#endif

