/****************************************************************************
 *	SCCS:		%W%		%G%		%U%
 *
 *	Last Modified :	<990714.0815>
 *
 ****************************************************************************
 *
 * The functions in this file negotiate with the operating system for
 * characters, and write characters in a barely buffered fashion on the display.
 * All operating systems.
 *
 ****************************************************************************
 * 
 * Modifications to the original file by Jasspa. 
 * 
 * Copyright (C) 1988 - 1999, JASSPA 
 * The MicroEmacs Jasspa distribution can be copied and distributed freely for
 * any non-commercial purposes. The MicroEmacs Jasspa Distribution can only be
 * incorporated into commercial software with the expressed permission of
 * JASSPA.
 * 
 ****************************************************************************/
#include "emain.h"
#include "eskeys.h"
#ifndef _WIN32
#include <time.h>
#include <sys/time.h>
#endif

#if (defined _WIN32) || (defined _DOS)
/* executable extension list, in reverse order of dos priority
 * included 4dos's btm files for completeness */
char *execExtensions[noExecExtensions] = {".btm", ".bat", ".exe", ".com"} ;
#endif

/* SpecialChars; This is an array of special characters, below 32 used for
 * rendering boxes, lines etc. The following table provides a conversion
 * between those characters */
char ttSpeChars [TTSPECCHARS] =
{
    ' ', /*  0/0x00 - Undefined */
    ' ', /*  1/0x01 - Undefined */
    ' ', /*  2/0x02 - Undefined */
    'H', /*  3/0x03 - Graphic / Playing card Heart */
    'D', /*  4/0x04 - Graphic / Playing card Diamond */
    'C', /*  5/0x05 - Graphic / Playing card Club */
    'S', /*  6/0x06 - Graphic / Playing card Space */
    ' ', /*  7/0x07 - Undefined */
    ' ', /*  8/0x08 - Undefined */
    ' ', /*  9/0x09 - Undefined */
    ' ', /* 10/0x0a - Undefined */
    '+', /* 11/0x0b - Line Drawing / Bottom right _| */
    '+', /* 12/0x0c - Line Drawing / Top right */
    '+', /* 13/0x0d - Line Drawing / Top left */
    '+', /* 14/0x0e - Line Drawing / Bottom left |_ */
    '+', /* 15/0x0f - Line Drawing / Centre cross + */
    '>', /* 16/0x10 - Cursor Arrows / Right */
    '<', /* 17/0x11 - Cursor Arrows / Left */
    '-', /* 18/0x12 - Line Drawing / Horizontal line - */
    ' ', /* 19/0x13 - Undefined */
    ' ', /* 20/0x14 - Undefined */
    '+', /* 21/0x15 - Line Drawing / Left Tee |- */
    '+', /* 22/0x16 - Line Drawing / Right Tee -| */
    '+', /* 23/0x17 - Line Drawing / Bottom Tee _|_ */
    '+', /* 24/0x18 - Line Drawing / Top Tee -|- */
    '|', /* 25/0x19 - Line Drawing / Vertical Line | */
    ' ', /* 26/0x1a - Graphic  / Filled Gray Box */
    ' ', /* 27/0x1b - undefined */
    ' ', /* 28/0x1c - undefined */
    ' ', /* 28/0x1d - undefined */
    '^', /* 29/0x1e - Cursor Arrows / Up */
    'v'  /* 31/0x1f - Cursor Arrows / Down */
};

void
TTdoBell(int n)
{  
    unsigned int arg ;
    register int index;
    
    if(macbug < -1)
        macbug = 1 ;
    if((index = decode_key(ME_SPECIAL|SKEY_bell,&arg)) >= 0)
        execFuncHidden(ME_SPECIAL|SKEY_bell,index,arg) ;
    else if((n == 0) || ((n == 1) && meModeTest(globMode,MDQUIET)))
    {
        unsigned char scheme=(globScheme/HCTNUM) ;
        pokeScreen(POKE_NOMARK+0x10,TTnrow,TTncol-6,&scheme,
                   (unsigned char *) "[BELL]") ;
        vvideo.video[TTnrow].endp = TTncol ;
        resetCursor() ;
        mlStatus |= MLSTATUS_CLEAR ;
    }
    else
    {
        TTNbell() ;
    }
}
void
TTbell(void)
{  
    if(kbdmode == PLAY)
    {
        kbdmode = STOP ;
        mlwrite(0,"Macro terminated by ringing the bell") ;
    }
    TTinflush() ;
    TTdoBell(1) ;
}

unsigned char meTimerState[NUM_TIMERS]={0} ;/* State of the timers, set or expired */
long          meTimerTime[NUM_TIMERS]={0} ; /* Absolute time of the timers */

#ifdef _MULTI_TIMER

#ifdef _WIN32
static INT meTimerId[NUM_TIMERS]={0} ;      /* Windows timer handles          */
#define meSetMultiTimer(id,tim)  (meTimerId[id]=SetTimer(ttHwnd,id,tim,(TIMERPROC)(LPVOID)NULL))
#define meKillMultiTimer(id)                                                 \
do {                                                                         \
    if(meTimerId[id] != 0)                                                   \
    {                                                                        \
        KillTimer(ttHwnd,meTimerId[id]) ;                                    \
        meTimerId[id] = 0 ;                                                  \
    }                                                                        \
} while(0)

#endif

#endif

#ifdef _SINGLE_TIMER

#ifdef _UNIX
static struct itimerval itv={0} ;           /* itimer interval                */
#define meSetSingleTimer(tim)                                                \
((itv.it_value.tv_sec=tim/1000),(itv.it_value.tv_usec=(tim%1000)*1000),setitimer(ITIMER_REAL,&itv,0))
#define meKillSingleTimer()                                                  \
(itv.it_value.tv_sec=0,itv.it_value.tv_usec=0,setitimer(ITIMER_REAL,&itv,0))
#endif

#endif


#ifdef _MULTI_TIMER
#ifdef _MULTI_NOID_TIMER
void
timerAlarm(int id)
{
    /* the id is the timer id stored in meTimeId not the timer number, convert */
    int ii=NUM_TIMERS ;
    
    /* must kill the timer or it will come back!!! */
    KillTimer(NULL,id) ;
    while(--ii >= 0)
        if(meTimerId[ii] == id)
        {
            meTimerId[ii] = 0 ;
            meTimerState[ii] = (meTimerState[ii] & ~TIMER_SET) | TIMER_EXPIRED;
        }
}
#endif
#endif
    

#if (defined _CONST_TIMER) || (defined _SINGLE_TIMER)

static TIMERBLOCK tblock[NUM_TIMERS]={      /* Timer block storage            */
{ NULL, 0, 0 },
{ NULL, 0, 1 },
{ NULL, 0, 2 },
{ NULL, 0, 3 },
{ NULL, 0, 4 },
{ NULL, 0, 5 }
} ;
TIMERBLOCK *timers = NULL ;                 /* Head of timer list             */

/*FILE *fplog=NULL ;*/

/*
 * timerCheck
 * Check the status of the timers. Invokes the timer callback functions
 * if any of the timers have expired.
 *
 * timerCheck is invoked after a timer has changed status, it schedules
 * the timer fornext time, taking into account the order of the timers
 * on the list.
 *
 * tim may be specified as zero, in which case the current time is
 * obtained.
 */

void
timerCheck(long tim)
{
    TIMERBLOCK *tbp;                    /* Current timer block */
    TIMERBLOCK *ntbp;                   /* Next timer block */

    /* Make sure that there is something to do */
    if((tbp = timers) == NULL)          /* Timers scheduled ?? */
        return ;

    /* Get the current time if not supplied */
    if(tim == 0)
    {
        struct meTimeval tp ;             /* Time interval for current time */

        gettimeofday(&tp,NULL) ;
        tim = ((tp.tv_sec-startTime)*1000) + (tp.tv_usec/1000);
    }

    /* Iterate down the list of timers looking for any that have
     * expired. 
     */
    do
    {
        ntbp = tbp->next ;              /* Save the next timer in list */
        /* Timer expired ?? */
        if (tim+TIMER_MIN >= tbp->abstime)
        {
/*            fprintf(fplog,"Timer %d %d Expired\n",tbp->id,tbp->abstime) ;*/
            /* Report it as expired */
            meTimerState[tbp->id] = (meTimerState[tbp->id] & ~TIMER_SET) | TIMER_EXPIRED ;
            /* Now remove the timer - we know we're first in the list */
            timers = tbp->next ;
        }
        else
        {
#ifdef _SINGLE_TIMER
            /* Set up the timer for next time */
            tim = tbp->abstime - tim;        /* New offset time */
            /*            fprintf(fplog,"Timer %d %d Reset %d\n",tbp->id,tbp->abstime,tim) ;*/
            /* Set a new timer interval */
            meSetSingleTimer(tim) ;
#endif
            break ;
        }
    }
    while((tbp = ntbp) != NULL);
}

#endif


/*
 * timerSet
 * Start a timer, reset or continue the timer. In all eventualities
 * the timer is re-started and will run.
 *
 * id     - handle of the timer.
 * tim    - absolute time of alarm time.
 *          if (tim == 0) Absolute time is unknown but value
 *              is needed in meTimerTime.
 *          if (tim < 0)  Absolute time is unknown and value
                is not needed.
 * offset - offset from the current time to alarm.
 */
void
timerSet (int id, long tim, long offset)
{
#if (defined _CONST_TIMER) || (defined _SINGLE_TIMER)
    TIMERBLOCK *tbp;
    TIMERBLOCK *p;
    TIMERBLOCK *q;    
#ifdef _SINGLE_TIMER
    int mustSet = 0 ;
#endif
#endif
#ifdef _UNIX
    /* On unix, must hold back the signals else we can get in a right state */
    meSigHold() ;
#endif
    
    if(offset < TIMER_MIN)
    {
        /* must increase the time to the min */
        if(tim > 0)
            tim += TIMER_MIN - offset ;
        offset = TIMER_MIN ;
    }
#ifdef _MULTI_TIMER
    {
#ifdef _MULTI_NOID_TIMER
        /* each timer does not keep the id, so kill the current timer if set */
#ifdef _WINCON
        if (meSystemCfg & meSYSTEM_CONSOLE)
#endif
            meKillMultiTimer(id) ;
#endif
        meSetMultiTimer(id,offset) ;
        if(tim == 0)
        {
            struct meTimeval tp ;
            /* Get the absolute time */
            gettimeofday(&tp,NULL) ;
            tim = ((tp.tv_sec-startTime)*1000) + (tp.tv_usec/1000) + offset;
        }
    }
#endif
#if (defined _CONST_TIMER) || (defined _SINGLE_TIMER)
    {
        tbp = &tblock [id];
        /* Kill off the old timer, and unlink from list */
        if(meTimerState[id] & TIMER_SET)
        {
            if (tbp == timers)
            {
                timers = tbp->next;
#ifdef _SINGLE_TIMER
                mustSet = 1 ;
#endif
            }
            else
            {
                /* Search the list for the timer block */
                p = timers ;
                while(p != NULL)
                {
                    if(p->next == tbp)
                    {
                        p->next = tbp->next;
                        break ;
                    }
                    p = p->next ;
                }
            }
        }
        
        if(tim <= 0)
        {
            struct meTimeval tp ;
            /* Get the absolute time */
            gettimeofday(&tp,NULL) ;
            tim = ((tp.tv_sec-startTime)*1000) + (tp.tv_usec/1000) + offset ;
        }
        tbp->abstime = tim ;
        /*    fprintf(fplog,"Timer %d %d Set\n",tbp->id,tbp->abstime) ;*/
        /* Insertion sort the timer block into the list. */
        for (q = NULL, p = timers; p != NULL; p = p->next)
        {
            if (p->abstime > tbp->abstime)
                break;
            q = p;
        }
        
        if (q == NULL)
        {
            timers = tbp;                   /* Link to the head of the list */
#ifdef _SINGLE_TIMER
            /* Set a new timer interval */
            meSetSingleTimer(offset) ;
            mustSet = 0 ;
#endif
        }
        else
            q->next = tbp;                  /* Link into middle of list */
        tbp->next = p;                      /* Fix the next pointer */
    }
#endif
        
    meTimerState[id] = (meTimerState[id] & ~TIMER_EXPIRED) | TIMER_SET ;
    meTimerTime[id] = tim ;
        
#ifdef _SINGLE_TIMER
    if(mustSet)
        /* Call timerCheck to schedule the timer */
        timerCheck(tim-offset) ;
#endif
#ifdef _UNIX
    meSigRelease() ;
#endif
}

/*
 * timerKill
 * Kills a timer if it is currently running. The timer is stopped
 * and removed from the timer list.
 *
 * A macro wrap checks if the timer is running, if it is then
 * it is descheduled, otherwise there is no work to be done.
 */


int
_timerKill (int id)
{
    /* Kill off the old timer */
    if(isTimerSet(id))
    {
#if (defined _CONST_TIMER) || (defined _SINGLE_TIMER)
        TIMERBLOCK *tbp;
        TIMERBLOCK *p;
#endif
        
#ifdef _MULTI_TIMER
        {
            meKillMultiTimer(id) ;
        }
#endif
#if (defined _CONST_TIMER) || (defined _SINGLE_TIMER)
        {
            tbp = &tblock [id];
            
            if (tbp == timers)
            {
                timers = tbp->next;
#ifdef _SINGLE_TIMER
                if(timers == NULL)
                {
                    /* No other timer - kill the timer */
                    meKillSingleTimer() ;
                }
                else if(timers->abstime > tbp->abstime)
                    /* Must change the timer to the next timer which is
                     * a different time
                     */
                    timerCheck(0) ;
#endif
            }
            else
            {
                for (p = timers; p != NULL; p = p->next)
                {
                    if (p->next == tbp)
                    {
                        p->next = tbp->next;
                        break;
                    }
                }
            }
            tbp->next = NULL;
        }
#endif            
    }
    return (meTimerState[id] = meTimerState[id] & ~(TIMER_SET|TIMER_EXPIRED)) ;
}

/* translate-key support code */
#define ME_DELETE_KEY 0xffff           /* an invalid key number used to flag the key is deleted */
meTRANSKEY TTtransKey={0,0,ME_INVALID_KEY,NULL} ;

/* Keyboard buffer */
unsigned short TTkeyBuf [KEYBUFSIZ]={0};
unsigned char  TTnextKeyIdx = KEYBUFSIZ;
unsigned char  TTlastKeyIdx = KEYBUFSIZ;
unsigned char  TTnoKeys = 0;
unsigned char  TTbreakFlag = 0;
unsigned char  TTbreakCnt = TTBREAKCNT;
unsigned char  TTallKeys = 0;           /* Report all keys */

int TTcurr =  0;                 /* Windows current row. */
int TTcurc =  0;                 /* Windows current column */
int TTfocus = 1;                 /* Is the window the current focus */


void
TThandleBlink(int initFlag)
{
    if((cursorState < 0) || !TTfocus)
        timerClearExpired(CURSOR_TIMER_ID) ;
    else
    {
        int ii ;
        
        if(initFlag && blinkState)
        {
            if(initFlag & 2)
            {
                /* The cursor must be redrawn */
                TTshowCur() ;
                TTflush() ;
            }
            ii = cursorBlink & 0xffff ;
        }
        else
        {
            blinkState ^= 1 ;
            if(blinkState)
            {
                TTshowCur() ;
                ii = cursorBlink & 0xffff ;
            }
            else
            {
                TThideCur() ;
                if((ii = cursorBlink >> 16) == 0)
                    ii = cursorBlink ;
            }
            TTflush() ;
        }
        timerSet (CURSOR_TIMER_ID,-1,ii) ;
    }
}

void
TTmove(int row, int col)
{
    if ((row != TTcurr) || (col != TTcurc))
    {
        if((cursorState >= 0) && blinkState)
            TThideCur() ;
        TTcurr = row ;
        TTcurc = col ;
    }
    if((cursorState >= 0) && blinkState)
        TTshowCur() ;
}


#ifdef FREE_ALL_MEMORY

static void
TTfreeTranslateKeyChild(meTRANSKEY *tcapKeys)
{
    int ii ;
    for(ii=0 ; ii<tcapKeys->count ; ii++)
        TTfreeTranslateKeyChild(tcapKeys->child+ii) ;
    meNullFree(tcapKeys->child) ;
}

void
TTfreeTranslateKey(void)
{
    TTfreeTranslateKeyChild(&TTtransKey) ;
}
   
#endif

void
translateKeyAdd(meTRANSKEY *tcapKeys, int count, unsigned short *key, unsigned short map)
{
    if(--count < 0)
        tcapKeys->map = map ;
    else
    {
        unsigned short cc=*key++ ;
        int ii ;
        
        for(ii=0 ; ii<tcapKeys->count ; ii++)
            if(tcapKeys->child[ii].key == cc)
                break ;
        if(ii == tcapKeys->count)
        {
            tcapKeys->child = realloc(tcapKeys->child,(ii+1)*sizeof(meTRANSKEY)) ;
            (tcapKeys->count)++ ;
            tcapKeys->child[ii].key   = cc ;
            tcapKeys->child[ii].count = 0 ;
            tcapKeys->child[ii].map   = ME_INVALID_KEY ;
            tcapKeys->child[ii].child = NULL ;
        }
        translateKeyAdd(tcapKeys->child+ii,count,key,map) ;
    }
}

int
keyListToShorts(unsigned short *sl, unsigned char *kl)
{
    unsigned char dd ;
    int ii=0 ;
    
    for(;;)
    {
        for(;;kl++)
        {
            if((dd=*kl) == '\0')
                return ii ;
            if(!isSpace(dd))
                break ;
        }
        sl[ii++] = getskey(&kl) ;
    }
}

int
charListToShorts(unsigned short *sl, unsigned char *cl)
{
    unsigned char cc ;
    int ii=0 ;
    
    while((cc=*cl++) != 0)
        sl[ii++] = (unsigned short) cc ;
    return ii ;
}

void
translateKeyShow(meTRANSKEY *tcapKeys, BUFFER *bp, char *key, int depth)
{
    int ii, nd ;
    
    if(tcapKeys->map != ME_INVALID_KEY)
    {
        char buf[MAXBUF] ;
        
        strcpy(buf,"    \"") ;
        ii = 5 ;
        strncpy(buf+ii,key,depth) ;
        ii += depth ;
        buf[ii++] = '"' ;
        buf[ii++] = ' ' ;
        if(ii < 35)
        {
            memset(buf+ii,'.',35-ii) ;
            ii = 35 ;
        }
        buf[ii++] = ' ' ;
        buf[ii++] = '"' ;
        if(tcapKeys->map != ME_DELETE_KEY)
            ii += cmdchar(tcapKeys->map,buf+ii);
        buf[ii++] = '"' ;
        buf[ii] = '\0' ;
        addLineToEob(bp,buf) ;
    }
    if(depth)
        key[depth++] = ' ' ;
    for(ii=0 ; ii<tcapKeys->count ; ii++)
    {
        nd = depth + cmdchar(tcapKeys->child[ii].key,key+depth);
        translateKeyShow(tcapKeys->child+ii,bp,key,nd) ;
    }
}

int
translateKey(int f, int n)
{
    static   char  tnkyPrompt[]="translate-tcap-key code" ;
    unsigned char  buf[128] ;
    unsigned short c_from[20], c_to[20] ;
    int ii ;
    
    if(n == 0)
    {
        WINDOW *wp ;
        BUFFER *bp ;
    
        if((wp = wpopup("*translate-key*",(BFND_CREAT|BFND_CLEAR|WPOP_USESTR))) == NULL)
            return FALSE ;
        bp = wp->w_bufp ;
        
        translateKeyShow(&TTtransKey,bp,(char *)buf,0) ;
        bp->b_dotp = lforw(bp->b_linep);
        bp->b_doto = 0 ;
        bp->line_no = 0 ;
        meModeClear(bp->b_mode,MDEDIT) ;    /* don't flag this as a change */
        meModeSet(bp->b_mode,MDVIEW) ;      /* put this buffer view mode */
        resetBufferWindows(bp) ;            /* Update the window */
        mlerase(MWCLEXEC);	                /* clear the mode line */
    }
    else
    {
        strcpy(tnkyPrompt+19,"code") ;
        if((mlreply(tnkyPrompt,0,0,(char *)buf,128) != TRUE) ||
           ((ii=keyListToShorts(c_from,buf)) <= 0))
            return FALSE ;

        if(n == -1)
            c_to[0] = ME_INVALID_KEY ;
        else
        {
            strcpy(tnkyPrompt+19,"to") ;
            if((mlreply(tnkyPrompt,0,0,(char *)buf,128) != TRUE) ||
               ((n=keyListToShorts(c_to,buf)) < 0))
                return FALSE ;
            if(n == 0)
                c_to[0] = ME_DELETE_KEY ;
        }
        translateKeyAdd(&TTtransKey,ii,c_from,*c_to) ;
    }
    return TRUE ;
}

unsigned short
TTgetc(void)
{
    unsigned short cc ;
    
    do {
        if(!TTahead())
            TTwaitForChar() ;
        
        if(TTtransKey.count)
        {
            meTRANSKEY *tt ;
            int         ii, nextKeyIdx, keyNo ;
            
            tt = &TTtransKey ;
            nextKeyIdx = TTnextKeyIdx ;
            keyNo = 0 ;
            
            while(tt->count != 0)
            {
                if(keyNo == TTnoKeys)
                {
                    /* Wait for a quarter a second to see if any more keys are available */
                    /* Only gocha here is that preferably you want an interuptable wait
                     * therefore we must set the TTnoKeys to 0 so we only interrupt on
                     * a receipt of a new key. Must remember to reset it as well.
                     */
                    TTnoKeys = 0 ;
                    TTsleep(250,1) ;
                    TTnoKeys += keyNo ;
                    if(keyNo == TTnoKeys)
                        break ;
                }
                keyNo++ ;
                if(!nextKeyIdx)
                    nextKeyIdx = KEYBUFSIZ ;
                cc = TTkeyBuf[--nextKeyIdx] ;
                for(ii=0 ; ii<tt->count ; ii++)
                {
                    if(tt->child[ii].key == cc)
                    {
                        tt = tt->child+ii ;
                        if(tt->map == ME_DELETE_KEY)
                        {
                            TTnextKeyIdx = nextKeyIdx ;
                            TTnoKeys -= keyNo ;
                            keyNo = 0 ;
                        }
                        else if(tt->map != ME_INVALID_KEY)
                        {
                            TTnextKeyIdx = nextKeyIdx + 1 ;
                            TTnoKeys -= keyNo-1 ;
                            TTkeyBuf[nextKeyIdx] = tt->map ;
                            keyNo = 1 ;
                        }
                        ii = 0 ;
                        break ;
                    }
                }
                if(ii == tt->count)
                    break ;
            }
        }
    } while(!TTnoKeys) ;
    
    if(!TTnextKeyIdx)
        TTnextKeyIdx = KEYBUFSIZ ;
    cc = TTkeyBuf[--TTnextKeyIdx] ;
    TTnoKeys-- ;
    /* remove the abort flag if we are waiting for the key */
    if((cc == 0x07) && TTbreakFlag)
        TTbreakFlag = 0 ;
    return cc ;
}

/*
 * doIdlePickEvent
 * Handle the idle event.
 */
void
doIdlePickEvent(void)
{
    unsigned int arg;           /* Decode key argument */
    register int index;

    /* See if there is any binding on the idle event */
    if((index=decode_key(ME_SPECIAL|SKEY_idle_pick,&arg)) != -1)
    {
        /* Execute the idle-pick key */
        execFuncHidden(ME_SPECIAL|SKEY_idle_pick,index,arg) ;
        
        /* Now see if theres a idle-time event, if so start the timer */
        if(decode_key(ME_SPECIAL|SKEY_idle_time, &arg) != -1)
            timerSet(IDLE_TIMER_ID,-1,idletime);
        else if(decode_key(ME_SPECIAL|SKEY_idle_drop,&arg) != -1)
            meTimerState[IDLE_TIMER_ID] = IDLE_STATE_DROP ;
    }
}

/*
 * doIdleDropEvent
 * Handle the end of an idle event.
 */
void
doIdleDropEvent(void)
{
    unsigned int arg;           /* Decode key argument */
    register int index;

    if(isTimerActive(IDLE_TIMER_ID))
        /* Kill off the timer */
        timerKill(IDLE_TIMER_ID) ;
    
    /* do the idle-drop if bound */
    if((index=decode_key(ME_SPECIAL|SKEY_idle_drop,&arg)) != -1)
        /* Execute the idle-pick key */
        execFuncHidden(ME_SPECIAL|SKEY_idle_drop,index,arg) ;
    meTimerState[IDLE_TIMER_ID] = 0;
}

/*
 * addKeyToBuffer
 * Add a new keyboard character to the keyboard buffer.
 */
void
addKeyToBuffer(unsigned short cc)
{
    static unsigned short lastcc=0 ;
#if MOUSE
    /* Reset the mouse context if mouse interaction is outstanding. 
     * This should only be timer keys
     * Kill off the timer */
    timerKill(MOUSE_TIMER_ID);
#endif
    /* Flag the end of the idle if necessary */
    if(meTimerState[IDLE_TIMER_ID] != 0)
        doIdleDropEvent() ;
    
    /* If cursor is blinking, ensure the cursor is visible */
    if((cursorState >= 0) && cursorBlink)
        TThandleBlink(1) ;
    
    if(!TTallKeys)
    {
        unsigned short modCnt=0 ;
        unsigned short modKey[3] ;
        unsigned int arg ;
        unsigned short kk, ccx=lastcc ^ cc ;

        if(ccx & ME_SHIFT)
        {
            if(lastcc & ME_SHIFT)
                kk = ME_SPECIAL|ME_SHIFT|SKEY_drop ;
            else
                kk = ME_SPECIAL|ME_SHIFT|SKEY_pick ;
            if(decode_key(kk,&arg) != -1)
                modKey[modCnt++] = kk ;
        }
        if(ccx & ME_CONTROL)
        {
            if(lastcc & ME_CONTROL)
                kk = ME_SPECIAL|ME_CONTROL|SKEY_drop ;
            else
                kk = ME_SPECIAL|ME_CONTROL|SKEY_pick ;
            if(decode_key(kk,&arg) != -1)
                modKey[modCnt++] = kk ;
        }
        if(ccx & ME_ALT)
        {
            if(lastcc & ME_ALT)
                kk = ME_SPECIAL|ME_ALT|SKEY_drop ;
            else
                kk = ME_SPECIAL|ME_ALT|SKEY_pick ;
            if(decode_key(kk,&arg) != -1)
                modKey[modCnt++] = kk ;
        }
        lastcc = cc ;
        if(modCnt)
        {
            int ii ;
            if (TTnoKeys+modCnt < KEYBUFSIZ)
            {
                modKey[modCnt] = cc ;
                for(ii=0 ; ii<=modCnt ;ii++)
                {
                    if (!TTlastKeyIdx)
                    TTlastKeyIdx = KEYBUFSIZ;
                    TTkeyBuf [--TTlastKeyIdx] = modKey[ii] ;
                    TTnoKeys++;
                }
            }
            else
                TTNbell() ;
            return ;
        }
    }
    else
        lastcc = cc ;
    /* Add the key to the buffer */
    if (TTnoKeys < KEYBUFSIZ)
    {
        if (!TTlastKeyIdx)
            TTlastKeyIdx = KEYBUFSIZ;
        TTkeyBuf [--TTlastKeyIdx] = cc;
        TTnoKeys++;
    }
    else
        TTNbell() ;
}

